/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#ifndef _HEAVY_WSTD_M3NGLR_H_
#define _HEAVY_WSTD_M3NGLR_H_

#include "HvHeavy.h"

#ifdef __cplusplus
extern "C" {
#endif

#if HV_APPLE
#pragma mark - Heavy Context
#endif

typedef enum {
  HV_WSTD_M3NGLR_PARAM_IN_HIGH = 0xC9CE7B11, // High
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_CRSHR = 0xCFF7A234, // High_Crshr
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_FLDR = 0x12EF45B9, // High_Fldr
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_GAIN = 0xC05E9A15, // High_Gain
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_LMTR = 0x8E596A67, // High_Lmtr
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_MIX = 0xAC5FA57F, // High_Mix
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_SMTHR = 0x585E4A54, // High_Smthr
  HV_WSTD_M3NGLR_PARAM_IN_HIGH_SQNC = 0x14A9747F, // High_Sqnc
  HV_WSTD_M3NGLR_PARAM_IN_LOW = 0x87744544, // Low
  HV_WSTD_M3NGLR_PARAM_IN_LOW_CRSHR = 0xEBDB0C5E, // Low_Crshr
  HV_WSTD_M3NGLR_PARAM_IN_LOW_FLDR = 0xF974E926, // Low_Fldr
  HV_WSTD_M3NGLR_PARAM_IN_LOW_GAIN = 0xBB56D13C, // Low_Gain
  HV_WSTD_M3NGLR_PARAM_IN_LOW_LMTR = 0xB4EFBAC1, // Low_Lmtr
  HV_WSTD_M3NGLR_PARAM_IN_LOW_MIX = 0x91FEF7A, // Low_Mix
  HV_WSTD_M3NGLR_PARAM_IN_LOW_SMTHR = 0x1665AEE9, // Low_Smthr
  HV_WSTD_M3NGLR_PARAM_IN_LOW_SQNC = 0x860CA1D9, // Low_Sqnc
  HV_WSTD_M3NGLR_PARAM_IN_MID = 0xBFDC48B6, // Mid
  HV_WSTD_M3NGLR_PARAM_IN_MID_CRSHR = 0xEA9DEF12, // Mid_Crshr
  HV_WSTD_M3NGLR_PARAM_IN_MID_FLDR = 0xDBF6FE0A, // Mid_Fldr
  HV_WSTD_M3NGLR_PARAM_IN_MID_FREQ = 0x59CF21AC, // Mid_Freq
  HV_WSTD_M3NGLR_PARAM_IN_MID_GAIN = 0x23F5E910, // Mid_Gain
  HV_WSTD_M3NGLR_PARAM_IN_MID_LMTR = 0x33C126B6, // Mid_Lmtr
  HV_WSTD_M3NGLR_PARAM_IN_MID_MIX = 0x676FE235, // Mid_Mix
  HV_WSTD_M3NGLR_PARAM_IN_MID_SMTHR = 0x64B1268F, // Mid_Smthr
  HV_WSTD_M3NGLR_PARAM_IN_MID_SQNC = 0xA91022F8, // Mid_Sqnc
} Hv_WSTD_M3NGLR_ParameterIn;


/**
 * Creates a new patch instance.
 * Sample rate should be positive and in Hertz, e.g. 44100.0.
 */
HeavyContextInterface *hv_WSTD_M3NGLR_new(double sampleRate);

/**
 * Creates a new patch instance.
 * @param sampleRate  Sample rate should be positive (> 0) and in Hertz, e.g. 48000.0.
 * @param poolKb  Pool size is in kilobytes, and determines the maximum amount of memory
 *   allocated to messages at any time. By default this is 10 KB.
 * @param inQueueKb  The size of the input message queue in kilobytes. It determines the
 *   amount of memory dedicated to holding scheduled messages between calls to
 *   process(). Default is 2 KB.
 * @param outQueueKb  The size of the output message queue in kilobytes. It determines the
 *   amount of memory dedicated to holding scheduled messages to the default sendHook.
 *   See getNextSentMessage() for info on accessing these messages. Default is 0 KB.
 */
HeavyContextInterface *hv_WSTD_M3NGLR_new_with_options(double sampleRate, int poolKb, int inQueueKb, int outQueueKb);

/**
 * Free the patch instance.
 */
void hv_WSTD_M3NGLR_free(HeavyContextInterface *instance);


#ifdef __cplusplus
} // extern "C"
#endif

#endif // _HEAVY_WSTD_M3NGLR_H_
