/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "Heavy_WSTD_M3NGLR.hpp"

#include <new>

#define Context(_c) static_cast<Heavy_WSTD_M3NGLR *>(_c)


/*
 * C Functions
 */

extern "C" {
  HV_EXPORT HeavyContextInterface *hv_WSTD_M3NGLR_new(double sampleRate) {
    // allocate aligned memory
    void *ptr = hv_malloc(sizeof(Heavy_WSTD_M3NGLR));
    // ensure non-null
    if (!ptr) return nullptr;
    // call constructor
    new(ptr) Heavy_WSTD_M3NGLR(sampleRate);
    return Context(ptr);
  }

  HV_EXPORT HeavyContextInterface *hv_WSTD_M3NGLR_new_with_options(double sampleRate,
      int poolKb, int inQueueKb, int outQueueKb) {
    // allocate aligned memory
    void *ptr = hv_malloc(sizeof(Heavy_WSTD_M3NGLR));
    // ensure non-null
    if (!ptr) return nullptr;
    // call constructor
    new(ptr) Heavy_WSTD_M3NGLR(sampleRate, poolKb, inQueueKb, outQueueKb);
    return Context(ptr);
  }

  HV_EXPORT void hv_WSTD_M3NGLR_free(HeavyContextInterface *instance) {
    // call destructor
    Context(instance)->~Heavy_WSTD_M3NGLR();
    // free memory
    hv_free(instance);
  }
} // extern "C"







/*
 * Class Functions
 */

Heavy_WSTD_M3NGLR::Heavy_WSTD_M3NGLR(double sampleRate, int poolKb, int inQueueKb, int outQueueKb)
    : HeavyContext(sampleRate, poolKb, inQueueKb, outQueueKb) {
  numBytes += sLine_init(&sLine_F0JuHtHo);
  numBytes += sLine_init(&sLine_wp2rbr12);
  numBytes += sLine_init(&sLine_Ik8joamg);
  numBytes += sLine_init(&sLine_DpAfafPl);
  numBytes += sLine_init(&sLine_xglAVFA6);
  numBytes += sBiquad_init(&sBiquad_s_zodZRcjv);
  numBytes += sLine_init(&sLine_lZjeicNW);
  numBytes += sLine_init(&sLine_kimGO7Ur);
  numBytes += sLine_init(&sLine_nvirZ593);
  numBytes += sLine_init(&sLine_xnRwmx5a);
  numBytes += sLine_init(&sLine_l6oALZOK);
  numBytes += sBiquad_init(&sBiquad_s_Rh0i4upA);
  numBytes += sLine_init(&sLine_Rf2hEX6Q);
  numBytes += sLine_init(&sLine_ZJPBoN8j);
  numBytes += sLine_init(&sLine_dd6zbHbI);
  numBytes += sLine_init(&sLine_fnxLh9cA);
  numBytes += sLine_init(&sLine_IZ2Om1Jb);
  numBytes += sLine_init(&sLine_JINsEOAB);
  numBytes += sBiquad_init(&sBiquad_s_Jo96Moux);
  numBytes += sLine_init(&sLine_7rgd3VNh);
  numBytes += sLine_init(&sLine_ZixC1Pet);
  numBytes += sLine_init(&sLine_S4rC6zLn);
  numBytes += sLine_init(&sLine_gVCakBg5);
  numBytes += sLine_init(&sLine_WEkTRzNX);
  numBytes += sLine_init(&sLine_BYtOCRgy);
  numBytes += sBiquad_init(&sBiquad_s_Hi1jQCzP);
  numBytes += sLine_init(&sLine_wpzdnGce);
  numBytes += sLine_init(&sLine_bx236IpN);
  numBytes += sLine_init(&sLine_j2GoFRCw);
  numBytes += sLine_init(&sLine_NlY0VCZV);
  numBytes += sLine_init(&sLine_dkHqOCXr);
  numBytes += sBiquad_init(&sBiquad_s_Aw4Dg0sr);
  numBytes += sLine_init(&sLine_LCfTd9ID);
  numBytes += sLine_init(&sLine_4VaKHD1e);
  numBytes += sLine_init(&sLine_DciG475c);
  numBytes += sLine_init(&sLine_0WUCz3An);
  numBytes += sLine_init(&sLine_2er8Yv1c);
  numBytes += sLine_init(&sLine_G5R6lu62);
  numBytes += sBiquad_init(&sBiquad_s_jRhPNL2g);
  numBytes += sLine_init(&sLine_3F4GMRkF);
  numBytes += sLine_init(&sLine_rby24EH1);
  numBytes += sLine_init(&sLine_3k5g1IdH);
  numBytes += sLine_init(&sLine_tI24KERb);
  numBytes += sLine_init(&sLine_BVWA1YXY);
  numBytes += sLine_init(&sLine_MoU8urdf);
  numBytes += sLine_init(&sLine_QTfLgJMX);
  numBytes += sLine_init(&sLine_orqh9TpR);
  numBytes += sLine_init(&sLine_NWlJaD2v);
  numBytes += sLine_init(&sLine_UERuc2TT);
  numBytes += sLine_init(&sLine_qEVXWEBh);
  numBytes += sLine_init(&sLine_TxjPaoTU);
  numBytes += sLine_init(&sLine_vGHuA5zu);
  numBytes += sLine_init(&sLine_sQzKqn9s);
  numBytes += sLine_init(&sLine_CoxZr5Ze);
  numBytes += sEnv_init(&sEnv_7DcPoE4W, 512, 256);
  numBytes += sEnv_init(&sEnv_opuIzkiy, 512, 256);
  numBytes += sLine_init(&sLine_kx2wIf3T);
  numBytes += sLine_init(&sLine_YtDohQAi);
  numBytes += sLine_init(&sLine_65Wdeb4Z);
  numBytes += sLine_init(&sLine_qgzzHDCM);
  numBytes += sLine_init(&sLine_1RbEMkiL);
  numBytes += sLine_init(&sLine_dqMzI0vv);
  numBytes += sLine_init(&sLine_pp5JfdTE);
  numBytes += sLine_init(&sLine_YV3v2uOo);
  numBytes += sLine_init(&sLine_QOge56TJ);
  numBytes += sLine_init(&sLine_WyFfcG1z);
  numBytes += sLine_init(&sLine_ZPW00dSA);
  numBytes += sLine_init(&sLine_gBD9a38k);
  numBytes += sLine_init(&sLine_K9QPJHn2);
  numBytes += sLine_init(&sLine_UCPF3frM);
  numBytes += sLine_init(&sLine_f0HiPJD1);
  numBytes += sLine_init(&sLine_JpdM51Pu);
  numBytes += sLine_init(&sLine_gCHyjQC6);
  numBytes += sLine_init(&sLine_BizsZvdS);
  numBytes += sEnv_init(&sEnv_1TvLgphb, 512, 256);
  numBytes += sEnv_init(&sEnv_p5pbPN1Q, 512, 256);
  numBytes += sLine_init(&sLine_UgCbAvGG);
  numBytes += sLine_init(&sLine_8SoWHIbM);
  numBytes += sLine_init(&sLine_dhOv2gwo);
  numBytes += sLine_init(&sLine_TG9l5fhg);
  numBytes += sLine_init(&sLine_3SItWbzE);
  numBytes += sLine_init(&sLine_1ZQImshA);
  numBytes += sLine_init(&sLine_I7Tzfx8d);
  numBytes += sLine_init(&sLine_436i1fPs);
  numBytes += sLine_init(&sLine_yOW9VxDn);
  numBytes += sLine_init(&sLine_XvhwWoGj);
  numBytes += sLine_init(&sLine_TFhGkTXs);
  numBytes += sLine_init(&sLine_NoLSESq7);
  numBytes += sLine_init(&sLine_hT4P1smt);
  numBytes += sLine_init(&sLine_hn3MwQZ7);
  numBytes += sLine_init(&sLine_9zwD0dMe);
  numBytes += sLine_init(&sLine_CTmP5gSj);
  numBytes += sLine_init(&sLine_gsvLugpM);
  numBytes += sLine_init(&sLine_uuoTx4ER);
  numBytes += sLine_init(&sLine_50dw8X8z);
  numBytes += sLine_init(&sLine_6UeFb5mq);
  numBytes += sEnv_init(&sEnv_002yFlyr, 512, 256);
  numBytes += sEnv_init(&sEnv_EYQBnZ5o, 512, 256);
  numBytes += sLine_init(&sLine_QJkbFjMq);
  numBytes += sLine_init(&sLine_ei6ijquT);
  numBytes += sLine_init(&sLine_q3t8fdtk);
  numBytes += sLine_init(&sLine_LGpLTtuO);
  numBytes += sLine_init(&sLine_adrmUAO1);
  numBytes += sLine_init(&sLine_VSisnH6c);
  numBytes += sLine_init(&sLine_T4AXGh4t);
  numBytes += sLine_init(&sLine_Ed0waiOo);
  numBytes += sLine_init(&sLine_cVGc06ZX);
  numBytes += sLine_init(&sLine_PBMJsRgR);
  numBytes += sLine_init(&sLine_ogPoI4Gm);
  numBytes += sLine_init(&sLine_QDmTEdfD);
  numBytes += sLine_init(&sLine_IJBRAbAK);
  numBytes += sLine_init(&sLine_UXOMloGm);
  numBytes += sLine_init(&sLine_4ncutjso);
  numBytes += sLine_init(&sLine_8iFLSOGV);
  numBytes += sLine_init(&sLine_maSlUTpQ);
  numBytes += sLine_init(&sLine_X2BBOH47);
  numBytes += sEnv_init(&sEnv_CMuR8qRn, 512, 256);
  numBytes += sEnv_init(&sEnv_45tOAySj, 512, 256);
  numBytes += sLine_init(&sLine_t0DjmaiP);
  numBytes += sLine_init(&sLine_225hgjbz);
  numBytes += sLine_init(&sLine_6VexcZJF);
  numBytes += sLine_init(&sLine_zT4WjDAp);
  numBytes += sLine_init(&sLine_D3xFLs9K);
  numBytes += sLine_init(&sLine_n4y0vey0);
  numBytes += sLine_init(&sLine_jkHkClBU);
  numBytes += sLine_init(&sLine_rbElxj1N);
  numBytes += sLine_init(&sLine_YTWmwqf5);
  numBytes += sLine_init(&sLine_8b1XMKYt);
  numBytes += sLine_init(&sLine_gpmk9514);
  numBytes += sLine_init(&sLine_TkJqBeom);
  numBytes += sLine_init(&sLine_VnrnqskK);
  numBytes += sLine_init(&sLine_6MXdMybS);
  numBytes += sLine_init(&sLine_JgktvltE);
  numBytes += sLine_init(&sLine_c6oIfRZl);
  numBytes += sLine_init(&sLine_w0usi9ji);
  numBytes += sLine_init(&sLine_JSIk82TP);
  numBytes += sLine_init(&sLine_FqCHeqQA);
  numBytes += sLine_init(&sLine_3ZMyjMzM);
  numBytes += sEnv_init(&sEnv_UvKpHaWA, 512, 256);
  numBytes += sEnv_init(&sEnv_xQgzju4H, 512, 256);
  numBytes += sLine_init(&sLine_2dWBcy6Z);
  numBytes += sLine_init(&sLine_lLssYiPy);
  numBytes += sLine_init(&sLine_emHMr1W4);
  numBytes += sLine_init(&sLine_Xt4uhHeO);
  numBytes += sLine_init(&sLine_RLNFQdZ1);
  numBytes += sLine_init(&sLine_DSPEe7yf);
  numBytes += sLine_init(&sLine_pKRE2uPH);
  numBytes += sLine_init(&sLine_sj6MO7R2);
  numBytes += sLine_init(&sLine_YtNcxyvu);
  numBytes += sLine_init(&sLine_We3kXJi3);
  numBytes += sLine_init(&sLine_yw79kbKp);
  numBytes += sLine_init(&sLine_coxYZNAo);
  numBytes += sLine_init(&sLine_KAUdhPPU);
  numBytes += sLine_init(&sLine_aG95Bhw8);
  numBytes += sLine_init(&sLine_lOjst2XV);
  numBytes += sLine_init(&sLine_OPvamDGw);
  numBytes += sLine_init(&sLine_B2BiVPLR);
  numBytes += sLine_init(&sLine_y4bKytXG);
  numBytes += sEnv_init(&sEnv_8SZBwKcI, 512, 256);
  numBytes += sEnv_init(&sEnv_bxy3cjVB, 512, 256);
  numBytes += sLine_init(&sLine_AmjVNRMc);
  numBytes += sLine_init(&sLine_qNgfQORs);
  numBytes += sLine_init(&sLine_15uNlUt4);
  numBytes += sLine_init(&sLine_hbYCTr0Q);
  numBytes += sLine_init(&sLine_36L2nBRt);
  numBytes += sLine_init(&sLine_sbsm5hgo);
  numBytes += cVar_init_f(&cVar_Uudm7beN, 0.0f);
  numBytes += cVar_init_f(&cVar_WEbLxt2K, 0.0f);
  numBytes += cVar_init_f(&cVar_h8fgoD7X, 0.0f);
  numBytes += cVar_init_f(&cVar_dZUNub7K, 0.0f);
  numBytes += cVar_init_f(&cVar_B2SJsuMT, 0.0f);
  numBytes += cVar_init_f(&cVar_vksrWJt7, 1.0f);
  numBytes += cVar_init_f(&cVar_pF65netw, 0.0f);
  numBytes += cVar_init_f(&cVar_nFdg7mi8, 0.0f);
  numBytes += cVar_init_f(&cVar_BAig76qw, 0.0f);
  numBytes += cVar_init_f(&cVar_lNtHYdg0, 0.0f);
  numBytes += cSlice_init(&cSlice_IPFUKDbP, 1, -1);
  numBytes += cVar_init_f(&cVar_uU9rdlxg, 1337.0f);
  numBytes += cVar_init_f(&cVar_gAg6qjBM, 0.0f);
  numBytes += cIf_init(&cIf_X56DICyB, false);
  numBytes += cIf_init(&cIf_yLeN1oPI, false);
  numBytes += cBinop_init(&cBinop_hTaK6H9K, 0.0f); // __pow
  numBytes += cIf_init(&cIf_SvcoDBo5, false);
  numBytes += cBinop_init(&cBinop_lsxyR1MR, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_xtedFMWw, -15.0f); // __eq
  numBytes += cVar_init_f(&cVar_ISmXXI6S, 0.0f);
  numBytes += cIf_init(&cIf_C7TUshBf, false);
  numBytes += cIf_init(&cIf_5aEBXhFJ, false);
  numBytes += cBinop_init(&cBinop_rIvpTJ8k, 0.0f); // __pow
  numBytes += cIf_init(&cIf_wns5xxaK, false);
  numBytes += cBinop_init(&cBinop_HEozZ6nC, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_0v6ClKby, -15.0f); // __eq
  numBytes += cVar_init_f(&cVar_bXjj2nBB, 0.0f);
  numBytes += cIf_init(&cIf_TJXofeju, false);
  numBytes += cIf_init(&cIf_UwVBHk82, false);
  numBytes += cBinop_init(&cBinop_3TCvFMH7, 0.0f); // __pow
  numBytes += cIf_init(&cIf_oOJZhzKm, false);
  numBytes += cBinop_init(&cBinop_MErk5WkG, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_pbxZIMdS, -15.0f); // __eq
  numBytes += cIf_init(&cIf_0058jvgH, false);
  numBytes += cSlice_init(&cSlice_pRF4sV0R, 1, -1);
  numBytes += cVar_init_f(&cVar_xxyCg8kW, 0.0f);
  numBytes += cSlice_init(&cSlice_tA7sCMKP, 1, -1);
  numBytes += cVar_init_f(&cVar_2o3gsLoE, 0.0f);
  numBytes += cSlice_init(&cSlice_qDbdejEZ, 1, -1);
  numBytes += cVar_init_f(&cVar_FMyaqghN, 0.0f);
  numBytes += cSlice_init(&cSlice_igYS0Laz, 1, -1);
  numBytes += cVar_init_f(&cVar_IvPeriUE, 0.313f);
  numBytes += cBinop_init(&cBinop_DIt2TF3W, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_5JplgOXI, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_fVwWMyv9, 0.0f);
  numBytes += cVar_init_f(&cVar_uz9xgzOh, 1337.0f);
  numBytes += cVar_init_f(&cVar_alITRf8y, 0.3f);
  numBytes += cBinop_init(&cBinop_IpwSQqMb, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_qe16fU4M, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_3KcoqOX8, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_eYOnIcXC, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_TrHSnBRS, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_RHbWx4ks, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_ZGf4lvFy, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_QU34jZCu, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_LBMbN51Y, 0.0f);
  numBytes += cVar_init_f(&cVar_Ux8L40qC, 1337.0f);
  numBytes += cVar_init_f(&cVar_1SQqlECF, 0.313f);
  numBytes += cBinop_init(&cBinop_8E1V1uNq, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_ARLz0Qgv, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_D3Vhpdco, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_Knrcs4kQ, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_sHOHcMnl, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_xv6VVk5p, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_sjKYFhKY, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_a51maKhF, 0.0f);
  numBytes += cVar_init_f(&cVar_cl6eqZL4, 1037.0f);
  numBytes += cVar_init_f(&cVar_GNL5bhuc, 0.4f);
  numBytes += cBinop_init(&cBinop_1AcyTjjy, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_XYD9pYXV, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_dUTZaPVx, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_7N5TSUPw, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_YANWlh3H, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_YBZfyTRS, 0.0f); // __mul
  numBytes += cSlice_init(&cSlice_2eovY5Jx, 1, -1);
  numBytes += cVar_init_f(&cVar_g7qup9cy, 1337.0f);
  numBytes += cVar_init_f(&cVar_6zbdNNKu, 0.0f);
  numBytes += cIf_init(&cIf_A8pa85ir, false);
  numBytes += cIf_init(&cIf_LMW8uNah, false);
  numBytes += cBinop_init(&cBinop_lvVaTXpv, 0.0f); // __pow
  numBytes += cIf_init(&cIf_8zZ4xUKY, false);
  numBytes += cBinop_init(&cBinop_fNmBxXCI, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_ZXltfU7q, -15.0f); // __eq
  numBytes += cVar_init_f(&cVar_w1SnImWl, 0.0f);
  numBytes += cIf_init(&cIf_LQRoR72w, false);
  numBytes += cIf_init(&cIf_RQJoknLz, false);
  numBytes += cBinop_init(&cBinop_y3u75JIv, 0.0f); // __pow
  numBytes += cIf_init(&cIf_xHBugYMj, false);
  numBytes += cBinop_init(&cBinop_WucXNllF, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_umhq6qFE, -15.0f); // __eq
  numBytes += cVar_init_f(&cVar_VDjwZrPZ, 0.0f);
  numBytes += cIf_init(&cIf_5bGc1Dcd, false);
  numBytes += cIf_init(&cIf_qPMXU2FV, false);
  numBytes += cBinop_init(&cBinop_TgrnBniG, 0.0f); // __pow
  numBytes += cIf_init(&cIf_nkLx9320, false);
  numBytes += cBinop_init(&cBinop_qDv83sJf, -15.0f); // __gt
  numBytes += cBinop_init(&cBinop_1PH5ytvI, -15.0f); // __eq
  numBytes += cIf_init(&cIf_7fr7Qn5C, false);
  numBytes += cSlice_init(&cSlice_J9vLCqLb, 1, -1);
  numBytes += cVar_init_f(&cVar_kD4dVJ2R, 0.0f);
  numBytes += cSlice_init(&cSlice_CTAIy3oK, 1, -1);
  numBytes += cVar_init_f(&cVar_xdySbQS6, 0.0f);
  numBytes += cSlice_init(&cSlice_TmcgSdi5, 1, -1);
  numBytes += cVar_init_f(&cVar_nlMVnmMc, 0.0f);
  numBytes += cSlice_init(&cSlice_OISHlI0z, 1, -1);
  numBytes += cVar_init_f(&cVar_nK7oTbDD, 0.313f);
  numBytes += cBinop_init(&cBinop_W2Vq8JlY, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_chC5bbOG, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_Ab5OQAIz, 0.0f);
  numBytes += cVar_init_f(&cVar_gcmj1Fey, 1337.0f);
  numBytes += cVar_init_f(&cVar_wgdWg9QQ, 0.3f);
  numBytes += cBinop_init(&cBinop_T9CfvvHV, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_YRCuchwE, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_TYPqN1Jk, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_3ZOv4dp6, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_6SExkTXH, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_ZeoDh5Nd, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_d6I1KWhm, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_iloRtr90, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_PoDQzqLM, 0.0f);
  numBytes += cVar_init_f(&cVar_Td91SR6z, 1337.0f);
  numBytes += cVar_init_f(&cVar_NAYNFJiB, 0.313f);
  numBytes += cBinop_init(&cBinop_CY1gO20Q, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_ssTyQGf0, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_Bcwpkm8M, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_Gm0WPrd4, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_ztg8Y2RH, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_bYVs0LCV, 44100.0f); // __div
  numBytes += cBinop_init(&cBinop_xbYRiY2A, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_qdMs3Pba, 0.0f);
  numBytes += cVar_init_f(&cVar_FjuiMfmI, 1037.0f);
  numBytes += cVar_init_f(&cVar_hrQgH3PA, 0.4f);
  numBytes += cBinop_init(&cBinop_jCxaUzve, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_0fxlH6ci, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_YmV68aP8, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_JPOhQzbY, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_WsZvETzu, 0.0f); // __mul
  numBytes += cBinop_init(&cBinop_g6EfD2Se, 0.0f); // __mul
  numBytes += cVar_init_f(&cVar_cDTlciWN, 5705.6f);
  numBytes += cVar_init_f(&cVar_xI7olurA, 313.3f);
  numBytes += cBinop_init(&cBinop_TVyITclt, 0.0f); // __div
  numBytes += cBinop_init(&cBinop_C0DHWCua, 0.0f); // __div
  numBytes += cBinop_init(&cBinop_IlYsLoX0, 0.0f); // __div
  numBytes += cSlice_init(&cSlice_OxSDdlgO, 1, -1);
  numBytes += cVar_init_f(&cVar_ocFZCKCX, 0.5f);
  numBytes += cVar_init_f(&cVar_M0mXCjwu, 0.0f);
  numBytes += cVar_init_f(&cVar_zQsN3mug, 0.0f);
  numBytes += cIf_init(&cIf_dcGLzOhc, false);
  numBytes += cIf_init(&cIf_pz7vYx32, false);
  numBytes += cDelay_init(this, &cDelay_Dd1djVr2, 10.0f);
  numBytes += cIf_init(&cIf_KTG9UN1r, false);
  numBytes += cIf_init(&cIf_uyvAbvrm, false);
  numBytes += cDelay_init(this, &cDelay_QAWOwtbO, 10.0f);
  numBytes += cIf_init(&cIf_VMrczBRp, false);
  numBytes += cIf_init(&cIf_wMV7ZqjE, false);
  numBytes += cDelay_init(this, &cDelay_1i5PSYUi, 10.0f);
  numBytes += cIf_init(&cIf_XFGbSp0l, false);
  numBytes += cIf_init(&cIf_YMfxl5VD, false);
  numBytes += cDelay_init(this, &cDelay_9AUmmWYN, 10.0f);
  numBytes += cIf_init(&cIf_cm407SPY, false);
  numBytes += cIf_init(&cIf_35BCguHV, false);
  numBytes += cDelay_init(this, &cDelay_NXAAAIsF, 10.0f);
  numBytes += cIf_init(&cIf_msWET4AO, false);
  numBytes += cIf_init(&cIf_pD2pLrCn, false);
  numBytes += cDelay_init(this, &cDelay_AAu5QIB9, 10.0f);
  numBytes += cIf_init(&cIf_7Ha14ZXT, false);
  numBytes += cIf_init(&cIf_9GO6Epdq, false);
  numBytes += cDelay_init(this, &cDelay_RedtM1XZ, 10.0f);
  numBytes += cIf_init(&cIf_UN5jR3iw, false);
  numBytes += cIf_init(&cIf_mwJgj6L0, false);
  numBytes += cDelay_init(this, &cDelay_LOktevFb, 10.0f);
  numBytes += cIf_init(&cIf_MqBXDyjt, false);
  numBytes += cIf_init(&cIf_EGmZTA6H, false);
  numBytes += cDelay_init(this, &cDelay_daTlKgQK, 10.0f);
  numBytes += cIf_init(&cIf_Qg3vZOmZ, false);
  numBytes += cIf_init(&cIf_WzR4oIST, false);
  numBytes += cDelay_init(this, &cDelay_Oz5OsZI8, 10.0f);
  numBytes += cIf_init(&cIf_0lypMwYR, false);
  numBytes += cIf_init(&cIf_UAMs6wuO, false);
  numBytes += cDelay_init(this, &cDelay_Oh732mEI, 10.0f);
  numBytes += cIf_init(&cIf_HlwQDSAa, false);
  numBytes += cIf_init(&cIf_3OOSkc0b, false);
  numBytes += cDelay_init(this, &cDelay_ST5gBwa1, 10.0f);
  numBytes += sVarf_init(&sVarf_jGYsXzV5, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_WxH6j9Pq, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_1XkO61QH, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_pMLCtIQp, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_tQAEmRSA, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_HdSGIVvE, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_peXENc61, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_gmMYFTyo, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_YU5vles9, 0.0f);
  numBytes += cSlice_init(&cSlice_YmHwPFj1, 1, -1);
  numBytes += cVar_init_f(&cVar_TF1GwCm9, 1.0f);
  numBytes += cSlice_init(&cSlice_OoH1Qp4T, 1, -1);
  numBytes += cVar_init_f(&cVar_iCMuYcKz, 0.0f);
  numBytes += cSlice_init(&cSlice_axsa2X0x, 1, -1);
  numBytes += cVar_init_f(&cVar_xJ0RrljN, 512.0f);
  numBytes += cSlice_init(&cSlice_3C0Gowsc, 1, -1);
  numBytes += cVar_init_f(&cVar_8FTilLlK, 1.0f);
  numBytes += cSlice_init(&cSlice_7yCldHiv, 1, -1);
  numBytes += cVar_init_f(&cVar_OUnPsAXz, 1.0f);
  numBytes += cDelay_init(this, &cDelay_arZ18Ie0, 20.0f);
  numBytes += cVar_init_f(&cVar_d9lPcQTB, 0.0f);
  numBytes += cIf_init(&cIf_iNJz9FPv, false);
  numBytes += cIf_init(&cIf_JYzx8HgW, false);
  numBytes += cBinop_init(&cBinop_YEeVHznb, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_cc4nAxNa, 1, -1);
  numBytes += cVar_init_f(&cVar_K1MN64FH, 0.0f);
  numBytes += sVarf_init(&sVarf_MtkhUlyb, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_V3ETqKtH, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_tuMg71Ok, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_8Jy9FumC, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_V2V8jseP, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_EFqy6zps, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_Od1XQm8K, 0.0f);
  numBytes += cVar_init_f(&cVar_VA6MSucp, 0.0f);
  numBytes += cIf_init(&cIf_iTxAl3iV, false);
  numBytes += cIf_init(&cIf_LjZYhgAo, false);
  numBytes += cDelay_init(this, &cDelay_d0fwqjZW, 10.0f);
  numBytes += cIf_init(&cIf_eyRD5zIA, false);
  numBytes += cIf_init(&cIf_fYvVQyP3, false);
  numBytes += cDelay_init(this, &cDelay_sVCi7rnF, 10.0f);
  numBytes += cIf_init(&cIf_lb40nRAH, false);
  numBytes += cIf_init(&cIf_xGCbN65e, false);
  numBytes += cDelay_init(this, &cDelay_oSEiY2QC, 10.0f);
  numBytes += cIf_init(&cIf_vuyGY4tQ, false);
  numBytes += cIf_init(&cIf_zguhKFjd, false);
  numBytes += cDelay_init(this, &cDelay_pURr00UJ, 10.0f);
  numBytes += cIf_init(&cIf_C0pt2DSn, false);
  numBytes += cIf_init(&cIf_k2IXdWKO, false);
  numBytes += cDelay_init(this, &cDelay_eMslrVLX, 10.0f);
  numBytes += cIf_init(&cIf_MkmJb04z, false);
  numBytes += cIf_init(&cIf_zHNoxGm9, false);
  numBytes += cDelay_init(this, &cDelay_OAyLJnpj, 10.0f);
  numBytes += cIf_init(&cIf_Uaj76DAW, false);
  numBytes += cIf_init(&cIf_5EiyQegT, false);
  numBytes += cDelay_init(this, &cDelay_Lu1uYppo, 10.0f);
  numBytes += cIf_init(&cIf_yyNLMG3E, false);
  numBytes += cIf_init(&cIf_DnjKhNWM, false);
  numBytes += cDelay_init(this, &cDelay_0TglfJOl, 10.0f);
  numBytes += cIf_init(&cIf_d5YIoBZr, false);
  numBytes += cIf_init(&cIf_QPDQzXpP, false);
  numBytes += cDelay_init(this, &cDelay_3E28iQhu, 10.0f);
  numBytes += cIf_init(&cIf_GsbZmXYx, false);
  numBytes += cIf_init(&cIf_Mw4bkTZg, false);
  numBytes += cDelay_init(this, &cDelay_XEm3K8l9, 10.0f);
  numBytes += cIf_init(&cIf_g2fHodJV, false);
  numBytes += cIf_init(&cIf_oMUBFsK8, false);
  numBytes += cDelay_init(this, &cDelay_5w0pn1Uj, 10.0f);
  numBytes += cIf_init(&cIf_YUYQPCFB, false);
  numBytes += cIf_init(&cIf_lHiQBhaH, false);
  numBytes += cDelay_init(this, &cDelay_ArnGXUJo, 10.0f);
  numBytes += sVarf_init(&sVarf_kMHdIB36, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_a4KnZP56, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_taRtdGB2, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_YpoOV4SO, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Zmj32A7k, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_kx32E46k, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_ObU8hxTg, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_b0hNdWBf, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_CunFlHmT, 0.0f);
  numBytes += cSlice_init(&cSlice_WPX31R9R, 1, -1);
  numBytes += cVar_init_f(&cVar_35veTBZ2, 1.0f);
  numBytes += cSlice_init(&cSlice_5LJSIxSV, 1, -1);
  numBytes += cVar_init_f(&cVar_ZLxderQc, 0.0f);
  numBytes += cSlice_init(&cSlice_7jSXxCnZ, 1, -1);
  numBytes += cVar_init_f(&cVar_7gvbRUho, 512.0f);
  numBytes += cSlice_init(&cSlice_xb3PVWVd, 1, -1);
  numBytes += cVar_init_f(&cVar_U8f1GYSP, 1.0f);
  numBytes += cSlice_init(&cSlice_y1ZiOLig, 1, -1);
  numBytes += cVar_init_f(&cVar_2XTjAuWf, 1.0f);
  numBytes += cDelay_init(this, &cDelay_EPtySQS8, 20.0f);
  numBytes += cVar_init_f(&cVar_N9g6K5gT, 0.0f);
  numBytes += cIf_init(&cIf_2GPfXjW3, false);
  numBytes += cIf_init(&cIf_UMU9VdC7, false);
  numBytes += cBinop_init(&cBinop_5b0lFFpz, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_rtlMGM82, 1, -1);
  numBytes += cVar_init_f(&cVar_5HobttMZ, 0.0f);
  numBytes += sVarf_init(&sVarf_lSwq8mza, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_T0xgSH8z, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Y7F7L9m7, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_FOfAkieo, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_jomfJ2KG, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_LSwUi9Zg, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_JBIq6J7w, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_3x25e2jw, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_uT3DcEaj, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_n28bgRkl, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_3e1g4Oco, 0.0f);
  numBytes += cVar_init_f(&cVar_96UmmIHQ, 0.0f);
  numBytes += cVar_init_f(&cVar_XTEu5phC, 0.0f);
  numBytes += cVar_init_f(&cVar_KcCE9QCM, 0.0f);
  numBytes += cVar_init_f(&cVar_6pLq9WCy, 0.0f);
  numBytes += cVar_init_f(&cVar_iDWrxXdj, 1.0f);
  numBytes += cSlice_init(&cSlice_cnysjtO6, 1, -1);
  numBytes += cVar_init_f(&cVar_dpdFeYaw, 0.5f);
  numBytes += cVar_init_f(&cVar_IKg4xYBS, 0.0f);
  numBytes += cVar_init_f(&cVar_UzVH5hEP, 0.0f);
  numBytes += cIf_init(&cIf_L8wtyRkz, false);
  numBytes += cIf_init(&cIf_0UziSUln, false);
  numBytes += cDelay_init(this, &cDelay_Zz4G0H1s, 10.0f);
  numBytes += cIf_init(&cIf_MSJ2rhuD, false);
  numBytes += cIf_init(&cIf_IjxalCac, false);
  numBytes += cDelay_init(this, &cDelay_3CI2ijKb, 10.0f);
  numBytes += cIf_init(&cIf_A7xbyVIp, false);
  numBytes += cIf_init(&cIf_fgtl7ym7, false);
  numBytes += cDelay_init(this, &cDelay_afD7ypye, 10.0f);
  numBytes += cIf_init(&cIf_6bFljx5e, false);
  numBytes += cIf_init(&cIf_9L2IP2FC, false);
  numBytes += cDelay_init(this, &cDelay_udYkQLCD, 10.0f);
  numBytes += cIf_init(&cIf_5VbxKJwu, false);
  numBytes += cIf_init(&cIf_lAQTjRw3, false);
  numBytes += cDelay_init(this, &cDelay_CrjxS59O, 10.0f);
  numBytes += cIf_init(&cIf_vSYmOfBc, false);
  numBytes += cIf_init(&cIf_ve8qkMIr, false);
  numBytes += cDelay_init(this, &cDelay_f1rjqsra, 10.0f);
  numBytes += cIf_init(&cIf_zM00CmwM, false);
  numBytes += cIf_init(&cIf_d8rxEpyn, false);
  numBytes += cDelay_init(this, &cDelay_kIJpx1QL, 10.0f);
  numBytes += cIf_init(&cIf_m937nwGI, false);
  numBytes += cIf_init(&cIf_mPuzOd7I, false);
  numBytes += cDelay_init(this, &cDelay_4ZUuceVg, 10.0f);
  numBytes += cIf_init(&cIf_UblHqpPZ, false);
  numBytes += cIf_init(&cIf_zAu43UMQ, false);
  numBytes += cDelay_init(this, &cDelay_vISuwMCw, 10.0f);
  numBytes += cIf_init(&cIf_BW6Km1Rc, false);
  numBytes += cIf_init(&cIf_8sJEbFQd, false);
  numBytes += cDelay_init(this, &cDelay_7YSZMAR9, 10.0f);
  numBytes += cIf_init(&cIf_Fnrjvx0J, false);
  numBytes += cIf_init(&cIf_KBs2cKeO, false);
  numBytes += cDelay_init(this, &cDelay_co7FWI8X, 10.0f);
  numBytes += cIf_init(&cIf_Q86cQbEO, false);
  numBytes += cIf_init(&cIf_nozFIrBj, false);
  numBytes += cDelay_init(this, &cDelay_7iwVzLMc, 10.0f);
  numBytes += sVarf_init(&sVarf_wDDwfKGv, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_uejxXEBp, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_U9U3U44s, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_jgqydgcA, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_eW3QVMwf, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_4lIqFWRH, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Flh0Ug9o, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_aG2OgZqR, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_RrKG4xNY, 0.0f);
  numBytes += cSlice_init(&cSlice_ORV7sPxf, 1, -1);
  numBytes += cVar_init_f(&cVar_GcjTuTPW, 1.0f);
  numBytes += cSlice_init(&cSlice_ZQBi7ycl, 1, -1);
  numBytes += cVar_init_f(&cVar_TqBzNWYb, 0.0f);
  numBytes += cSlice_init(&cSlice_gGsz0OV6, 1, -1);
  numBytes += cVar_init_f(&cVar_EMHmhOt3, 512.0f);
  numBytes += cSlice_init(&cSlice_ZjcWOpCE, 1, -1);
  numBytes += cVar_init_f(&cVar_wnRiXiy4, 1.0f);
  numBytes += cSlice_init(&cSlice_OzKhaLG7, 1, -1);
  numBytes += cVar_init_f(&cVar_zIcGzlGa, 1.0f);
  numBytes += cDelay_init(this, &cDelay_ZMFtjtjR, 20.0f);
  numBytes += cVar_init_f(&cVar_3oZhXiWn, 0.0f);
  numBytes += cIf_init(&cIf_qj2CWrww, false);
  numBytes += cIf_init(&cIf_OTZsxrs1, false);
  numBytes += cBinop_init(&cBinop_vdf5PFgh, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_Xg6TJzfE, 1, -1);
  numBytes += cVar_init_f(&cVar_GAdyyTVt, 0.0f);
  numBytes += sVarf_init(&sVarf_4RrETmif, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_2Ndn9KQG, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_JmPsMI7s, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Vg5GgGs1, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_P1GSsvHB, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_7RHfUzV5, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_ZhevNCBF, 0.0f);
  numBytes += cVar_init_f(&cVar_HRbU7CeG, 0.0f);
  numBytes += cIf_init(&cIf_caZyFUBz, false);
  numBytes += cIf_init(&cIf_OFuOxfHS, false);
  numBytes += cDelay_init(this, &cDelay_EWBaq5zv, 10.0f);
  numBytes += cIf_init(&cIf_4CIBc9GO, false);
  numBytes += cIf_init(&cIf_BHcKLl7Z, false);
  numBytes += cDelay_init(this, &cDelay_UcdJVJ4g, 10.0f);
  numBytes += cIf_init(&cIf_nQICfSTM, false);
  numBytes += cIf_init(&cIf_6J2nIAA1, false);
  numBytes += cDelay_init(this, &cDelay_Ab2V712U, 10.0f);
  numBytes += cIf_init(&cIf_4KqHx45W, false);
  numBytes += cIf_init(&cIf_9gYmSYUC, false);
  numBytes += cDelay_init(this, &cDelay_0kDvDfOB, 10.0f);
  numBytes += cIf_init(&cIf_t1IWtrGQ, false);
  numBytes += cIf_init(&cIf_95YLLLcm, false);
  numBytes += cDelay_init(this, &cDelay_BqYhunFc, 10.0f);
  numBytes += cIf_init(&cIf_GXFf6yqN, false);
  numBytes += cIf_init(&cIf_WLOUqG7R, false);
  numBytes += cDelay_init(this, &cDelay_FkVbRG5i, 10.0f);
  numBytes += cIf_init(&cIf_6rLq8vL2, false);
  numBytes += cIf_init(&cIf_yiiFDu2y, false);
  numBytes += cDelay_init(this, &cDelay_btzddTrh, 10.0f);
  numBytes += cIf_init(&cIf_6FzwUjcc, false);
  numBytes += cIf_init(&cIf_XBugU6D8, false);
  numBytes += cDelay_init(this, &cDelay_nlYULWDE, 10.0f);
  numBytes += cIf_init(&cIf_Wnf71lMp, false);
  numBytes += cIf_init(&cIf_DNK6Qfnu, false);
  numBytes += cDelay_init(this, &cDelay_Mo0r9DGV, 10.0f);
  numBytes += cIf_init(&cIf_S9GtnnA1, false);
  numBytes += cIf_init(&cIf_dBf3MJOh, false);
  numBytes += cDelay_init(this, &cDelay_3GoOmsEk, 10.0f);
  numBytes += cIf_init(&cIf_sj1MqZ0s, false);
  numBytes += cIf_init(&cIf_7W2L7zt1, false);
  numBytes += cDelay_init(this, &cDelay_bSz74kbb, 10.0f);
  numBytes += cIf_init(&cIf_MSPSodx5, false);
  numBytes += cIf_init(&cIf_wnBD7SOE, false);
  numBytes += cDelay_init(this, &cDelay_N9fWxzNY, 10.0f);
  numBytes += sVarf_init(&sVarf_4xOJDS4F, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_EhlJ6qIm, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_ID0Faufj, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Fwr3BRtP, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_T5EXXI9b, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_YNpMpOkZ, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_4ByEU52T, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_X8KJBSnl, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_nPytwmId, 0.0f);
  numBytes += cSlice_init(&cSlice_2L1iEtek, 1, -1);
  numBytes += cVar_init_f(&cVar_tjwCGUUl, 1.0f);
  numBytes += cSlice_init(&cSlice_JAQZg1fS, 1, -1);
  numBytes += cVar_init_f(&cVar_OkxvPbCq, 0.0f);
  numBytes += cSlice_init(&cSlice_J2NH60K7, 1, -1);
  numBytes += cVar_init_f(&cVar_PP4tW2MC, 512.0f);
  numBytes += cSlice_init(&cSlice_9ffvyDTC, 1, -1);
  numBytes += cVar_init_f(&cVar_hhvHu7AA, 1.0f);
  numBytes += cSlice_init(&cSlice_wYaC3vOp, 1, -1);
  numBytes += cVar_init_f(&cVar_AehKCGix, 1.0f);
  numBytes += cDelay_init(this, &cDelay_eBvITYP9, 20.0f);
  numBytes += cVar_init_f(&cVar_ETd0PQUF, 0.0f);
  numBytes += cIf_init(&cIf_SQw7j4CQ, false);
  numBytes += cIf_init(&cIf_58L0i0bl, false);
  numBytes += cBinop_init(&cBinop_Yryw9a9t, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_8DYE7hEE, 1, -1);
  numBytes += cVar_init_f(&cVar_WEJNRm8c, 0.0f);
  numBytes += sVarf_init(&sVarf_8rWqOBQG, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Zv8fG32K, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_dnuNoNdK, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_PAZ7iB8d, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_za0OzJrY, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_cn3rsmrh, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_bk9mmres, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_pc28uIXQ, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_DNU44Dvk, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_48Gbr0d4, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_KI0DoguC, 0.0f);
  numBytes += cVar_init_f(&cVar_s7gCHDLK, 0.0f);
  numBytes += cVar_init_f(&cVar_atwbwh98, 0.0f);
  numBytes += cVar_init_f(&cVar_NLJ5dtoh, 0.0f);
  numBytes += cVar_init_f(&cVar_UxYXUtWZ, 0.0f);
  numBytes += cVar_init_f(&cVar_MEeaWoyX, 1.0f);
  numBytes += cSlice_init(&cSlice_FTxEpQYV, 1, -1);
  numBytes += cVar_init_f(&cVar_2IIWTMSO, 0.5f);
  numBytes += cVar_init_f(&cVar_V63JdNu0, 0.0f);
  numBytes += cVar_init_f(&cVar_U4imW5zu, 0.0f);
  numBytes += cIf_init(&cIf_hRuxZLH4, false);
  numBytes += cIf_init(&cIf_0u47Q93g, false);
  numBytes += cDelay_init(this, &cDelay_TKHxeOY0, 10.0f);
  numBytes += cIf_init(&cIf_OjfOlQA4, false);
  numBytes += cIf_init(&cIf_cmSooiFj, false);
  numBytes += cDelay_init(this, &cDelay_RYKcHfG2, 10.0f);
  numBytes += cIf_init(&cIf_cbvnHtRb, false);
  numBytes += cIf_init(&cIf_BO8YMi5m, false);
  numBytes += cDelay_init(this, &cDelay_VIsEsFSd, 10.0f);
  numBytes += cIf_init(&cIf_DxP0nUcX, false);
  numBytes += cIf_init(&cIf_THQTpFV7, false);
  numBytes += cDelay_init(this, &cDelay_45PVxpEO, 10.0f);
  numBytes += cIf_init(&cIf_epnifYji, false);
  numBytes += cIf_init(&cIf_WkaOda4Z, false);
  numBytes += cDelay_init(this, &cDelay_1dFP3pfI, 10.0f);
  numBytes += cIf_init(&cIf_GwLfWoaq, false);
  numBytes += cIf_init(&cIf_VQZG141H, false);
  numBytes += cDelay_init(this, &cDelay_PIlmjk51, 10.0f);
  numBytes += cIf_init(&cIf_2ctlVBqJ, false);
  numBytes += cIf_init(&cIf_Yp3U8Zld, false);
  numBytes += cDelay_init(this, &cDelay_aVjOEWaY, 10.0f);
  numBytes += cIf_init(&cIf_vI5gjHdj, false);
  numBytes += cIf_init(&cIf_eTwT6q0C, false);
  numBytes += cDelay_init(this, &cDelay_mxjt0dz0, 10.0f);
  numBytes += cIf_init(&cIf_cNSl2xfz, false);
  numBytes += cIf_init(&cIf_zZSWcAwy, false);
  numBytes += cDelay_init(this, &cDelay_g9nnTwsO, 10.0f);
  numBytes += cIf_init(&cIf_wMn8wSN9, false);
  numBytes += cIf_init(&cIf_risoWZno, false);
  numBytes += cDelay_init(this, &cDelay_hURFSPh0, 10.0f);
  numBytes += cIf_init(&cIf_ai5Ea2jd, false);
  numBytes += cIf_init(&cIf_ofd9W6lr, false);
  numBytes += cDelay_init(this, &cDelay_Ax5oo9Yw, 10.0f);
  numBytes += cIf_init(&cIf_8xTDJBHs, false);
  numBytes += cIf_init(&cIf_dOUAAhdB, false);
  numBytes += cDelay_init(this, &cDelay_kvWP1jXg, 10.0f);
  numBytes += sVarf_init(&sVarf_2Id1rHJp, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_WuCVYk7a, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_ueUGLXBc, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Gi6co6Sw, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_G1t2167a, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_fPkbEyM4, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_7Snmmtkh, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_VRWRTgoG, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_HRZqLKqd, 0.0f);
  numBytes += cSlice_init(&cSlice_0ZHN1N8B, 1, -1);
  numBytes += cVar_init_f(&cVar_itik8I1i, 1.0f);
  numBytes += cSlice_init(&cSlice_tw7jn2zk, 1, -1);
  numBytes += cVar_init_f(&cVar_zJ0DTtW8, 0.0f);
  numBytes += cSlice_init(&cSlice_zUVq5yTi, 1, -1);
  numBytes += cVar_init_f(&cVar_8qFyfhTt, 512.0f);
  numBytes += cSlice_init(&cSlice_8WPMLH0M, 1, -1);
  numBytes += cVar_init_f(&cVar_5WboOmCL, 1.0f);
  numBytes += cSlice_init(&cSlice_CYoZ61Af, 1, -1);
  numBytes += cVar_init_f(&cVar_CmgJsT2G, 1.0f);
  numBytes += cDelay_init(this, &cDelay_d81baJ7W, 20.0f);
  numBytes += cVar_init_f(&cVar_eQp8tbe3, 0.0f);
  numBytes += cIf_init(&cIf_VzvOtQkL, false);
  numBytes += cIf_init(&cIf_ypzjjpo4, false);
  numBytes += cBinop_init(&cBinop_1jln4NUD, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_qWV0nL4O, 1, -1);
  numBytes += cVar_init_f(&cVar_oY29yRWB, 0.0f);
  numBytes += sVarf_init(&sVarf_LiCkyoL0, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_z9KQRjps, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_VkqYtiKj, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_EBebXL78, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_oOVEUhU7, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_05FRLHZ2, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_u6HPV7b8, 0.0f);
  numBytes += cVar_init_f(&cVar_hJTkuq3x, 0.0f);
  numBytes += cIf_init(&cIf_CgkF4SEh, false);
  numBytes += cIf_init(&cIf_eLbVDaVI, false);
  numBytes += cDelay_init(this, &cDelay_huU0tQ7j, 10.0f);
  numBytes += cIf_init(&cIf_so9jvf99, false);
  numBytes += cIf_init(&cIf_AY6HjMZS, false);
  numBytes += cDelay_init(this, &cDelay_3u1NkgB6, 10.0f);
  numBytes += cIf_init(&cIf_9uEltQ6o, false);
  numBytes += cIf_init(&cIf_IhU68FEl, false);
  numBytes += cDelay_init(this, &cDelay_dU8Vp9gl, 10.0f);
  numBytes += cIf_init(&cIf_WVUMLUcL, false);
  numBytes += cIf_init(&cIf_Rd4sPitH, false);
  numBytes += cDelay_init(this, &cDelay_R5NjUSdF, 10.0f);
  numBytes += cIf_init(&cIf_DvTMFpFs, false);
  numBytes += cIf_init(&cIf_wydTbgUV, false);
  numBytes += cDelay_init(this, &cDelay_Et29iBH1, 10.0f);
  numBytes += cIf_init(&cIf_SSl9in7x, false);
  numBytes += cIf_init(&cIf_l03sluUu, false);
  numBytes += cDelay_init(this, &cDelay_G9mGlW6m, 10.0f);
  numBytes += cIf_init(&cIf_CsvwRLqz, false);
  numBytes += cIf_init(&cIf_qskZZuWi, false);
  numBytes += cDelay_init(this, &cDelay_05ou24Aq, 10.0f);
  numBytes += cIf_init(&cIf_TvH7a5lK, false);
  numBytes += cIf_init(&cIf_kuhc0Rct, false);
  numBytes += cDelay_init(this, &cDelay_gSyQyjWP, 10.0f);
  numBytes += cIf_init(&cIf_VVakyEXH, false);
  numBytes += cIf_init(&cIf_iTG4cmUt, false);
  numBytes += cDelay_init(this, &cDelay_8RC82jrB, 10.0f);
  numBytes += cIf_init(&cIf_hnW2RDF1, false);
  numBytes += cIf_init(&cIf_un7OCP5w, false);
  numBytes += cDelay_init(this, &cDelay_NrfF1O3K, 10.0f);
  numBytes += cIf_init(&cIf_ImiqxEp5, false);
  numBytes += cIf_init(&cIf_2RJnOJgw, false);
  numBytes += cDelay_init(this, &cDelay_LixpknZy, 10.0f);
  numBytes += cIf_init(&cIf_KIefjQYA, false);
  numBytes += cIf_init(&cIf_iCmgiWL2, false);
  numBytes += cDelay_init(this, &cDelay_XYZd6bwI, 10.0f);
  numBytes += sVarf_init(&sVarf_XAr338KV, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_vpdEas9q, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_xRsmaKFr, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_HEmOZanm, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_4H4yP1z5, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_kLtPkszU, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_Z0ksGvbs, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_xNccdKfL, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_tZFaMLqy, 0.0f);
  numBytes += cSlice_init(&cSlice_ncSalD50, 1, -1);
  numBytes += cVar_init_f(&cVar_4tAy6BVl, 1.0f);
  numBytes += cSlice_init(&cSlice_DiXe5zAq, 1, -1);
  numBytes += cVar_init_f(&cVar_FHXPm0ZB, 0.0f);
  numBytes += cSlice_init(&cSlice_HSPt5mGs, 1, -1);
  numBytes += cVar_init_f(&cVar_xweTUVZA, 512.0f);
  numBytes += cSlice_init(&cSlice_ZCrol453, 1, -1);
  numBytes += cVar_init_f(&cVar_Za0eD1qF, 1.0f);
  numBytes += cSlice_init(&cSlice_v3vavNxi, 1, -1);
  numBytes += cVar_init_f(&cVar_muKdJB5v, 1.0f);
  numBytes += cDelay_init(this, &cDelay_cKWHqMCx, 20.0f);
  numBytes += cVar_init_f(&cVar_gQd62AtM, 0.0f);
  numBytes += cIf_init(&cIf_rAn8mkqv, false);
  numBytes += cIf_init(&cIf_V1yXuthD, false);
  numBytes += cBinop_init(&cBinop_tC0L9hZm, 0.0f); // __sub
  numBytes += cSlice_init(&cSlice_9aADY5jx, 1, -1);
  numBytes += cVar_init_f(&cVar_Rcs5mJTj, 0.0f);
  numBytes += sVarf_init(&sVarf_7GfR5yr8, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_0LBjjQmS, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_saq1Q8WE, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_3MJfekF8, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_CH5wshEn, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_efaBIdJA, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_NaU2Rj6s, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_XKKLtN6L, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_mjGnpXDN, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_bxAz3npH, 0.0f, 0.0f, false);
  numBytes += cVar_init_f(&cVar_oLzkn6qr, 0.0f);
  numBytes += cVar_init_f(&cVar_n3jzoP0d, 0.0f);
  numBytes += cVar_init_f(&cVar_X6yszb3Y, 0.0f);
  numBytes += sVarf_init(&sVarf_41BxaYgu, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_pFxFrJAB, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_pqyGbl7v, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_v0TpMypi, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_HxKXweV4, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_qp6MUgTI, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_na221FsL, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_sIlCiHDe, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_HRDkZKL3, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_D9WIgVKX, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_1nmUGqu9, 0.0f, 0.0f, false);
  numBytes += sVarf_init(&sVarf_EbruvE0R, 0.0f, 0.0f, false);
  
  // schedule a message to trigger all loadbangs via the __hv_init receiver
  scheduleMessageForReceiver(0xCE5CC65B, msg_initWithBang(HV_MESSAGE_ON_STACK(1), 0));
}

Heavy_WSTD_M3NGLR::~Heavy_WSTD_M3NGLR() {
  sEnv_free(&sEnv_7DcPoE4W);
  sEnv_free(&sEnv_opuIzkiy);
  sEnv_free(&sEnv_1TvLgphb);
  sEnv_free(&sEnv_p5pbPN1Q);
  sEnv_free(&sEnv_002yFlyr);
  sEnv_free(&sEnv_EYQBnZ5o);
  sEnv_free(&sEnv_CMuR8qRn);
  sEnv_free(&sEnv_45tOAySj);
  sEnv_free(&sEnv_UvKpHaWA);
  sEnv_free(&sEnv_xQgzju4H);
  sEnv_free(&sEnv_8SZBwKcI);
  sEnv_free(&sEnv_bxy3cjVB);
}

HvTable *Heavy_WSTD_M3NGLR::getTableForHash(hv_uint32_t tableHash) {
  return nullptr;
}

void Heavy_WSTD_M3NGLR::scheduleMessageForReceiver(hv_uint32_t receiverHash, HvMessage *m) {
  switch (receiverHash) {
    case 0xC9CE7B11: { // High
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_Wj2JpUW3_sendMessage);
      break;
    }
    case 0xCFF7A234: { // High_Crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_MvFpJRdZ_sendMessage);
      break;
    }
    case 0x12EF45B9: { // High_Fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_waxN58Ni_sendMessage);
      break;
    }
    case 0xC05E9A15: { // High_Gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_5TBiuZl5_sendMessage);
      break;
    }
    case 0x8E596A67: { // High_Lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_yRvUW9ON_sendMessage);
      break;
    }
    case 0xAC5FA57F: { // High_Mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_RQ8lnqju_sendMessage);
      break;
    }
    case 0x585E4A54: { // High_Smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_LdKvmYkh_sendMessage);
      break;
    }
    case 0x14A9747F: { // High_Sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_lk6laZNw_sendMessage);
      break;
    }
    case 0x87744544: { // Low
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_h6wmJtSK_sendMessage);
      break;
    }
    case 0xEBDB0C5E: { // Low_Crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_6clWHYdo_sendMessage);
      break;
    }
    case 0xF974E926: { // Low_Fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_HlKHEeVN_sendMessage);
      break;
    }
    case 0xBB56D13C: { // Low_Gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_lzBJpNp2_sendMessage);
      break;
    }
    case 0xB4EFBAC1: { // Low_Lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_edLW5Ok2_sendMessage);
      break;
    }
    case 0x91FEF7A: { // Low_Mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_tozHwuH9_sendMessage);
      break;
    }
    case 0x1665AEE9: { // Low_Smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_8iCGnXsL_sendMessage);
      break;
    }
    case 0x860CA1D9: { // Low_Sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_rcFogwHW_sendMessage);
      break;
    }
    case 0xBFDC48B6: { // Mid
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_66Uw0S6A_sendMessage);
      break;
    }
    case 0xEA9DEF12: { // Mid_Crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_2wWaOjBZ_sendMessage);
      break;
    }
    case 0xDBF6FE0A: { // Mid_Fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_ERaNPiVQ_sendMessage);
      break;
    }
    case 0x59CF21AC: { // Mid_Freq
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_M7vmizJ4_sendMessage);
      break;
    }
    case 0x23F5E910: { // Mid_Gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_5NV2HJGm_sendMessage);
      break;
    }
    case 0x33C126B6: { // Mid_Lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_g4fXhjYP_sendMessage);
      break;
    }
    case 0x676FE235: { // Mid_Mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_rEDPzlnV_sendMessage);
      break;
    }
    case 0x64B1268F: { // Mid_Smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_HNfz30dK_sendMessage);
      break;
    }
    case 0xA91022F8: { // Mid_Sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_guIVuCz8_sendMessage);
      break;
    }
    case 0x1AFCE6EB: { // 1025-highvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_imIWhje3_sendMessage);
      break;
    }
    case 0xD8B5B1DB: { // 1025-lowvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_43ZddC2v_sendMessage);
      break;
    }
    case 0x3794F656: { // 1025-midfreq
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_tTQhvy6q_sendMessage);
      break;
    }
    case 0x94AE0ED9: { // 1025-midq
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_1QQNA9M3_sendMessage);
      break;
    }
    case 0xCC4DB41E: { // 1025-midvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_p64sWTJL_sendMessage);
      break;
    }
    case 0xAFE114F6: { // 1084-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_MDOMLgPe_sendMessage);
      break;
    }
    case 0x98647BAD: { // 1084-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_GdVfSCBI_sendMessage);
      break;
    }
    case 0xA4661944: { // 1084-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_H6MsTrQq_sendMessage);
      break;
    }
    case 0x505CED7B: { // 1125-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_l4rrY6rq_sendMessage);
      break;
    }
    case 0x82611A6: { // 1125-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_PdTy37pm_sendMessage);
      break;
    }
    case 0x1B990984: { // 1125-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_MP6FeVQM_sendMessage);
      break;
    }
    case 0x5F0AFAC5: { // 1166-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_IqW4d26x_sendMessage);
      break;
    }
    case 0x25349984: { // 1166-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_GOQ8YMqR_sendMessage);
      break;
    }
    case 0x8258EF5C: { // 1166-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_SQamgB1D_sendMessage);
      break;
    }
    case 0xBC1DF870: { // 1181-highvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_RFdY0G3h_sendMessage);
      break;
    }
    case 0x9A5A2ED: { // 1181-lowvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_9SEoXHoU_sendMessage);
      break;
    }
    case 0x6714DCEB: { // 1181-midfreq
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_QeesCwR7_sendMessage);
      break;
    }
    case 0xE583CFFE: { // 1181-midq
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_vPtFdr5P_sendMessage);
      break;
    }
    case 0xA6D36AFA: { // 1181-midvol
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_BbgOv7yo_sendMessage);
      break;
    }
    case 0x2AD225: { // 1240-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_URR6d0NG_sendMessage);
      break;
    }
    case 0x7DD55B03: { // 1240-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_ReTLutGG_sendMessage);
      break;
    }
    case 0xCF4D157B: { // 1240-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_UT34Tkd2_sendMessage);
      break;
    }
    case 0x4DF9D195: { // 1281-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_oJhZG77o_sendMessage);
      break;
    }
    case 0x58CB5D1B: { // 1281-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_OLCCp2mq_sendMessage);
      break;
    }
    case 0xE926ADAA: { // 1281-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_N9y7codF_sendMessage);
      break;
    }
    case 0x839669AF: { // 1322-alpha
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_dsYqG1bK_sendMessage);
      break;
    }
    case 0x89AA8991: { // 1322-wcos
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_Zx7qRj0a_sendMessage);
      break;
    }
    case 0xC43C6A92: { // 1322-wsin
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_scIqzx6W_sendMessage);
      break;
    }
    case 0x44C46A9A: { // 1345-mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_e2HxpxhL_sendMessage);
      break;
    }
    case 0x5B005CAA: { // 1352-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_J4se2wAL_sendMessage);
      break;
    }
    case 0xBBD23460: { // 1352-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_44TTkb8o_sendMessage);
      break;
    }
    case 0x23884B06: { // 1352-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_WnXNA3q8_sendMessage);
      break;
    }
    case 0x91BE5650: { // 1352-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_aIQ9p1Wr_sendMessage);
      break;
    }
    case 0x51E8B18D: { // 1352-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_4gMPFIf6_sendMessage);
      break;
    }
    case 0x2CAD1402: { // 1352-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_sRMDF2PX_sendMessage);
      break;
    }
    case 0x8CE949DD: { // 1470-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_YV9QQVZn_sendMessage);
      break;
    }
    case 0x1A98A8C5: { // 1470-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_J8JX4lev_sendMessage);
      break;
    }
    case 0x4211FC92: { // 1470-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_8GN31U0Z_sendMessage);
      break;
    }
    case 0x86F6710B: { // 1470-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_zXbmkPu5_sendMessage);
      break;
    }
    case 0x102FAA7E: { // 1470-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_S4Fliwp5_sendMessage);
      break;
    }
    case 0x1A325FAC: { // 1470-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_A4fKPXII_sendMessage);
      break;
    }
    case 0x5FF45DD7: { // 1602-mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_3CoH8hpN_sendMessage);
      break;
    }
    case 0xA63D49D8: { // 1609-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_DNvwUqp6_sendMessage);
      break;
    }
    case 0xFB1C4FF4: { // 1609-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_a9dr0eMq_sendMessage);
      break;
    }
    case 0x2D7D2212: { // 1609-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_DY3OxtQJ_sendMessage);
      break;
    }
    case 0xB4610071: { // 1609-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_x4AqsGDQ_sendMessage);
      break;
    }
    case 0x47BBE59E: { // 1609-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_PPp5Evmm_sendMessage);
      break;
    }
    case 0xD51403EB: { // 1609-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_tUPKHbIY_sendMessage);
      break;
    }
    case 0x559ECA90: { // 1727-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_Wkpv9Vsl_sendMessage);
      break;
    }
    case 0xB4F19FB9: { // 1727-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_tlvShJ1H_sendMessage);
      break;
    }
    case 0x7CE4DF88: { // 1727-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_QbQDH3nH_sendMessage);
      break;
    }
    case 0x9FD7A64D: { // 1727-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_Fejs2us2_sendMessage);
      break;
    }
    case 0x52C25AFE: { // 1727-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_v2skW5DT_sendMessage);
      break;
    }
    case 0x801C27A5: { // 1727-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_XGxhpSj1_sendMessage);
      break;
    }
    case 0xE5875B65: { // 1859-mix
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_3lKBO5ei_sendMessage);
      break;
    }
    case 0x40898D28: { // 1866-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_JRXNrRlD_sendMessage);
      break;
    }
    case 0x9D616407: { // 1866-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_fbjDC3zp_sendMessage);
      break;
    }
    case 0xE8DBDB0: { // 1866-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_L2MVKoIl_sendMessage);
      break;
    }
    case 0xD770731A: { // 1866-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_g7Xyl5Ts_sendMessage);
      break;
    }
    case 0x124CD18B: { // 1866-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_69voIQmF_sendMessage);
      break;
    }
    case 0x4017FCCD: { // 1866-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_aZYlD2gk_sendMessage);
      break;
    }
    case 0x1B1DD5FB: { // 1984-crshr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_AaWtmiYG_sendMessage);
      break;
    }
    case 0xA478FA4C: { // 1984-fldr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_nYfQhPMr_sendMessage);
      break;
    }
    case 0x97EB3671: { // 1984-gain
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_a9AmNNPu_sendMessage);
      break;
    }
    case 0xF3FECDEE: { // 1984-lmtr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_8wn4a4Mx_sendMessage);
      break;
    }
    case 0x9558C335: { // 1984-smthr
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_rnNNEjfA_sendMessage);
      break;
    }
    case 0x4131288C: { // 1984-sqnc
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_STQOLGXP_sendMessage);
      break;
    }
    case 0xCE5CC65B: { // __hv_init
      mq_addMessageByTimestamp(&mq, m, 0, &cReceive_SotfbAwO_sendMessage);
      break;
    }
    default: return;
  }
}

int Heavy_WSTD_M3NGLR::getParameterInfo(int index, HvParameterInfo *info) {
  if (info != nullptr) {
    switch (index) {
      case 0: {
        info->name = "High";
        info->hash = 0xC9CE7B11;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -15.0f;
        info->maxVal = 15.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 1: {
        info->name = "High_Crshr";
        info->hash = 0xCFF7A234;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 2.0f;
        info->maxVal = 512.0f;
        info->defaultVal = 512.0f;
        break;
      }
      case 2: {
        info->name = "High_Fldr";
        info->hash = 0x12EF45B9;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 3: {
        info->name = "High_Gain";
        info->hash = 0xC05E9A15;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -25.0f;
        info->maxVal = 0.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 4: {
        info->name = "High_Lmtr";
        info->hash = 0x8E596A67;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 1.0f;
        info->defaultVal = 1.0f;
        break;
      }
      case 5: {
        info->name = "High_Mix";
        info->hash = 0xAC5FA57F;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 100.0f;
        info->defaultVal = 50.0f;
        break;
      }
      case 6: {
        info->name = "High_Smthr";
        info->hash = 0x585E4A54;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 7: {
        info->name = "High_Sqnc";
        info->hash = 0x14A9747F;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 5.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 8: {
        info->name = "Low";
        info->hash = 0x87744544;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -15.0f;
        info->maxVal = 15.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 9: {
        info->name = "Low_Crshr";
        info->hash = 0xEBDB0C5E;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 2.0f;
        info->maxVal = 512.0f;
        info->defaultVal = 512.0f;
        break;
      }
      case 10: {
        info->name = "Low_Fldr";
        info->hash = 0xF974E926;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 11: {
        info->name = "Low_Gain";
        info->hash = 0xBB56D13C;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -25.0f;
        info->maxVal = 0.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 12: {
        info->name = "Low_Lmtr";
        info->hash = 0xB4EFBAC1;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 1.0f;
        info->defaultVal = 1.0f;
        break;
      }
      case 13: {
        info->name = "Low_Mix";
        info->hash = 0x91FEF7A;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 100.0f;
        info->defaultVal = 50.0f;
        break;
      }
      case 14: {
        info->name = "Low_Smthr";
        info->hash = 0x1665AEE9;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 15: {
        info->name = "Low_Sqnc";
        info->hash = 0x860CA1D9;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 5.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 16: {
        info->name = "Mid";
        info->hash = 0xBFDC48B6;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -15.0f;
        info->maxVal = 15.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 17: {
        info->name = "Mid_Crshr";
        info->hash = 0xEA9DEF12;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 2.0f;
        info->maxVal = 512.0f;
        info->defaultVal = 512.0f;
        break;
      }
      case 18: {
        info->name = "Mid_Fldr";
        info->hash = 0xDBF6FE0A;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 19: {
        info->name = "Mid_Freq";
        info->hash = 0x59CF21AC;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 313.3f;
        info->maxVal = 5705.6f;
        info->defaultVal = 1337.0f;
        break;
      }
      case 20: {
        info->name = "Mid_Gain";
        info->hash = 0x23F5E910;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = -25.0f;
        info->maxVal = 0.0f;
        info->defaultVal = 0.0f;
        break;
      }
      case 21: {
        info->name = "Mid_Lmtr";
        info->hash = 0x33C126B6;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 1.0f;
        info->defaultVal = 1.0f;
        break;
      }
      case 22: {
        info->name = "Mid_Mix";
        info->hash = 0x676FE235;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 100.0f;
        info->defaultVal = 50.0f;
        break;
      }
      case 23: {
        info->name = "Mid_Smthr";
        info->hash = 0x64B1268F;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 1.0f;
        info->maxVal = 13.37f;
        info->defaultVal = 1.0f;
        break;
      }
      case 24: {
        info->name = "Mid_Sqnc";
        info->hash = 0xA91022F8;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 5.0f;
        info->defaultVal = 0.0f;
        break;
      }
      default: {
        info->name = "invalid parameter index";
        info->hash = 0;
        info->type = HvParameterType::HV_PARAM_TYPE_PARAMETER_IN;
        info->minVal = 0.0f;
        info->maxVal = 0.0f;
        info->defaultVal = 0.0f;
        break;
      }
    }
  }
  return 25;
}



/*
 * Send Function Implementations
 */


void Heavy_WSTD_M3NGLR::cVar_Uudm7beN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_l8cRGrNQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_WEbLxt2K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_Bi4ShFsU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_h8fgoD7X_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_DtdQVDRt_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_dZUNub7K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_munyIyvq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_B2SJsuMT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uUTCtfBo_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_vksrWJt7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_jXleI2Um_sendMessage);
  cMsg_oqAqAIEo_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_jXleI2Um_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_vksrWJt7, 1, m, &cVar_vksrWJt7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_PZD91OeI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uYYB9uJL_sendMessage(_c, 0, m);
  cMsg_rI49sh6k_sendMessage(_c, 0, m);
  cMsg_CMSDTOOJ_sendMessage(_c, 0, m);
  cMsg_b2vyTGUK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_pF65netw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_k1YiH2TN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_nFdg7mi8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4o5i0Dh8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_BAig76qw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_eIbtwcrF_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_lNtHYdg0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_hkvz5J2l_sendMessage(_c, 0, m);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CcYPcft5_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_8Ob29ulP_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Mrjn5aCX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_kX50YYKG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xE9AAC520: { // "midfreq"
      cSlice_onMessage(_c, &Context(_c)->cSlice_IPFUKDbP, 0, m, &cSlice_IPFUKDbP_sendMessage);
      break;
    }
    default: {
      cSwitchcase_b309qL9h_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_IPFUKDbP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_9l0xoHNE_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_9l0xoHNE_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_uU9rdlxg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_9l0xoHNE_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_9l0xoHNE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_tTQhvy6q_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_gAg6qjBM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_bTeFaQS8_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_X56DICyB, 0, m, &cIf_X56DICyB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_X56DICyB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_nHHj8bhk_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_yLeN1oPI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_D7w5Qvj6_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_fYW25CCx_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_Z9hsIwRQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_hTaK6H9K, HV_BINOP_POW, 0, m, &cBinop_hTaK6H9K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_hTaK6H9K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_SvcoDBo5, 0, m, &cIf_SvcoDBo5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Q797ig8e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_IQBUlSIQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZknKVCeC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_oiO0ONBo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_hBKBqkrf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_yLeN1oPI, 0, m, &cIf_yLeN1oPI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_oiO0ONBo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_yLeN1oPI, 1, m, &cIf_yLeN1oPI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_fYW25CCx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_Q797ig8e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_D7w5Qvj6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_SvcoDBo5, 0, m, &cIf_SvcoDBo5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_IQBUlSIQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_hTaK6H9K, HV_BINOP_POW, 1, m, &cBinop_hTaK6H9K_sendMessage);
  cMsg_Z9hsIwRQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_SvcoDBo5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_SYWrgmJX_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_bTeFaQS8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_X56DICyB, 1, m, &cIf_X56DICyB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_nHHj8bhk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_9xjY193Z_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_sTLE5bvz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_BdoT5CJo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ZknKVCeC_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_hBKBqkrf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lsxyR1MR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_SvcoDBo5, 1, m, &cIf_SvcoDBo5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xtedFMWw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_baI36yQQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_baI36yQQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_SYWrgmJX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_8ckYSsjz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_BdoT5CJo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sTLE5bvz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_lsxyR1MR, HV_BINOP_GREATER_THAN, 1, m, &cBinop_lsxyR1MR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_9xjY193Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xtedFMWw, HV_BINOP_EQ, 1, m, &cBinop_xtedFMWw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zT2j1wAY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_gAg6qjBM, 0, m, &cVar_gAg6qjBM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Wd9udIyI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_lsxyR1MR, HV_BINOP_GREATER_THAN, 0, m, &cBinop_lsxyR1MR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_hNuQUVtn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xtedFMWw, HV_BINOP_EQ, 0, m, &cBinop_xtedFMWw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zT0kruKr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_8ckYSsjz_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_ISmXXI6S_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_W2L8GPdV_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_C7TUshBf, 0, m, &cIf_C7TUshBf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_C7TUshBf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_XvQwQjfm_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_5aEBXhFJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_Sgte01Lg_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_9kYw2XXV_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_wil7h67o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_rIvpTJ8k, HV_BINOP_POW, 0, m, &cBinop_rIvpTJ8k_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rIvpTJ8k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wns5xxaK, 0, m, &cIf_wns5xxaK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_VckztNK3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_d4yUvl1o_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Zu2IltQW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_mBxZOUMl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zCaBbbDr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5aEBXhFJ, 0, m, &cIf_5aEBXhFJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_mBxZOUMl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5aEBXhFJ, 1, m, &cIf_5aEBXhFJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_9kYw2XXV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_VckztNK3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Sgte01Lg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_wns5xxaK, 0, m, &cIf_wns5xxaK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_d4yUvl1o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_rIvpTJ8k, HV_BINOP_POW, 1, m, &cBinop_rIvpTJ8k_sendMessage);
  cMsg_wil7h67o_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_wns5xxaK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DwRrXqtE_sendMessage);
      cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0H9i61QM_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_W2L8GPdV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_C7TUshBf, 1, m, &cIf_C7TUshBf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XvQwQjfm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_EDzQXL1Q_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ZdU7N0T1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_RXdjBUot_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Zu2IltQW_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zCaBbbDr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_HEozZ6nC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wns5xxaK, 1, m, &cIf_wns5xxaK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_0v6ClKby_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uVczJ1bU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_uVczJ1bU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DwRrXqtE_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0H9i61QM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_groL4nUX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_RXdjBUot_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_EDzQXL1Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_0v6ClKby, HV_BINOP_EQ, 1, m, &cBinop_0v6ClKby_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZdU7N0T1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_HEozZ6nC, HV_BINOP_GREATER_THAN, 1, m, &cBinop_HEozZ6nC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bH7gdO1c_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_0v6ClKby, HV_BINOP_EQ, 0, m, &cBinop_0v6ClKby_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_4oZeZiLj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_groL4nUX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_sxexITpI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_ISmXXI6S, 0, m, &cVar_ISmXXI6S_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_XulKMGFa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_HEozZ6nC, HV_BINOP_GREATER_THAN, 0, m, &cBinop_HEozZ6nC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_bXjj2nBB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_JfHxYzjD_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_TJXofeju, 0, m, &cIf_TJXofeju_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_TJXofeju_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_mAcLKxWm_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_UwVBHk82_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_oiarpPk9_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_Xgs8uXM7_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_as7UEYlG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_3TCvFMH7, HV_BINOP_POW, 0, m, &cBinop_3TCvFMH7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_3TCvFMH7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_oOJZhzKm, 0, m, &cIf_oOJZhzKm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FKs1hnjT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_zsdbi4a2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_KqzMOQYH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UwVBHk82, 0, m, &cIf_UwVBHk82_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ImsWAvcQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_FYxF9jOo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FYxF9jOo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UwVBHk82, 1, m, &cIf_UwVBHk82_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Xgs8uXM7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_FKs1hnjT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oiarpPk9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_oOJZhzKm, 0, m, &cIf_oOJZhzKm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_zsdbi4a2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_3TCvFMH7, HV_BINOP_POW, 1, m, &cBinop_3TCvFMH7_sendMessage);
  cMsg_as7UEYlG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_oOJZhzKm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_PcHtUGzf_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_JfHxYzjD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_TJXofeju, 1, m, &cIf_TJXofeju_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_mAcLKxWm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_jzpDq6vI_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_sDyVQxFE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Qoc54MXu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ImsWAvcQ_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_KqzMOQYH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MErk5WkG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_oOJZhzKm, 1, m, &cIf_oOJZhzKm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_pbxZIMdS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_lUofWyWz_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_lUofWyWz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_PcHtUGzf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_xCiGi9r2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_Qoc54MXu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jzpDq6vI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_pbxZIMdS, HV_BINOP_EQ, 1, m, &cBinop_pbxZIMdS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sDyVQxFE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_MErk5WkG, HV_BINOP_GREATER_THAN, 1, m, &cBinop_MErk5WkG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_uHyprpIa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_pbxZIMdS, HV_BINOP_EQ, 0, m, &cBinop_pbxZIMdS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_A85DFdiu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_xCiGi9r2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_e97CnG4f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_bXjj2nBB, 0, m, &cVar_bXjj2nBB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xVBHH04v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_MErk5WkG, HV_BINOP_GREATER_THAN, 0, m, &cBinop_MErk5WkG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_0058jvgH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_1swmRkSQ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_q6dw9q4N_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x58220DA4: { // "highvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_pRF4sV0R, 0, m, &cSlice_pRF4sV0R_sendMessage);
      break;
    }
    default: {
      cSwitchcase_O05JfZ4r_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_pRF4sV0R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_c6wDpAdw_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_c6wDpAdw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_xxyCg8kW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_c6wDpAdw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_c6wDpAdw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_imIWhje3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_O05JfZ4r_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xBDABF734: { // "midvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_tA7sCMKP, 0, m, &cSlice_tA7sCMKP_sendMessage);
      break;
    }
    default: {
      cSwitchcase_kX50YYKG_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_tA7sCMKP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_yabBzQmT_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_yabBzQmT_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_2o3gsLoE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_yabBzQmT_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_yabBzQmT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_p64sWTJL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_b309qL9h_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x1FF74688: { // "lowvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_qDbdejEZ, 0, m, &cSlice_qDbdejEZ_sendMessage);
      break;
    }
    default: {
      cSwitchcase_b0JoSgtt_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_qDbdejEZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_D3ZVb5qf_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_D3ZVb5qf_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_FMyaqghN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_D3ZVb5qf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_D3ZVb5qf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_43ZddC2v_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_b0JoSgtt_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x98D37786: { // "midq"
      cSlice_onMessage(_c, &Context(_c)->cSlice_igYS0Laz, 0, m, &cSlice_igYS0Laz_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_igYS0Laz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_k678y7PJ_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_k678y7PJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_IvPeriUE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_k678y7PJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_k678y7PJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_1QQNA9M3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_pFczbUvG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_F0JuHtHo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_rG7ZnrFt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_wp2rbr12, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_TCQTYvdP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_Ik8joamg, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_qvCc9vtL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_DpAfafPl, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_xMG2nIp5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_xglAVFA6, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4AvLmp5R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_xiQfo7iN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_xiQfo7iN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_DIt2TF3W, HV_BINOP_DIVIDE, 1, m, &cBinop_DIt2TF3W_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_O7Yoaapa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_vK4FSmYj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_sCxXjkDV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_LfVqWrCA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_vK4FSmYj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_sCxXjkDV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_DIt2TF3W, HV_BINOP_DIVIDE, 0, m, &cBinop_DIt2TF3W_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_O7Yoaapa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_LfVqWrCA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_DIt2TF3W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_5JplgOXI, HV_BINOP_MULTIPLY, 1, m, &cBinop_5JplgOXI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_5JplgOXI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_XBWjegn8_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_xsNcw5bI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_PiJzfjkZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_lKfjUmHJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_Qff54zSd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_tpfj1tTO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_Rw3eMf3Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_63EFgYKA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_63EFgYKA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_5JplgOXI, HV_BINOP_MULTIPLY, 0, m, &cBinop_5JplgOXI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Alygj2rY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_Diqp7s0n_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Diqp7s0n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ydnEQr0Z_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_7wow9NCq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_fVwWMyv9, 1, m, &cVar_fVwWMyv9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_fVwWMyv9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_WboIsQZ8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_4Ugyg8k5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_qe16fU4M, HV_BINOP_MULTIPLY, 1, m, &cBinop_qe16fU4M_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_wRsRKMUr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_4Ugyg8k5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_hgklUeqD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_aUcyKWfs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_aUcyKWfs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_3KcoqOX8, HV_BINOP_MULTIPLY, 1, m, &cBinop_3KcoqOX8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_neejbN0l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_eYOnIcXC, HV_BINOP_MULTIPLY, 1, m, &cBinop_eYOnIcXC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_LgwrMfe7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_neejbN0l_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_S8OmFtjH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TrHSnBRS, HV_BINOP_MULTIPLY, 1, m, &cBinop_TrHSnBRS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_maF2aIb6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_yb3KIA3K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_yb3KIA3K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_RHbWx4ks, HV_BINOP_MULTIPLY, 1, m, &cBinop_RHbWx4ks_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_uz9xgzOh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_BKSplMDf_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_Rw3eMf3Z_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_RgaMirKa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_alITRf8y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_Alygj2rY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xsNcw5bI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_Qff54zSd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_XBWjegn8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_PiJzfjkZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_lKfjUmHJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_H6MsTrQq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_tpfj1tTO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_GdVfSCBI_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_lml1V5dX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_MDOMLgPe_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ydnEQr0Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_4ZLAtFlX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4ZLAtFlX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_IpwSQqMb, HV_BINOP_MULTIPLY, 1, m, &cBinop_IpwSQqMb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_IpwSQqMb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_GCrCWRtj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_GCrCWRtj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_lml1V5dX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_WboIsQZ8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_jtsC4Ex9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_jtsC4Ex9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6ghPkUEs_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wmu1eF9d_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_84ZOQoxX_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_K5VHT6S9_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DVrUWfGh_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_pkwfXCtS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_qe16fU4M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_pFczbUvG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_3KcoqOX8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_rG7ZnrFt_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_eYOnIcXC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_TCQTYvdP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_TrHSnBRS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_qvCc9vtL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_RHbWx4ks_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_xMG2nIp5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_1tQOwWXM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_uz9xgzOh, 0, m, &cVar_uz9xgzOh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_qc7Jwb9E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_alITRf8y, 1, m, &cVar_alITRf8y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_K5VHT6S9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_eYOnIcXC, HV_BINOP_MULTIPLY, 0, m, &cBinop_eYOnIcXC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6ghPkUEs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_pkwfXCtS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_qe16fU4M, HV_BINOP_MULTIPLY, 0, m, &cBinop_qe16fU4M_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_84ZOQoxX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TrHSnBRS, HV_BINOP_MULTIPLY, 0, m, &cBinop_TrHSnBRS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DVrUWfGh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_3KcoqOX8, HV_BINOP_MULTIPLY, 0, m, &cBinop_3KcoqOX8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wmu1eF9d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_RHbWx4ks, HV_BINOP_MULTIPLY, 0, m, &cBinop_RHbWx4ks_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_RgaMirKa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_fVwWMyv9, 0, m, &cVar_fVwWMyv9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_BKSplMDf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_alITRf8y, 0, m, &cVar_alITRf8y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_olD97KzV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_kimGO7Ur, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_WG4TTCHH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_nvirZ593, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_QfCovRDc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_xnRwmx5a, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_i0K3RKhA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_l6oALZOK, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_Xxi37HEJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_AMBiD38k_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_AMBiD38k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZGf4lvFy, HV_BINOP_DIVIDE, 1, m, &cBinop_ZGf4lvFy_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_51xqOBV7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_v7JYYGyE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_mdCNsxix_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_HAHUfJzB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_v7JYYGyE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_mdCNsxix_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZGf4lvFy, HV_BINOP_DIVIDE, 0, m, &cBinop_ZGf4lvFy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_51xqOBV7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_HAHUfJzB_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_ZGf4lvFy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_QU34jZCu, HV_BINOP_MULTIPLY, 1, m, &cBinop_QU34jZCu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_QU34jZCu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gOfVMP5U_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_NbWDjW0z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_39SH8RTN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_KEeTDlmG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_YzzIuHci_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_thmnFktR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_5ltHYuoO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_kxqybaME_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_kxqybaME_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_QU34jZCu, HV_BINOP_MULTIPLY, 0, m, &cBinop_QU34jZCu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ONd4t05u_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_vQJn8EcW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_vQJn8EcW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_F0QZJDGZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_RYseM1Ry_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_D3Vhpdco, HV_BINOP_MULTIPLY, 1, m, &cBinop_D3Vhpdco_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_UOYvWYah_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_LBMbN51Y, 1, m, &cVar_LBMbN51Y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_GoYp8bYa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Knrcs4kQ, HV_BINOP_MULTIPLY, 1, m, &cBinop_Knrcs4kQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EHKzmQNe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_WJZb4Tiz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WJZb4Tiz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_sHOHcMnl, HV_BINOP_MULTIPLY, 1, m, &cBinop_sHOHcMnl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_LBMbN51Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_FpihG7bL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_Ux8L40qC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_SbYg44Tm_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_5ltHYuoO_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_DL5leyXV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_1SQqlECF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_ONd4t05u_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gOfVMP5U_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_39SH8RTN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NbWDjW0z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_YzzIuHci_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_KEeTDlmG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_MP6FeVQM_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_thmnFktR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_PdTy37pm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_uVcntqKq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_l4rrY6rq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_F0QZJDGZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_6UXtLgmY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_6UXtLgmY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_8E1V1uNq, HV_BINOP_MULTIPLY, 1, m, &cBinop_8E1V1uNq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_8E1V1uNq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_LdlzJCwF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_LdlzJCwF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_uVcntqKq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_FpihG7bL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_4I6wCi7C_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4I6wCi7C_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_HDUaxF2i_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_RuhhcCIv_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Bfe1OrEU_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_heXT7pAm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ARLz0Qgv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_olD97KzV_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_D3Vhpdco_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_WG4TTCHH_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_Knrcs4kQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_QfCovRDc_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_sHOHcMnl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_i0K3RKhA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_ueyLGSqM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_1SQqlECF, 1, m, &cVar_1SQqlECF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6qBeUsyh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Ux8L40qC, 0, m, &cVar_Ux8L40qC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_SbYg44Tm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_1SQqlECF, 0, m, &cVar_1SQqlECF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DL5leyXV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_LBMbN51Y, 0, m, &cVar_LBMbN51Y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Bfe1OrEU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_D3Vhpdco, HV_BINOP_MULTIPLY, 0, m, &cBinop_D3Vhpdco_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_RuhhcCIv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Knrcs4kQ, HV_BINOP_MULTIPLY, 0, m, &cBinop_Knrcs4kQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_heXT7pAm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ARLz0Qgv, HV_BINOP_MULTIPLY, 0, m, &cBinop_ARLz0Qgv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HDUaxF2i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_sHOHcMnl, HV_BINOP_MULTIPLY, 0, m, &cBinop_sHOHcMnl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_gZ0T9MEa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ZJPBoN8j, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_2BugWlPZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_dd6zbHbI, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_sokL5Cxk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_fnxLh9cA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_JnRdJhTO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_IZ2Om1Jb, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_YE4ALcIP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JINsEOAB, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uka3hlWw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_EcOp74LL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_EcOp74LL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xv6VVk5p, HV_BINOP_DIVIDE, 1, m, &cBinop_xv6VVk5p_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CyOS6A5A_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_hV5Wct7G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_FM3CBJ1T_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_qX2yK3U3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_hV5Wct7G_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FM3CBJ1T_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xv6VVk5p, HV_BINOP_DIVIDE, 0, m, &cBinop_xv6VVk5p_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CyOS6A5A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_qX2yK3U3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_xv6VVk5p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_sjKYFhKY, HV_BINOP_MULTIPLY, 1, m, &cBinop_sjKYFhKY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_sjKYFhKY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_q7M2YZSU_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LgUAbKRJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_2fnnXhLp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_xWuhdySk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_zWjFSyZH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_qclWrAxq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_jt3joafz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_ZXxu9FGM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZXxu9FGM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_sjKYFhKY, HV_BINOP_MULTIPLY, 0, m, &cBinop_sjKYFhKY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_nLctOJOM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_SzRjYUho_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_SzRjYUho_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_v3GXNq7Y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_9At0rYQ6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_XYD9pYXV, HV_BINOP_MULTIPLY, 1, m, &cBinop_XYD9pYXV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_stl2HTHH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_FAqn96oV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FAqn96oV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_9At0rYQ6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_iBfdIBBv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_glpYvKWL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_glpYvKWL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_dUTZaPVx, HV_BINOP_MULTIPLY, 1, m, &cBinop_dUTZaPVx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_NX8VdMXM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_7N5TSUPw, HV_BINOP_MULTIPLY, 1, m, &cBinop_7N5TSUPw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_htRDScqH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_pRRlDkrC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_pRRlDkrC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_NX8VdMXM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Bc2yqogs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_a51maKhF, 1, m, &cVar_a51maKhF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FqAkigqz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YANWlh3H, HV_BINOP_MULTIPLY, 1, m, &cBinop_YANWlh3H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_onkF0dFu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_0RqstqLX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_0RqstqLX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YBZfyTRS, HV_BINOP_MULTIPLY, 1, m, &cBinop_YBZfyTRS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_a51maKhF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_vOBOVrKy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_cl6eqZL4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0aPPUxc1_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_jt3joafz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_aBQWOvCo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_GNL5bhuc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_nLctOJOM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_q7M2YZSU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_2fnnXhLp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LgUAbKRJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_zWjFSyZH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_xWuhdySk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_SQamgB1D_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_qclWrAxq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_GOQ8YMqR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_Hqv9Sgoa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_IqW4d26x_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_v3GXNq7Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_9UwUGXS3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_9UwUGXS3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1AcyTjjy, HV_BINOP_MULTIPLY, 1, m, &cBinop_1AcyTjjy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1AcyTjjy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_teRHDmwo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_teRHDmwo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_Hqv9Sgoa_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vOBOVrKy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_c2Ebt1Cm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_c2Ebt1Cm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CEhbxv7G_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LF4MCOii_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_m041FXWe_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_GTN85OaT_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_JLp0vAEl_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_dEiBUHHx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XYD9pYXV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_gZ0T9MEa_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_dUTZaPVx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_2BugWlPZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_7N5TSUPw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_sokL5Cxk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_YANWlh3H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_JnRdJhTO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_YBZfyTRS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_YE4ALcIP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_YCze8nUv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_cl6eqZL4, 0, m, &cVar_cl6eqZL4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_erdNPgSG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_GNL5bhuc, 1, m, &cVar_GNL5bhuc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CEhbxv7G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_dEiBUHHx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_XYD9pYXV, HV_BINOP_MULTIPLY, 0, m, &cBinop_XYD9pYXV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_m041FXWe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YANWlh3H, HV_BINOP_MULTIPLY, 0, m, &cBinop_YANWlh3H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_JLp0vAEl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_dUTZaPVx, HV_BINOP_MULTIPLY, 0, m, &cBinop_dUTZaPVx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LF4MCOii_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YBZfyTRS, HV_BINOP_MULTIPLY, 0, m, &cBinop_YBZfyTRS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_GTN85OaT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_7N5TSUPw, HV_BINOP_MULTIPLY, 0, m, &cBinop_7N5TSUPw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aBQWOvCo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_a51maKhF, 0, m, &cVar_a51maKhF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0aPPUxc1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_GNL5bhuc, 0, m, &cVar_GNL5bhuc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_EwAVRbOB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0058jvgH, 1, m, &cIf_0058jvgH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_OqCOmAsk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_1swmRkSQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_NBwehghK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_OqCOmAsk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_A4FmVmUM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0058jvgH, 0, m, &cIf_0058jvgH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SYWrgmJX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_lZjeicNW, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_1swmRkSQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_Rf2hEX6Q, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_PcHtUGzf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_7rgd3VNh, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_DwRrXqtE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_EwAVRbOB_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_NBwehghK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0H9i61QM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0313f, 0, m, &cBinop_A4FmVmUM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jM4W40KI_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xE9AAC520: { // "midfreq"
      cSlice_onMessage(_c, &Context(_c)->cSlice_2eovY5Jx, 0, m, &cSlice_2eovY5Jx_sendMessage);
      break;
    }
    default: {
      cSwitchcase_35J1tPF9_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_2eovY5Jx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_7vTjHtts_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_7vTjHtts_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_g7qup9cy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_7vTjHtts_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_7vTjHtts_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_QeesCwR7_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_6zbdNNKu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_2TDnAq2t_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_A8pa85ir, 0, m, &cIf_A8pa85ir_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_A8pa85ir_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_XmZqJA5J_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_LMW8uNah_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_Trd2ao6h_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_ZTMGWf7p_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_CeAjAxXv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_lvVaTXpv, HV_BINOP_POW, 0, m, &cBinop_lvVaTXpv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lvVaTXpv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_8zZ4xUKY, 0, m, &cIf_8zZ4xUKY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_J47LmBXx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_X1VR1HLA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vwuojnMA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_tHbHuUSP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_L2ACUihC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_LMW8uNah, 0, m, &cIf_LMW8uNah_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_tHbHuUSP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_LMW8uNah, 1, m, &cIf_LMW8uNah_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZTMGWf7p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_J47LmBXx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Trd2ao6h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_8zZ4xUKY, 0, m, &cIf_8zZ4xUKY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_X1VR1HLA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_lvVaTXpv, HV_BINOP_POW, 1, m, &cBinop_lvVaTXpv_sendMessage);
  cMsg_CeAjAxXv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_8zZ4xUKY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_I0LRTHja_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_2TDnAq2t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_A8pa85ir, 1, m, &cIf_A8pa85ir_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XmZqJA5J_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_b0jJXgnL_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Z6tJHpet_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_B4nNgEJL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vwuojnMA_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_L2ACUihC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_fNmBxXCI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_8zZ4xUKY, 1, m, &cIf_8zZ4xUKY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZXltfU7q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_8JPDh2h9_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_8JPDh2h9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_I0LRTHja_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Ns26xmyq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_B4nNgEJL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Z6tJHpet_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_fNmBxXCI, HV_BINOP_GREATER_THAN, 1, m, &cBinop_fNmBxXCI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_b0jJXgnL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZXltfU7q, HV_BINOP_EQ, 1, m, &cBinop_ZXltfU7q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2yKnTu0G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_Ns26xmyq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_ae0EBhG8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_fNmBxXCI, HV_BINOP_GREATER_THAN, 0, m, &cBinop_fNmBxXCI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6rkaZsed_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_6zbdNNKu, 0, m, &cVar_6zbdNNKu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wNQxy0Za_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZXltfU7q, HV_BINOP_EQ, 0, m, &cBinop_ZXltfU7q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_w1SnImWl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_hjB19TH2_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_LQRoR72w, 0, m, &cIf_LQRoR72w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_LQRoR72w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_MHm8VGem_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_RQJoknLz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_9pASetU3_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_JtVIGSCD_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_C3AU6Jhi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_y3u75JIv, HV_BINOP_POW, 0, m, &cBinop_y3u75JIv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_y3u75JIv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_xHBugYMj, 0, m, &cIf_xHBugYMj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_jZlnmIov_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_EHjQDg6r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Pfit9Cm2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_vHTAqqeU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_mOqRoEd1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_RQJoknLz, 0, m, &cIf_RQJoknLz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_vHTAqqeU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_RQJoknLz, 1, m, &cIf_RQJoknLz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_JtVIGSCD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_jZlnmIov_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_9pASetU3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_xHBugYMj, 0, m, &cIf_xHBugYMj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_EHjQDg6r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_y3u75JIv, HV_BINOP_POW, 1, m, &cBinop_y3u75JIv_sendMessage);
  cMsg_C3AU6Jhi_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_xHBugYMj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gQIOdeFJ_sendMessage);
      cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bhG88dQ3_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_hjB19TH2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_LQRoR72w, 1, m, &cIf_LQRoR72w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MHm8VGem_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_NGwveijH_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_BhxQBcAM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_5MXgmfcw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Pfit9Cm2_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_mOqRoEd1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WucXNllF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_xHBugYMj, 1, m, &cIf_xHBugYMj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_umhq6qFE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_Cf5eQhXz_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Cf5eQhXz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gQIOdeFJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bhG88dQ3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_pi6WbtKl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_5MXgmfcw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_BhxQBcAM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_WucXNllF, HV_BINOP_GREATER_THAN, 1, m, &cBinop_WucXNllF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NGwveijH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_umhq6qFE, HV_BINOP_EQ, 1, m, &cBinop_umhq6qFE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zucpqpUi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_w1SnImWl, 0, m, &cVar_w1SnImWl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Sala7CIy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_WucXNllF, HV_BINOP_GREATER_THAN, 0, m, &cBinop_WucXNllF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_q3ZRd5Ge_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_umhq6qFE, HV_BINOP_EQ, 0, m, &cBinop_umhq6qFE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_3B5hXH9F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_pi6WbtKl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_VDjwZrPZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_BwEV2Xus_sendMessage);
  cIf_onMessage(_c, &Context(_c)->cIf_5bGc1Dcd, 0, m, &cIf_5bGc1Dcd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_5bGc1Dcd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_QtTQk0dD_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_qPMXU2FV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cMsg_KK24adbA_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 485.0f, 0, m, &cBinop_maK0MOZ4_sendMessage);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cMsg_9XkLhH64_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 10.0f);
  cBinop_onMessage(_c, &Context(_c)->cBinop_TgrnBniG, HV_BINOP_POW, 0, m, &cBinop_TgrnBniG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_TgrnBniG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_nkLx9320, 0, m, &cIf_nkLx9320_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_HBjJBSMj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 20.0f, 0, m, &cBinop_wIuYQUY9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vXJwK8TH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_qPMXU2FV, 0, m, &cIf_qPMXU2FV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zzDdQBsA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_MvzV73kb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MvzV73kb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_qPMXU2FV, 1, m, &cIf_qPMXU2FV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_maK0MOZ4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 100.0f, 0, m, &cBinop_HBjJBSMj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KK24adbA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_nkLx9320, 0, m, &cIf_nkLx9320_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_wIuYQUY9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TgrnBniG, HV_BINOP_POW, 1, m, &cBinop_TgrnBniG_sendMessage);
  cMsg_9XkLhH64_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_nkLx9320_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_4cBnBp94_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cBinop_BwEV2Xus_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5bGc1Dcd, 1, m, &cIf_5bGc1Dcd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_QtTQk0dD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_izP4vnX1_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Jc2AfzSY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_kTx9bWUR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zzDdQBsA_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vXJwK8TH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_qDv83sJf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_nkLx9320, 1, m, &cIf_nkLx9320_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1PH5ytvI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_GddgUo6n_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_GddgUo6n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_4cBnBp94_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_g94BXK9y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_kTx9bWUR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Jc2AfzSY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_qDv83sJf, HV_BINOP_GREATER_THAN, 1, m, &cBinop_qDv83sJf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_izP4vnX1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1PH5ytvI, HV_BINOP_EQ, 1, m, &cBinop_1PH5ytvI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_e45dtl0N_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1PH5ytvI, HV_BINOP_EQ, 0, m, &cBinop_1PH5ytvI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gnhAuVBc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_g94BXK9y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_Jl9vc1mJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_VDjwZrPZ, 0, m, &cVar_VDjwZrPZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_7VtKQzLI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_qDv83sJf, HV_BINOP_GREATER_THAN, 0, m, &cBinop_qDv83sJf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_7fr7Qn5C_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_3dHg4aNQ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_etpstFjG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x58220DA4: { // "highvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_J9vLCqLb, 0, m, &cSlice_J9vLCqLb_sendMessage);
      break;
    }
    default: {
      cSwitchcase_ckBHH9NU_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_J9vLCqLb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_7fZ0G7VS_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_7fZ0G7VS_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_kD4dVJ2R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_7fZ0G7VS_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_7fZ0G7VS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_RFdY0G3h_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ckBHH9NU_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xBDABF734: { // "midvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_CTAIy3oK, 0, m, &cSlice_CTAIy3oK_sendMessage);
      break;
    }
    default: {
      cSwitchcase_jM4W40KI_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_CTAIy3oK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_WoiEOOzi_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_WoiEOOzi_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_xdySbQS6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_WoiEOOzi_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_WoiEOOzi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_BbgOv7yo_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_35J1tPF9_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x1FF74688: { // "lowvol"
      cSlice_onMessage(_c, &Context(_c)->cSlice_TmcgSdi5, 0, m, &cSlice_TmcgSdi5_sendMessage);
      break;
    }
    default: {
      cSwitchcase_H7pkSiUO_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_TmcgSdi5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_Q1dVFqrJ_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_Q1dVFqrJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_nlMVnmMc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_Q1dVFqrJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_Q1dVFqrJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_9SEoXHoU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_H7pkSiUO_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x98D37786: { // "midq"
      cSlice_onMessage(_c, &Context(_c)->cSlice_OISHlI0z, 0, m, &cSlice_OISHlI0z_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_OISHlI0z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_arfRUQTw_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_arfRUQTw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_nK7oTbDD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_arfRUQTw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_arfRUQTw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_vPtFdr5P_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_aPXH0UWd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ZixC1Pet, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_fM8muHLk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_S4rC6zLn, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_lIbVvUrL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gVCakBg5, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_hNEnT9oL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_WEkTRzNX, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_R9wGNvov_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_BYtOCRgy, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_TZXtMJ9Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_O7FGAfW1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_O7FGAfW1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_W2Vq8JlY, HV_BINOP_DIVIDE, 1, m, &cBinop_W2Vq8JlY_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_AjH7gRgA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_X2brpkoY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_nK89fYUP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_sVCgYjo2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_X2brpkoY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_nK89fYUP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_W2Vq8JlY, HV_BINOP_DIVIDE, 0, m, &cBinop_W2Vq8JlY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_AjH7gRgA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_sVCgYjo2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_W2Vq8JlY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_chC5bbOG, HV_BINOP_MULTIPLY, 1, m, &cBinop_chC5bbOG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_chC5bbOG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_z5bkfDqo_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_K2ZwDAao_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_Er2X2T0V_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_lZ2etS8D_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_NyC7HLZL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_mHmz9reX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_ATRyjES5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_ZtE4Zzxn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZtE4Zzxn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_chC5bbOG, HV_BINOP_MULTIPLY, 0, m, &cBinop_chC5bbOG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_0Q6F2s4z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_psWJEYLr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_psWJEYLr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_9lubIyab_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_KaNCp8Jk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Ab5OQAIz, 1, m, &cVar_Ab5OQAIz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_Ab5OQAIz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ltRzNvl3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_gYyfgJxv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YRCuchwE, HV_BINOP_MULTIPLY, 1, m, &cBinop_YRCuchwE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_C9yKus9m_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_gYyfgJxv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_JnHsxTiR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_S9sGiMmL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_S9sGiMmL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TYPqN1Jk, HV_BINOP_MULTIPLY, 1, m, &cBinop_TYPqN1Jk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_GKhaHnDn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_3ZOv4dp6, HV_BINOP_MULTIPLY, 1, m, &cBinop_3ZOv4dp6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_q5WTJwHo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_GKhaHnDn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4AMm70Da_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_6SExkTXH, HV_BINOP_MULTIPLY, 1, m, &cBinop_6SExkTXH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_X9jCmQhl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_pIwbdBHg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_pIwbdBHg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZeoDh5Nd, HV_BINOP_MULTIPLY, 1, m, &cBinop_ZeoDh5Nd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_gcmj1Fey_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ZiIWy17Q_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_ATRyjES5_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_t30J6559_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_wgdWg9QQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_0Q6F2s4z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_z5bkfDqo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_Er2X2T0V_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_K2ZwDAao_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_NyC7HLZL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_lZ2etS8D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_UT34Tkd2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_mHmz9reX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_ReTLutGG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_PtwOShgC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_URR6d0NG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9lubIyab_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_Wise8xfc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Wise8xfc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_T9CfvvHV, HV_BINOP_MULTIPLY, 1, m, &cBinop_T9CfvvHV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_T9CfvvHV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_XN9Ef90F_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XN9Ef90F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_PtwOShgC_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ltRzNvl3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_fS7bDVXx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_fS7bDVXx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CD7pz67H_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_aaTjzxyN_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_uZEgL0VP_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LocCgY9Y_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_2SIg55bL_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_EM1oN0gx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_YRCuchwE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_aPXH0UWd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_TYPqN1Jk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_fM8muHLk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_3ZOv4dp6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_lIbVvUrL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_6SExkTXH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_hNEnT9oL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_ZeoDh5Nd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_R9wGNvov_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_P8lb723y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_gcmj1Fey, 0, m, &cVar_gcmj1Fey_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ry0hqWLL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_wgdWg9QQ, 1, m, &cVar_wgdWg9QQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CD7pz67H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_EM1oN0gx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YRCuchwE, HV_BINOP_MULTIPLY, 0, m, &cBinop_YRCuchwE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_uZEgL0VP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_6SExkTXH, HV_BINOP_MULTIPLY, 0, m, &cBinop_6SExkTXH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2SIg55bL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TYPqN1Jk, HV_BINOP_MULTIPLY, 0, m, &cBinop_TYPqN1Jk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aaTjzxyN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ZeoDh5Nd, HV_BINOP_MULTIPLY, 0, m, &cBinop_ZeoDh5Nd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LocCgY9Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_3ZOv4dp6, HV_BINOP_MULTIPLY, 0, m, &cBinop_3ZOv4dp6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_t30J6559_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Ab5OQAIz, 0, m, &cVar_Ab5OQAIz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZiIWy17Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_wgdWg9QQ, 0, m, &cVar_wgdWg9QQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_44UxCuUK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_bx236IpN, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_ltjRhbHY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_j2GoFRCw, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_QIxZNfH9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_NlY0VCZV, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4J2m67pZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_dkHqOCXr, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_h2BYsSjj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_TsbLGK8A_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_TsbLGK8A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_d6I1KWhm, HV_BINOP_DIVIDE, 1, m, &cBinop_d6I1KWhm_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_KkihJ995_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_2x6vobT8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_ZKbrAn0r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_0bwtzSRU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_2x6vobT8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZKbrAn0r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_d6I1KWhm, HV_BINOP_DIVIDE, 0, m, &cBinop_d6I1KWhm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_KkihJ995_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_0bwtzSRU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_d6I1KWhm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_iloRtr90, HV_BINOP_MULTIPLY, 1, m, &cBinop_iloRtr90_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_iloRtr90_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Kwi4Kv5v_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_2EHmpSER_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_kN9mIl8f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_iuMbHyxf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_lwU7y8Ql_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_MY5tfoKX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_fVzksULH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_NfHfO6SX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_NfHfO6SX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_iloRtr90, HV_BINOP_MULTIPLY, 0, m, &cBinop_iloRtr90_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_vryHoFGV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_KsE4jLNB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_KsE4jLNB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_teO2jz7b_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_l092jTZK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Bcwpkm8M, HV_BINOP_MULTIPLY, 1, m, &cBinop_Bcwpkm8M_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AgoAhiOw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_PoDQzqLM, 1, m, &cVar_PoDQzqLM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_YZhNLIar_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Gm0WPrd4, HV_BINOP_MULTIPLY, 1, m, &cBinop_Gm0WPrd4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zZAObVjr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_jPvsXtp6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_jPvsXtp6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ztg8Y2RH, HV_BINOP_MULTIPLY, 1, m, &cBinop_ztg8Y2RH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_PoDQzqLM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_s6ByXUXJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_Td91SR6z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zdXx47j3_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_fVzksULH_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_31CA8ngH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_NAYNFJiB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_vryHoFGV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Kwi4Kv5v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_kN9mIl8f_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2EHmpSER_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_lwU7y8Ql_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_iuMbHyxf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_N9y7codF_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_MY5tfoKX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_OLCCp2mq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_t55ibX5H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_oJhZG77o_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_teO2jz7b_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_PZUohPz0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_PZUohPz0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_CY1gO20Q, HV_BINOP_MULTIPLY, 1, m, &cBinop_CY1gO20Q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_CY1gO20Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_vTpkGXRy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_vTpkGXRy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_t55ibX5H_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_s6ByXUXJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_bAHgE2Uw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_bAHgE2Uw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_C8GOsqQM_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_PVGTALQ6_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_jULDWvRZ_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_sXRhPOUq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ssTyQGf0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_44UxCuUK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_Bcwpkm8M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ltjRhbHY_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_Gm0WPrd4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_QIxZNfH9_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_ztg8Y2RH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4J2m67pZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_QtjZCZ64_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_NAYNFJiB, 1, m, &cVar_NAYNFJiB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aq66E8BJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Td91SR6z, 0, m, &cVar_Td91SR6z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_31CA8ngH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_PoDQzqLM, 0, m, &cVar_PoDQzqLM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zdXx47j3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_NAYNFJiB, 0, m, &cVar_NAYNFJiB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jULDWvRZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Bcwpkm8M, HV_BINOP_MULTIPLY, 0, m, &cBinop_Bcwpkm8M_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_PVGTALQ6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Gm0WPrd4, HV_BINOP_MULTIPLY, 0, m, &cBinop_Gm0WPrd4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sXRhPOUq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ssTyQGf0, HV_BINOP_MULTIPLY, 0, m, &cBinop_ssTyQGf0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_C8GOsqQM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ztg8Y2RH, HV_BINOP_MULTIPLY, 0, m, &cBinop_ztg8Y2RH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_HqkPmZRq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_4VaKHD1e, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_hFsuCw2V_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_DciG475c, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_2ruBeCWf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_0WUCz3An, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_tECHbf1n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_2er8Yv1c, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_V4RuJIwh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_G5R6lu62, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_slxOIbhy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "samplerate");
  cSystem_onMessage(_c, NULL, 0, m, &cSystem_IPz49zqq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSystem_IPz49zqq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_bYVs0LCV, HV_BINOP_DIVIDE, 1, m, &cBinop_bYVs0LCV_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_x0rJCLZE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_GXDgRxzh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 8.0f, 0, m, &cBinop_yEuOD5K5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ldTXAvYf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cUnop_onMessage(_c, HV_UNOP_ATAN, m, &cUnop_GXDgRxzh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_yEuOD5K5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_bYVs0LCV, HV_BINOP_DIVIDE, 0, m, &cBinop_bYVs0LCV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_x0rJCLZE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ldTXAvYf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_bYVs0LCV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xbYRiY2A, HV_BINOP_MULTIPLY, 1, m, &cBinop_xbYRiY2A_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xbYRiY2A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Z56tdqhN_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_GKlAwYPt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_ixe0sUZq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_cqWaNJ58_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cUnop_ZqtaXyW9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_sfqahQdQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_7RX3ubOm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 20.0f, 0, m, &cBinop_FoYaMkTP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FoYaMkTP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_xbYRiY2A, HV_BINOP_MULTIPLY, 0, m, &cBinop_xbYRiY2A_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MI2tV2It_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.1f, 0, m, &cBinop_wKqMsVWX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_wKqMsVWX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_KIrgQXRL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_SgLN79XS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_0fxlH6ci, HV_BINOP_MULTIPLY, 1, m, &cBinop_0fxlH6ci_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_nQ6k9lYQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_qCEQzVVO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_qCEQzVVO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_SgLN79XS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_bULR6DH9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_yGSZkWUd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_yGSZkWUd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YmV68aP8, HV_BINOP_MULTIPLY, 1, m, &cBinop_YmV68aP8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WwRyZufC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_JPOhQzbY, HV_BINOP_MULTIPLY, 1, m, &cBinop_JPOhQzbY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_E2p3pDhh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_Du8XRaeU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Du8XRaeU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_WwRyZufC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XWZCxUU1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_qdMs3Pba, 1, m, &cVar_qdMs3Pba_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_sKJcIsIm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_WsZvETzu, HV_BINOP_MULTIPLY, 1, m, &cBinop_WsZvETzu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3JTJ98Me_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_lmO7ViuN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lmO7ViuN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_g6EfD2Se, HV_BINOP_MULTIPLY, 1, m, &cBinop_g6EfD2Se_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_qdMs3Pba_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_Q3jJjhHR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_FjuiMfmI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_I5q59aew_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 20000.0f, 0, m, &cBinop_7RX3ubOm_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_pcqjUwf3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_hrQgH3PA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_MI2tV2It_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_GKlAwYPt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_COS, m, &cUnop_ZqtaXyW9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Z56tdqhN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_SIN, m, &cUnop_ixe0sUZq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cSend_cqWaNJ58_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_scIqzx6W_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_sfqahQdQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_Zx7qRj0a_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_sbY4NYrr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_dsYqG1bK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_KIrgQXRL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_TE58ic1E_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_TE58ic1E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_jCxaUzve, HV_BINOP_MULTIPLY, 1, m, &cBinop_jCxaUzve_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_jCxaUzve_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_XeoenBYe_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XeoenBYe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_sbY4NYrr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Q3jJjhHR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 0.0f, 0, m, &cBinop_viIrEOrO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_viIrEOrO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_s0X6LIku_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zb382HSe_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_YfFdIsM5_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vBmwR6aC_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vXTXo51g_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wdqTyHhn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_0fxlH6ci_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_HqkPmZRq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_YmV68aP8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_hFsuCw2V_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_JPOhQzbY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_2ruBeCWf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_WsZvETzu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_tECHbf1n_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_g6EfD2Se_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_V4RuJIwh_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_W8gqUEvp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_hrQgH3PA, 1, m, &cVar_hrQgH3PA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0r5SXWUW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_FjuiMfmI, 0, m, &cVar_FjuiMfmI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vBmwR6aC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_JPOhQzbY, HV_BINOP_MULTIPLY, 0, m, &cBinop_JPOhQzbY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_s0X6LIku_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_wdqTyHhn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_0fxlH6ci, HV_BINOP_MULTIPLY, 0, m, &cBinop_0fxlH6ci_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_YfFdIsM5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_WsZvETzu, HV_BINOP_MULTIPLY, 0, m, &cBinop_WsZvETzu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vXTXo51g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YmV68aP8, HV_BINOP_MULTIPLY, 0, m, &cBinop_YmV68aP8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zb382HSe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_g6EfD2Se, HV_BINOP_MULTIPLY, 0, m, &cBinop_g6EfD2Se_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_I5q59aew_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_hrQgH3PA, 0, m, &cVar_hrQgH3PA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_pcqjUwf3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_qdMs3Pba, 0, m, &cVar_qdMs3Pba_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xFjgn0qj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_7fr7Qn5C, 1, m, &cIf_7fr7Qn5C_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_cnR9FF5r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cMsg_3dHg4aNQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_0UeyQ3Da_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_cnR9FF5r_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_nhwIatK0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_7fr7Qn5C, 0, m, &cIf_7fr7Qn5C_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_I0LRTHja_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_wpzdnGce, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_3dHg4aNQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_LCfTd9ID, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4cBnBp94_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_3F4GMRkF, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_gQIOdeFJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_GREATER_THAN, 0.0f, 0, m, &cBinop_xFjgn0qj_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_0UeyQ3Da_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bhG88dQ3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0313f, 0, m, &cBinop_nhwIatK0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_cDTlciWN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_C0DHWCua, HV_BINOP_DIVIDE, 0, m, &cBinop_C0DHWCua_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_xI7olurA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TVyITclt, HV_BINOP_DIVIDE, 1, m, &cBinop_TVyITclt_sendMessage);
  cBinop_onMessage(_c, &Context(_c)->cBinop_C0DHWCua, HV_BINOP_DIVIDE, 1, m, &cBinop_C0DHWCua_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_didzKcA2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_IlYsLoX0, HV_BINOP_DIVIDE, 0, m, &cBinop_IlYsLoX0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_QeBWX3o4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_IlYsLoX0, HV_BINOP_DIVIDE, 1, m, &cBinop_IlYsLoX0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_8Ob29ulP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_cDTlciWN, 0, m, &cVar_cDTlciWN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CcYPcft5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_xI7olurA, 0, m, &cVar_xI7olurA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Mrjn5aCX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_TVyITclt, HV_BINOP_DIVIDE, 0, m, &cBinop_TVyITclt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_TVyITclt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_LOG, m, &cUnop_didzKcA2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_C0DHWCua_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_LOG, m, &cUnop_QeBWX3o4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_IlYsLoX0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 2.0f, 0, m, &cBinop_WasDA64k_sendMessage);
}

void Heavy_WSTD_M3NGLR::cUnop_jct2rnHj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.313f, 0, m, &cBinop_HrUrIYEl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_7lHcFn9l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.313f, 0, m, &cBinop_5YDgQZXE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WasDA64k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 1.0f, 0, m, &cBinop_7lHcFn9l_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_5YDgQZXE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cUnop_onMessage(_c, HV_UNOP_ABS, m, &cUnop_jct2rnHj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_HrUrIYEl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ShTlb9K0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_3Lkzwf8D_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x764F1C36: { // "mix"
      cSlice_onMessage(_c, &Context(_c)->cSlice_OxSDdlgO, 0, m, &cSlice_OxSDdlgO_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_OxSDdlgO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_d7l7Cuii_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_d7l7Cuii_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_ocFZCKCX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_d7l7Cuii_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_d7l7Cuii_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_e2HxpxhL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_5OJXsxOJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.0f, 0, m, &cBinop_ozE6f9hb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ozE6f9hb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 2.0f, 0, m, &cBinop_7jbtVnnR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_7jbtVnnR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 1.0f, 0, m, &cBinop_xwJbZECQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xwJbZECQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0VxVazhW_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_QvmJikH4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QvmJikH4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_oWFsbSFn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_0VxVazhW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_1nOUWfTH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1nOUWfTH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_XB8xMI0J_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oWFsbSFn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_weOdNZ3m_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_weOdNZ3m_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_hxGg6hjX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XB8xMI0J_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_vDEsbsGw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_hxGg6hjX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_oszq0i3Y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vDEsbsGw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_3k5g1IdH, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_oszq0i3Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_rby24EH1, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_M0mXCjwu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_N7aMYqHc_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_zQsN3mug_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_u8ZuJVWn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_1SuaqSbl_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_efVb0Wbu_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_1IhjiWqt_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_j9hGj9HH_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_tAEoVmb2_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UeqAu0U9_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_YUV2JSrk_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_efVb0Wbu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_mw5vSFdB_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_1IhjiWqt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_z4Qd28Op_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_j9hGj9HH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_rG1fVpNv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_tAEoVmb2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uquytZev_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_UeqAu0U9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_BacoiGoS_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_YUV2JSrk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_0actBPwp_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_dcGLzOhc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_WC57BcLq_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_pz7vYx32_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_N1qz8Jh1_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_LHHCsrQt_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_MRTH5gDr_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_MRTH5gDr_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Dd1djVr2, 1, m, &cDelay_Dd1djVr2_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_OJuPbC5X_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Dd1djVr2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Dd1djVr2, m);
  cIf_onMessage(_c, &Context(_c)->cIf_dcGLzOhc, 0, m, &cIf_dcGLzOhc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_MRTH5gDr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Dd1djVr2, 0, m, &cDelay_Dd1djVr2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_OJuPbC5X_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Dd1djVr2, 0, m, &cDelay_Dd1djVr2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lJPhDTNO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_3sxePTFx_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_V0XSlxbU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lpwwbrYq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ZH6IOE9A_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CVI5MbER_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZH6IOE9A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_dcGLzOhc, 1, m, &cIf_dcGLzOhc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CVI5MbER_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_LHHCsrQt_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_3sxePTFx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_pz7vYx32, 1, m, &cIf_pz7vYx32_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_V0XSlxbU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_pz7vYx32, 0, m, &cIf_pz7vYx32_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_N1qz8Jh1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QTfLgJMX, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_WC57BcLq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QTfLgJMX, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_KTG9UN1r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_CLoxfWc6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_uyvAbvrm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_J1vSBc8j_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_tvalwGkv_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_IoFnwoRg_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_IoFnwoRg_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_QAWOwtbO, 1, m, &cDelay_QAWOwtbO_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6UXbMyPi_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_QAWOwtbO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_QAWOwtbO, m);
  cIf_onMessage(_c, &Context(_c)->cIf_KTG9UN1r, 0, m, &cIf_KTG9UN1r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_IoFnwoRg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_QAWOwtbO, 0, m, &cDelay_QAWOwtbO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6UXbMyPi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_QAWOwtbO, 0, m, &cDelay_QAWOwtbO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_VBoMi3Nm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ruryJJqT_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sEOdZfvB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MoOEPZ4q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_dJLr27Gr_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_bcdHbJja_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_dJLr27Gr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_KTG9UN1r, 1, m, &cIf_KTG9UN1r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bcdHbJja_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_tvalwGkv_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_ruryJJqT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_uyvAbvrm, 1, m, &cIf_uyvAbvrm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sEOdZfvB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_uyvAbvrm, 0, m, &cIf_uyvAbvrm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_J1vSBc8j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UERuc2TT, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_CLoxfWc6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UERuc2TT, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_VMrczBRp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_fdZzldL6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_wMV7ZqjE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_89OxhpHG_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_9cxizug0_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_0PVihfTd_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_0PVihfTd_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_1i5PSYUi, 1, m, &cDelay_1i5PSYUi_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_jpjZihlx_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_1i5PSYUi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_1i5PSYUi, m);
  cIf_onMessage(_c, &Context(_c)->cIf_VMrczBRp, 0, m, &cIf_VMrczBRp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_0PVihfTd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_1i5PSYUi, 0, m, &cDelay_1i5PSYUi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jpjZihlx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_1i5PSYUi, 0, m, &cDelay_1i5PSYUi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AfJIOB4U_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_7yqCEzd3_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_MsCjmUYS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AHWms0Mt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DCuBs4or_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_exzyA415_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DCuBs4or_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_VMrczBRp, 1, m, &cIf_VMrczBRp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_exzyA415_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_9cxizug0_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_MsCjmUYS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wMV7ZqjE, 0, m, &cIf_wMV7ZqjE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_7yqCEzd3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wMV7ZqjE, 1, m, &cIf_wMV7ZqjE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_89OxhpHG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_vGHuA5zu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_fdZzldL6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_vGHuA5zu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_XFGbSp0l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_9d2SlTBk_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_YMfxl5VD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_kWKa42Hw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_qbQCoCwI_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_GWnDuYtD_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_GWnDuYtD_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_9AUmmWYN, 1, m, &cDelay_9AUmmWYN_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HyDDGwcU_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_9AUmmWYN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_9AUmmWYN, m);
  cIf_onMessage(_c, &Context(_c)->cIf_XFGbSp0l, 0, m, &cIf_XFGbSp0l_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_GWnDuYtD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_9AUmmWYN, 0, m, &cDelay_9AUmmWYN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HyDDGwcU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_9AUmmWYN, 0, m, &cDelay_9AUmmWYN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_KcpVt3xn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_uWcJN9qR_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_aCiUaBN2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_99YZiQ76_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_kBFCUQOJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_cuE3BX4w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_cuE3BX4w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_qbQCoCwI_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_kBFCUQOJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_XFGbSp0l, 1, m, &cIf_XFGbSp0l_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aCiUaBN2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_YMfxl5VD, 0, m, &cIf_YMfxl5VD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_uWcJN9qR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_YMfxl5VD, 1, m, &cIf_YMfxl5VD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_kWKa42Hw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_orqh9TpR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_9d2SlTBk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_orqh9TpR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_cm407SPY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_gvPdvYSN_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_35BCguHV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_vAJ3MNCY_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_8DyYqgSH_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_epRJGG2j_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_epRJGG2j_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_NXAAAIsF, 1, m, &cDelay_NXAAAIsF_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_oQkYID0y_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_NXAAAIsF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_NXAAAIsF, m);
  cIf_onMessage(_c, &Context(_c)->cIf_cm407SPY, 0, m, &cIf_cm407SPY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_epRJGG2j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_NXAAAIsF, 0, m, &cDelay_NXAAAIsF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_oQkYID0y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_NXAAAIsF, 0, m, &cDelay_NXAAAIsF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_latzZbz4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_E354pcbZ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_PiD7oB2J_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_RryZD2AZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_9aB91nt4_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NdfJBG0c_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NdfJBG0c_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_8DyYqgSH_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_9aB91nt4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_cm407SPY, 1, m, &cIf_cm407SPY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_PiD7oB2J_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_35BCguHV, 0, m, &cIf_35BCguHV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_E354pcbZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_35BCguHV, 1, m, &cIf_35BCguHV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_vAJ3MNCY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_qEVXWEBh, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_gvPdvYSN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_qEVXWEBh, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_msWET4AO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_uwqr0Hmm_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_pD2pLrCn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_7lQvqihJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_2UracFtN_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_1FSSgVRl_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_1FSSgVRl_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_AAu5QIB9, 1, m, &cDelay_AAu5QIB9_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_1WKywMXh_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_AAu5QIB9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_AAu5QIB9, m);
  cIf_onMessage(_c, &Context(_c)->cIf_msWET4AO, 0, m, &cIf_msWET4AO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1FSSgVRl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_AAu5QIB9, 0, m, &cDelay_AAu5QIB9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_1WKywMXh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_AAu5QIB9, 0, m, &cDelay_AAu5QIB9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rEBOlXZ1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DANdR1eG_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_8xGBrd5L_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_dAV7dK9F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_8vQ8SuO9_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Kr2yeeiB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_8vQ8SuO9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_msWET4AO, 1, m, &cIf_msWET4AO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Kr2yeeiB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_2UracFtN_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_8xGBrd5L_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_pD2pLrCn, 0, m, &cIf_pD2pLrCn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DANdR1eG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_pD2pLrCn, 1, m, &cIf_pD2pLrCn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_7lQvqihJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_sQzKqn9s, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uwqr0Hmm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_sQzKqn9s, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_7Ha14ZXT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_z7X9nroL_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_9GO6Epdq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_2EhTeelE_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Yn8Voy5g_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Lfqkg7qi_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Lfqkg7qi_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_RedtM1XZ, 1, m, &cDelay_RedtM1XZ_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_cVnkBmEB_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_RedtM1XZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_RedtM1XZ, m);
  cIf_onMessage(_c, &Context(_c)->cIf_7Ha14ZXT, 0, m, &cIf_7Ha14ZXT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Lfqkg7qi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_RedtM1XZ, 0, m, &cDelay_RedtM1XZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_cVnkBmEB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_RedtM1XZ, 0, m, &cDelay_RedtM1XZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_y6nuFJKO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wkFHhBIl_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_eEWyZRTz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_DG4KPtHH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_J7zW50po_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LeobdC0Q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_J7zW50po_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_7Ha14ZXT, 1, m, &cIf_7Ha14ZXT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LeobdC0Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_Yn8Voy5g_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_wkFHhBIl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9GO6Epdq, 1, m, &cIf_9GO6Epdq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_eEWyZRTz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9GO6Epdq, 0, m, &cIf_9GO6Epdq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_2EhTeelE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_NWlJaD2v, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_z7X9nroL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_NWlJaD2v, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_UN5jR3iw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_sT1XXBlE_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_mwJgj6L0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_cAo12kxB_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_c9jriirJ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_XKGJLUPc_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_XKGJLUPc_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_LOktevFb, 1, m, &cDelay_LOktevFb_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HqZA5Bps_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_LOktevFb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_LOktevFb, m);
  cIf_onMessage(_c, &Context(_c)->cIf_UN5jR3iw, 0, m, &cIf_UN5jR3iw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_XKGJLUPc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_LOktevFb, 0, m, &cDelay_LOktevFb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HqZA5Bps_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_LOktevFb, 0, m, &cDelay_LOktevFb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_B1nEsm7j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_z5A2cL8e_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Jyi3wZdJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_K9a7qc13_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vDWk3T5d_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_DTrNJ7dQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vDWk3T5d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UN5jR3iw, 1, m, &cIf_UN5jR3iw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DTrNJ7dQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_c9jriirJ_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_z5A2cL8e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_mwJgj6L0, 1, m, &cIf_mwJgj6L0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Jyi3wZdJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_mwJgj6L0, 0, m, &cIf_mwJgj6L0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_cAo12kxB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TxjPaoTU, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_sT1XXBlE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TxjPaoTU, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_MqBXDyjt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_lH0LYCMg_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_EGmZTA6H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_y7UMGbn6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_BGlBRkjB_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_luqmUp1R_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_luqmUp1R_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_daTlKgQK, 1, m, &cDelay_daTlKgQK_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FnNyD4gG_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_daTlKgQK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_daTlKgQK, m);
  cIf_onMessage(_c, &Context(_c)->cIf_MqBXDyjt, 0, m, &cIf_MqBXDyjt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_luqmUp1R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_daTlKgQK, 0, m, &cDelay_daTlKgQK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FnNyD4gG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_daTlKgQK, 0, m, &cDelay_daTlKgQK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rywqI1Eq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_xBIdSt7E_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_wqL8FgPH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4HG5GhGg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_yKDa2uxc_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_dPyZGcQQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_dPyZGcQQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_BGlBRkjB_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_yKDa2uxc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_MqBXDyjt, 1, m, &cIf_MqBXDyjt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xBIdSt7E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_EGmZTA6H, 1, m, &cIf_EGmZTA6H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wqL8FgPH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_EGmZTA6H, 0, m, &cIf_EGmZTA6H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_y7UMGbn6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_CoxZr5Ze, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_lH0LYCMg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_CoxZr5Ze, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_Qg3vZOmZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_TqrY6zC2_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_WzR4oIST_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_BLN3i5oi_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_yQa64f1o_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_OA9CZbDI_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_OA9CZbDI_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Oz5OsZI8, 1, m, &cDelay_Oz5OsZI8_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Kv3iQvcO_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Oz5OsZI8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Oz5OsZI8, m);
  cIf_onMessage(_c, &Context(_c)->cIf_Qg3vZOmZ, 0, m, &cIf_Qg3vZOmZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_OA9CZbDI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Oz5OsZI8, 0, m, &cDelay_Oz5OsZI8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Kv3iQvcO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Oz5OsZI8, 0, m, &cDelay_Oz5OsZI8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rB3p3Qoy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_J1cEyKr0_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sRPxLWce_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_f3Elg7i8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_2tRH2uxk_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_iVcgveeF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_iVcgveeF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_yQa64f1o_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_2tRH2uxk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Qg3vZOmZ, 1, m, &cIf_Qg3vZOmZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sRPxLWce_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WzR4oIST, 0, m, &cIf_WzR4oIST_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_J1cEyKr0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WzR4oIST, 1, m, &cIf_WzR4oIST_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BLN3i5oi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_tI24KERb, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_TqrY6zC2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_tI24KERb, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_0lypMwYR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_4EH5XMuu_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_UAMs6wuO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_CCgkvemK_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ydshrnM3_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_EABWyVWO_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_EABWyVWO_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Oh732mEI, 1, m, &cDelay_Oh732mEI_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QYVFV3DM_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Oh732mEI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Oh732mEI, m);
  cIf_onMessage(_c, &Context(_c)->cIf_0lypMwYR, 0, m, &cIf_0lypMwYR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EABWyVWO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Oh732mEI, 0, m, &cDelay_Oh732mEI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QYVFV3DM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Oh732mEI, 0, m, &cDelay_Oh732mEI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_aHEMqln2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_k41dOuQK_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_615f0I1r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_R39NWHxv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_YyOzC60w_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NBpzhCU4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_YyOzC60w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0lypMwYR, 1, m, &cIf_0lypMwYR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NBpzhCU4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ydshrnM3_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_k41dOuQK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UAMs6wuO, 1, m, &cIf_UAMs6wuO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_615f0I1r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UAMs6wuO, 0, m, &cIf_UAMs6wuO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_CCgkvemK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_BVWA1YXY, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4EH5XMuu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_BVWA1YXY, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_HlwQDSAa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_CaAKZDhH_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_3OOSkc0b_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_WarVItu0_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_QJmsCgu6_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_quUYE4ny_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_quUYE4ny_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_ST5gBwa1, 1, m, &cDelay_ST5gBwa1_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FcLGZcFx_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_ST5gBwa1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_ST5gBwa1, m);
  cIf_onMessage(_c, &Context(_c)->cIf_HlwQDSAa, 0, m, &cIf_HlwQDSAa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_quUYE4ny_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_ST5gBwa1, 0, m, &cDelay_ST5gBwa1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FcLGZcFx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_ST5gBwa1, 0, m, &cDelay_ST5gBwa1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XrvAbw10_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_B3aYKELS_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_dTYfcuZt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_cJhffrLQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_SmPffCnJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_itcyzUva_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_SmPffCnJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_HlwQDSAa, 1, m, &cIf_HlwQDSAa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_itcyzUva_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_QJmsCgu6_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_B3aYKELS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_3OOSkc0b, 1, m, &cIf_3OOSkc0b_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_dTYfcuZt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_3OOSkc0b, 0, m, &cIf_3OOSkc0b_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_WarVItu0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_MoU8urdf, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_CaAKZDhH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_MoU8urdf, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_mw5vSFdB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_m1WbPhVf_sendMessage(_c, 0, m);
  cMsg_aPpuCA8W_sendMessage(_c, 0, m);
  cMsg_MqvoIhMD_sendMessage(_c, 0, m);
  cMsg_S1HkdkK1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_z4Qd28Op_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_m1WbPhVf_sendMessage(_c, 0, m);
  cMsg_q1QaeP0S_sendMessage(_c, 0, m);
  cMsg_gy46A8vi_sendMessage(_c, 0, m);
  cMsg_x4sBnmmf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_rG1fVpNv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_vgqKkE0l_sendMessage(_c, 0, m);
  cMsg_q1QaeP0S_sendMessage(_c, 0, m);
  cMsg_S1HkdkK1_sendMessage(_c, 0, m);
  cMsg_UvWHz1O6_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_uquytZev_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_vgqKkE0l_sendMessage(_c, 0, m);
  cMsg_5FEwg2XG_sendMessage(_c, 0, m);
  cMsg_MqvoIhMD_sendMessage(_c, 0, m);
  cMsg_qQenThyC_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_BacoiGoS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_4eonUGI9_sendMessage(_c, 0, m);
  cMsg_x4sBnmmf_sendMessage(_c, 0, m);
  cMsg_Kvd7gyYT_sendMessage(_c, 0, m);
  cMsg_MqvoIhMD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_0actBPwp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_4eonUGI9_sendMessage(_c, 0, m);
  cMsg_qQenThyC_sendMessage(_c, 0, m);
  cMsg_q1QaeP0S_sendMessage(_c, 0, m);
  cMsg_zhtqarA3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_NFfBQlmo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lpwwbrYq_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_lJPhDTNO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fXd3pLy2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lpwwbrYq_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_lJPhDTNO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_41I3ELYf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MoOEPZ4q_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_VBoMi3Nm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_TkrSEjmu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MoOEPZ4q_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_VBoMi3Nm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BEi10pTd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_AHWms0Mt_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AfJIOB4U_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_CnH9U67j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_AHWms0Mt_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AfJIOB4U_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_m1WbPhVf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_NFfBQlmo_sendMessage(_c, 0, m);
  cMsg_TkrSEjmu_sendMessage(_c, 0, m);
  cMsg_CnH9U67j_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vgqKkE0l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_fXd3pLy2_sendMessage(_c, 0, m);
  cMsg_41I3ELYf_sendMessage(_c, 0, m);
  cMsg_CnH9U67j_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_4eonUGI9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_fXd3pLy2_sendMessage(_c, 0, m);
  cMsg_TkrSEjmu_sendMessage(_c, 0, m);
  cMsg_BEi10pTd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_EYIvAG7B_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_99YZiQ76_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_KcpVt3xn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_FiHHE29p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_99YZiQ76_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_KcpVt3xn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_0YkpI1lh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_RryZD2AZ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_latzZbz4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_hjVqYinM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_RryZD2AZ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_latzZbz4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_6wp5xpOW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_dAV7dK9F_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rEBOlXZ1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3OuZlinl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_dAV7dK9F_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rEBOlXZ1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_q1QaeP0S_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_FiHHE29p_sendMessage(_c, 0, m);
  cMsg_hjVqYinM_sendMessage(_c, 0, m);
  cMsg_3OuZlinl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_aPpuCA8W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_FiHHE29p_sendMessage(_c, 0, m);
  cMsg_0YkpI1lh_sendMessage(_c, 0, m);
  cMsg_6wp5xpOW_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_5FEwg2XG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_EYIvAG7B_sendMessage(_c, 0, m);
  cMsg_hjVqYinM_sendMessage(_c, 0, m);
  cMsg_6wp5xpOW_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_EDx6LDF4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_f3Elg7i8_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rB3p3Qoy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_w7sAZQxF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_f3Elg7i8_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rB3p3Qoy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_4x780Cpf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_R39NWHxv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aHEMqln2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_OyQzuFLq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_R39NWHxv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aHEMqln2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Wh04ikhv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_cJhffrLQ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XrvAbw10_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_FJJHsj74_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_cJhffrLQ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XrvAbw10_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_qQenThyC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_EDx6LDF4_sendMessage(_c, 0, m);
  cMsg_OyQzuFLq_sendMessage(_c, 0, m);
  cMsg_FJJHsj74_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_x4sBnmmf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_w7sAZQxF_sendMessage(_c, 0, m);
  cMsg_4x780Cpf_sendMessage(_c, 0, m);
  cMsg_FJJHsj74_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_S1HkdkK1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_w7sAZQxF_sendMessage(_c, 0, m);
  cMsg_OyQzuFLq_sendMessage(_c, 0, m);
  cMsg_Wh04ikhv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_gnOGgciU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_DG4KPtHH_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_y6nuFJKO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EMSXdOhx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_DG4KPtHH_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_y6nuFJKO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BHPmos6O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_K9a7qc13_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_B1nEsm7j_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ntlEzEEt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_K9a7qc13_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_B1nEsm7j_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_bhnzRv9p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_4HG5GhGg_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rywqI1Eq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_WtG5StKq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_4HG5GhGg_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rywqI1Eq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_MqvoIhMD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_EMSXdOhx_sendMessage(_c, 0, m);
  cMsg_ntlEzEEt_sendMessage(_c, 0, m);
  cMsg_WtG5StKq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_gy46A8vi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_EMSXdOhx_sendMessage(_c, 0, m);
  cMsg_BHPmos6O_sendMessage(_c, 0, m);
  cMsg_bhnzRv9p_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_zhtqarA3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_gnOGgciU_sendMessage(_c, 0, m);
  cMsg_BHPmos6O_sendMessage(_c, 0, m);
  cMsg_WtG5StKq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Kvd7gyYT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_EYIvAG7B_sendMessage(_c, 0, m);
  cMsg_0YkpI1lh_sendMessage(_c, 0, m);
  cMsg_3OuZlinl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_UvWHz1O6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_gnOGgciU_sendMessage(_c, 0, m);
  cMsg_ntlEzEEt_sendMessage(_c, 0, m);
  cMsg_bhnzRv9p_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_YU5vles9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_VngnN4kn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_QzlsTstK_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_YmHwPFj1, 0, m, &cSlice_YmHwPFj1_sendMessage);
      break;
    }
    default: {
      cSwitchcase_5fBw89m2_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_YmHwPFj1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_PaEaFOtw_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_PaEaFOtw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_TF1GwCm9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_PaEaFOtw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_PaEaFOtw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_44TTkb8o_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_YsGlCJbz_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_OoH1Qp4T, 0, m, &cSlice_OoH1Qp4T_sendMessage);
      break;
    }
    default: {
      cSwitchcase_l0zWFZbW_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_OoH1Qp4T_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_ocUSZKq8_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_ocUSZKq8_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_iCMuYcKz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_ocUSZKq8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_ocUSZKq8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_sRMDF2PX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_l0zWFZbW_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_axsa2X0x, 0, m, &cSlice_axsa2X0x_sendMessage);
      break;
    }
    default: {
      cSwitchcase_QzlsTstK_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_axsa2X0x_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_oIpGKtwd_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_oIpGKtwd_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_xJ0RrljN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_oIpGKtwd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_oIpGKtwd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_J4se2wAL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_5fBw89m2_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_3C0Gowsc, 0, m, &cSlice_3C0Gowsc_sendMessage);
      break;
    }
    default: {
      cSwitchcase_2p2rYlGj_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_3C0Gowsc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_Dcsco9t0_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_Dcsco9t0_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_8FTilLlK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_Dcsco9t0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_Dcsco9t0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_4gMPFIf6_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_2p2rYlGj_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_7yCldHiv, 0, m, &cSlice_7yCldHiv_sendMessage);
      break;
    }
    default: {
      cSwitchcase_mDYcciUh_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_7yCldHiv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_jrdYiDSD_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_jrdYiDSD_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_OUnPsAXz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_jrdYiDSD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_jrdYiDSD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_aIQ9p1Wr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_s2qfDPHE_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_ZZrfw5vR_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_ZZrfw5vR_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_arZ18Ie0, 1, m, &cDelay_arZ18Ie0_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_7hQwfsGl_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_arZ18Ie0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_arZ18Ie0, m);
  cMsg_3VyF2v66_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ZZrfw5vR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_arZ18Ie0, 0, m, &cDelay_arZ18Ie0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_7hQwfsGl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_arZ18Ie0, 0, m, &cDelay_arZ18Ie0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_d9lPcQTB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YEeVHznb, HV_BINOP_SUBTRACT, 0, m, &cBinop_YEeVHznb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_iNJz9FPv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_GN5oraCz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_JYzx8HgW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_GN5oraCz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_7DcPoE4W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_d9lPcQTB, 0, m, &cVar_d9lPcQTB_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_opuIzkiy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DTlRLabP_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_I33yQIPJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_XNXak3Tq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_YEeVHznb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_cEnSQpV0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_cEnSQpV0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_9FgmOsIC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_9FgmOsIC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_LFzdqCiF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_2vBJDoM5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_q3RYZbqH_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0M2rAxCK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0M2rAxCK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_rxIaKOXZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_q3RYZbqH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iNJz9FPv, 1, m, &cIf_iNJz9FPv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_LyxbLvGZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_JYzx8HgW, 1, m, &cIf_JYzx8HgW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rxIaKOXZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_iNJz9FPv, 0, m, &cIf_iNJz9FPv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_LFzdqCiF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_JYzx8HgW, 0, m, &cIf_JYzx8HgW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_I33yQIPJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_YEeVHznb, HV_BINOP_SUBTRACT, 1, m, &cBinop_YEeVHznb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DTlRLabP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_XNXak3Tq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_d9lPcQTB, 0, m, &cVar_d9lPcQTB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_GN5oraCz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_kx2wIf3T, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_mDYcciUh_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_cc4nAxNa, 0, m, &cSlice_cc4nAxNa_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_cc4nAxNa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_7iBoRKvy_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_7iBoRKvy_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_K1MN64FH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_7iBoRKvy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_7iBoRKvy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_WnXNA3q8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_N7aMYqHc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_dqMzI0vv, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_u8ZuJVWn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_qgzzHDCM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_wsDP8LGG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_zQsN3mug, 0, m, &cVar_zQsN3mug_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_VngnN4kn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_1RbEMkiL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_bnA3puum_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_s2qfDPHE_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_UixkLqjv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_RJUHRUcK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_3VyF2v66_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YtDohQAi, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_RJUHRUcK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YtDohQAi, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_PXypk2IJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_FRKMWljl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FRKMWljl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_StwAKLp5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_StwAKLp5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_65Wdeb4Z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_Od1XQm8K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_NjuL8Qmj_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_VA6MSucp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_v4GmXdBD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_7ibaKm4w_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NcSnLJTg_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_avCZ21yo_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ZpAX5z0s_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_GTTkwRpW_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_y6afyHVE_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Tz0Fhf23_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_NcSnLJTg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_dBefGIFL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_avCZ21yo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_1R9On0lp_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_ZpAX5z0s_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_SRRdJwWn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_GTTkwRpW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_n89d6y3q_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_y6afyHVE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_vIZuLZNn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_Tz0Fhf23_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_pKd3umsy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_iTxAl3iV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_xwHUGik8_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_LjZYhgAo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Idxx2Erp_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_f8HFDTnu_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_xuPfW2Ht_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_xuPfW2Ht_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_d0fwqjZW, 1, m, &cDelay_d0fwqjZW_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_z6B4OtPG_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_d0fwqjZW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_d0fwqjZW, m);
  cIf_onMessage(_c, &Context(_c)->cIf_iTxAl3iV, 0, m, &cIf_iTxAl3iV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_xuPfW2Ht_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_d0fwqjZW, 0, m, &cDelay_d0fwqjZW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_z6B4OtPG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_d0fwqjZW, 0, m, &cDelay_d0fwqjZW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_QR5S2sf7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_d66DtBkl_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FF7QjNU2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_zIqdcK4a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_OvHGXrUv_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_7exfctXU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_7exfctXU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_f8HFDTnu_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_OvHGXrUv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iTxAl3iV, 1, m, &cIf_iTxAl3iV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FF7QjNU2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_LjZYhgAo, 0, m, &cIf_LjZYhgAo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_d66DtBkl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_LjZYhgAo, 1, m, &cIf_LjZYhgAo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Idxx2Erp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_WyFfcG1z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_xwHUGik8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_WyFfcG1z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_eyRD5zIA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_paBw0608_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_fYvVQyP3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_sT8h3EnT_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_WgiD9V4K_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_GkJHZMCw_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_GkJHZMCw_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_sVCi7rnF, 1, m, &cDelay_sVCi7rnF_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ls8iOMxi_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_sVCi7rnF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_sVCi7rnF, m);
  cIf_onMessage(_c, &Context(_c)->cIf_eyRD5zIA, 0, m, &cIf_eyRD5zIA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_GkJHZMCw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_sVCi7rnF, 0, m, &cDelay_sVCi7rnF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ls8iOMxi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_sVCi7rnF, 0, m, &cDelay_sVCi7rnF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Xqahgyov_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_WTxt0Lmk_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HL2mVEEG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_2GotZKWq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_jZHOejWk_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_nILOtgbr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jZHOejWk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_eyRD5zIA, 1, m, &cIf_eyRD5zIA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_nILOtgbr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_WgiD9V4K_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_WTxt0Lmk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_fYvVQyP3, 1, m, &cIf_fYvVQyP3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HL2mVEEG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_fYvVQyP3, 0, m, &cIf_fYvVQyP3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_sT8h3EnT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_K9QPJHn2, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_paBw0608_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_K9QPJHn2, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_lb40nRAH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_oCfoqzxR_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_xGCbN65e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_wysgWiqv_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_u8D4MdSQ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_3WrYsXgT_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_3WrYsXgT_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_oSEiY2QC, 1, m, &cDelay_oSEiY2QC_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_k6st31A4_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_oSEiY2QC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_oSEiY2QC, m);
  cIf_onMessage(_c, &Context(_c)->cIf_lb40nRAH, 0, m, &cIf_lb40nRAH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3WrYsXgT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_oSEiY2QC, 0, m, &cDelay_oSEiY2QC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_k6st31A4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_oSEiY2QC, 0, m, &cDelay_oSEiY2QC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MhNeQI6k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Ww29ksHO_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_iHb5c7IH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xTXhFFHk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zOMKZMDp_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ZtjwgOkj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZtjwgOkj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_u8D4MdSQ_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_zOMKZMDp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_lb40nRAH, 1, m, &cIf_lb40nRAH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Ww29ksHO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_xGCbN65e, 1, m, &cIf_xGCbN65e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_iHb5c7IH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_xGCbN65e, 0, m, &cIf_xGCbN65e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wysgWiqv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JpdM51Pu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_oCfoqzxR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JpdM51Pu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_vuyGY4tQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_9oNbw5Go_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_zguhKFjd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_eBjyKT8C_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_MHvf9VKQ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_M7H8sbIx_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_M7H8sbIx_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_pURr00UJ, 1, m, &cDelay_pURr00UJ_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zr2ZJkI7_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_pURr00UJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_pURr00UJ, m);
  cIf_onMessage(_c, &Context(_c)->cIf_vuyGY4tQ, 0, m, &cIf_vuyGY4tQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_M7H8sbIx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_pURr00UJ, 0, m, &cDelay_pURr00UJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zr2ZJkI7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_pURr00UJ, 0, m, &cDelay_pURr00UJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_pKh3xg9R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_fiwLpA2j_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sL7LSsAr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rpoFnSBA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_UTFrk2EK_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_2I3JdUvs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2I3JdUvs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_MHvf9VKQ_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_UTFrk2EK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_vuyGY4tQ, 1, m, &cIf_vuyGY4tQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sL7LSsAr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zguhKFjd, 0, m, &cIf_zguhKFjd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_fiwLpA2j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zguhKFjd, 1, m, &cIf_zguhKFjd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_eBjyKT8C_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ZPW00dSA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_9oNbw5Go_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ZPW00dSA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_C0pt2DSn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Z7TVmbNX_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_k2IXdWKO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_pfIeGhrh_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_XKWAqPOQ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_dKsSdZp4_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_dKsSdZp4_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_eMslrVLX, 1, m, &cDelay_eMslrVLX_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Dcx2PHFV_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_eMslrVLX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_eMslrVLX, m);
  cIf_onMessage(_c, &Context(_c)->cIf_C0pt2DSn, 0, m, &cIf_C0pt2DSn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dKsSdZp4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_eMslrVLX, 0, m, &cDelay_eMslrVLX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Dcx2PHFV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_eMslrVLX, 0, m, &cDelay_eMslrVLX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_7VMfPGHF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gtakAzeL_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_SmlXjjB5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lgK2cNH7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_op937nA1_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QlOxvtEA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QlOxvtEA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_XKWAqPOQ_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_op937nA1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_C0pt2DSn, 1, m, &cIf_C0pt2DSn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_SmlXjjB5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_k2IXdWKO, 0, m, &cIf_k2IXdWKO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gtakAzeL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_k2IXdWKO, 1, m, &cIf_k2IXdWKO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_pfIeGhrh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UCPF3frM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_Z7TVmbNX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UCPF3frM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_MkmJb04z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_TQIf8f9k_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_zHNoxGm9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_IBs3sIUJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_iiwZYCo0_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_85EQzucY_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_85EQzucY_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_OAyLJnpj, 1, m, &cDelay_OAyLJnpj_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Leu5kAXP_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_OAyLJnpj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_OAyLJnpj, m);
  cIf_onMessage(_c, &Context(_c)->cIf_MkmJb04z, 0, m, &cIf_MkmJb04z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_85EQzucY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_OAyLJnpj, 0, m, &cDelay_OAyLJnpj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Leu5kAXP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_OAyLJnpj, 0, m, &cDelay_OAyLJnpj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_7bkIzOVu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Eqi9iPaJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_vDKHYr5G_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ayiDMcET_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_WdumuCa5_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NUJGnz8f_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NUJGnz8f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_iiwZYCo0_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_WdumuCa5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_MkmJb04z, 1, m, &cIf_MkmJb04z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vDKHYr5G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zHNoxGm9, 0, m, &cIf_zHNoxGm9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Eqi9iPaJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zHNoxGm9, 1, m, &cIf_zHNoxGm9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_IBs3sIUJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gCHyjQC6, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_TQIf8f9k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gCHyjQC6, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_Uaj76DAW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_8mDNWcFF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_5EiyQegT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_h8dzkDhn_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_x7BNlZp7_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_eKCzHXIQ_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_eKCzHXIQ_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Lu1uYppo, 1, m, &cDelay_Lu1uYppo_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_MxYG8YwW_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Lu1uYppo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Lu1uYppo, m);
  cIf_onMessage(_c, &Context(_c)->cIf_Uaj76DAW, 0, m, &cIf_Uaj76DAW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_eKCzHXIQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Lu1uYppo, 0, m, &cDelay_Lu1uYppo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_MxYG8YwW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Lu1uYppo, 0, m, &cDelay_Lu1uYppo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_5Ulrtg1E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_HFzgCKTj_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_X5U6ih1e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Mvb8e4QF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_XILD7Tqp_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QaTs21FM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QaTs21FM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_x7BNlZp7_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_XILD7Tqp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Uaj76DAW, 1, m, &cIf_Uaj76DAW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_X5U6ih1e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5EiyQegT, 0, m, &cIf_5EiyQegT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HFzgCKTj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5EiyQegT, 1, m, &cIf_5EiyQegT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_h8dzkDhn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gBD9a38k, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_8mDNWcFF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gBD9a38k, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_yyNLMG3E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_USguGsCc_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_DnjKhNWM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_oZjl2Dmw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ZshLYUAm_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_rVIN283W_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_rVIN283W_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_0TglfJOl, 1, m, &cDelay_0TglfJOl_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6s5crQC1_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_0TglfJOl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_0TglfJOl, m);
  cIf_onMessage(_c, &Context(_c)->cIf_yyNLMG3E, 0, m, &cIf_yyNLMG3E_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rVIN283W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_0TglfJOl, 0, m, &cDelay_0TglfJOl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6s5crQC1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_0TglfJOl, 0, m, &cDelay_0TglfJOl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_e7GZ1JIN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_WN6d5H9q_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sCyiiRXg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_JYg5mQwL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wOQuXQft_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_pptBOt7b_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wOQuXQft_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_yyNLMG3E, 1, m, &cIf_yyNLMG3E_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_pptBOt7b_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ZshLYUAm_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_sCyiiRXg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DnjKhNWM, 0, m, &cIf_DnjKhNWM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WN6d5H9q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DnjKhNWM, 1, m, &cIf_DnjKhNWM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oZjl2Dmw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_f0HiPJD1, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_USguGsCc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_f0HiPJD1, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_d5YIoBZr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_yGc89eo8_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_QPDQzXpP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_dQYCZcnU_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_X8mhkxg3_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_JR2oqKaR_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_JR2oqKaR_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_3E28iQhu, 1, m, &cDelay_3E28iQhu_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_craF5vqI_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_3E28iQhu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_3E28iQhu, m);
  cIf_onMessage(_c, &Context(_c)->cIf_d5YIoBZr, 0, m, &cIf_d5YIoBZr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_JR2oqKaR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_3E28iQhu, 0, m, &cDelay_3E28iQhu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_craF5vqI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_3E28iQhu, 0, m, &cDelay_3E28iQhu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AhqpiOId_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bVjVYcmc_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_H0oZCng7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_QKgnepKu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_AM2eKKm2_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ngJrtY0d_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ngJrtY0d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_X8mhkxg3_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_AM2eKKm2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_d5YIoBZr, 1, m, &cIf_d5YIoBZr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_H0oZCng7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_QPDQzXpP, 0, m, &cIf_QPDQzXpP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bVjVYcmc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_QPDQzXpP, 1, m, &cIf_QPDQzXpP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dQYCZcnU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_BizsZvdS, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_yGc89eo8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_BizsZvdS, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_GsbZmXYx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_kC5iSgmf_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_Mw4bkTZg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_uLG7ZRLl_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ju5pu9VD_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Kh8smg3R_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Kh8smg3R_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_XEm3K8l9, 1, m, &cDelay_XEm3K8l9_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_curIjVHO_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_XEm3K8l9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_XEm3K8l9, m);
  cIf_onMessage(_c, &Context(_c)->cIf_GsbZmXYx, 0, m, &cIf_GsbZmXYx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Kh8smg3R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_XEm3K8l9, 0, m, &cDelay_XEm3K8l9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_curIjVHO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_XEm3K8l9, 0, m, &cDelay_XEm3K8l9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ngivbmgu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_GkI4mUQn_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UUnzitbi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_2a27WnkY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_KRtU5Jom_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UkeXFmUM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_UkeXFmUM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ju5pu9VD_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_KRtU5Jom_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_GsbZmXYx, 1, m, &cIf_GsbZmXYx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_UUnzitbi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Mw4bkTZg, 0, m, &cIf_Mw4bkTZg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_GkI4mUQn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Mw4bkTZg, 1, m, &cIf_Mw4bkTZg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_uLG7ZRLl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_pp5JfdTE, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_kC5iSgmf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_pp5JfdTE, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_g2fHodJV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_rSK3FRWh_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_oMUBFsK8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_vZW5qy1L_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_izfiYfJx_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_7kMVm01Y_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_7kMVm01Y_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_5w0pn1Uj, 1, m, &cDelay_5w0pn1Uj_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ALMllp94_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_5w0pn1Uj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_5w0pn1Uj, m);
  cIf_onMessage(_c, &Context(_c)->cIf_g2fHodJV, 0, m, &cIf_g2fHodJV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_7kMVm01Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_5w0pn1Uj, 0, m, &cDelay_5w0pn1Uj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ALMllp94_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_5w0pn1Uj, 0, m, &cDelay_5w0pn1Uj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Zaf0x0Zv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_UAH6go0q_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_AOyzUksh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_kHXJOOWo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Ug5eqcdJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_IW4P17yI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_IW4P17yI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_izfiYfJx_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_Ug5eqcdJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_g2fHodJV, 1, m, &cIf_g2fHodJV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_AOyzUksh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_oMUBFsK8, 0, m, &cIf_oMUBFsK8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_UAH6go0q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_oMUBFsK8, 1, m, &cIf_oMUBFsK8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_vZW5qy1L_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YV3v2uOo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_rSK3FRWh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YV3v2uOo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_YUYQPCFB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_uMOT0S9u_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_lHiQBhaH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_wfqoIBO3_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_tsjOUzp6_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_EWZnwOhe_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_EWZnwOhe_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_ArnGXUJo, 1, m, &cDelay_ArnGXUJo_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_IaRti4Wd_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_ArnGXUJo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_ArnGXUJo, m);
  cIf_onMessage(_c, &Context(_c)->cIf_YUYQPCFB, 0, m, &cIf_YUYQPCFB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EWZnwOhe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_ArnGXUJo, 0, m, &cDelay_ArnGXUJo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_IaRti4Wd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_ArnGXUJo, 0, m, &cDelay_ArnGXUJo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_j6PsPTyO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_jnGaTgN6_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_U1yOJvyN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Q6yHOLVI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_8vqm2BYJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_izgOKkK6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_8vqm2BYJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_YUYQPCFB, 1, m, &cIf_YUYQPCFB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_izgOKkK6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_tsjOUzp6_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_U1yOJvyN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_lHiQBhaH, 0, m, &cIf_lHiQBhaH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jnGaTgN6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_lHiQBhaH, 1, m, &cIf_lHiQBhaH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wfqoIBO3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QOge56TJ, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uMOT0S9u_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QOge56TJ, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_dBefGIFL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_bbCRhc0d_sendMessage(_c, 0, m);
  cMsg_fiIEAkpj_sendMessage(_c, 0, m);
  cMsg_SR8WNT7g_sendMessage(_c, 0, m);
  cMsg_hDpspC6p_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_1R9On0lp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_bbCRhc0d_sendMessage(_c, 0, m);
  cMsg_KgB7QJ0R_sendMessage(_c, 0, m);
  cMsg_CgQ1V1fg_sendMessage(_c, 0, m);
  cMsg_Vw0dDa6F_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_SRRdJwWn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_Shg6XAHJ_sendMessage(_c, 0, m);
  cMsg_KgB7QJ0R_sendMessage(_c, 0, m);
  cMsg_hDpspC6p_sendMessage(_c, 0, m);
  cMsg_o7exz0eS_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_n89d6y3q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_Shg6XAHJ_sendMessage(_c, 0, m);
  cMsg_qE4Qprjc_sendMessage(_c, 0, m);
  cMsg_SR8WNT7g_sendMessage(_c, 0, m);
  cMsg_tzHaMyKw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vIZuLZNn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_KGQ4AMq5_sendMessage(_c, 0, m);
  cMsg_Vw0dDa6F_sendMessage(_c, 0, m);
  cMsg_HuFcQIqG_sendMessage(_c, 0, m);
  cMsg_SR8WNT7g_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_pKd3umsy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_KGQ4AMq5_sendMessage(_c, 0, m);
  cMsg_tzHaMyKw_sendMessage(_c, 0, m);
  cMsg_KgB7QJ0R_sendMessage(_c, 0, m);
  cMsg_oKBsfdRn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Jqb56qnU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zIqdcK4a_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_QR5S2sf7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_m74EycTM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zIqdcK4a_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_QR5S2sf7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oBvzUxg0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2GotZKWq_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Xqahgyov_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_AvKUpqYu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2GotZKWq_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Xqahgyov_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_v6yHB7h9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_xTXhFFHk_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_MhNeQI6k_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_AKouOdnx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_xTXhFFHk_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_MhNeQI6k_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_bbCRhc0d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_Jqb56qnU_sendMessage(_c, 0, m);
  cMsg_AvKUpqYu_sendMessage(_c, 0, m);
  cMsg_AKouOdnx_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Shg6XAHJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_m74EycTM_sendMessage(_c, 0, m);
  cMsg_oBvzUxg0_sendMessage(_c, 0, m);
  cMsg_AKouOdnx_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_KGQ4AMq5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_m74EycTM_sendMessage(_c, 0, m);
  cMsg_AvKUpqYu_sendMessage(_c, 0, m);
  cMsg_v6yHB7h9_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_V9aFkZtK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_rpoFnSBA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_pKh3xg9R_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_GdjpYP6i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_rpoFnSBA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_pKh3xg9R_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_n3srV9vp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lgK2cNH7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_7VMfPGHF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_pFx7z45r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lgK2cNH7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_7VMfPGHF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_kj8jGsgI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ayiDMcET_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_7bkIzOVu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_g73pXvLq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ayiDMcET_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_7bkIzOVu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KgB7QJ0R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_GdjpYP6i_sendMessage(_c, 0, m);
  cMsg_pFx7z45r_sendMessage(_c, 0, m);
  cMsg_g73pXvLq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_fiIEAkpj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_GdjpYP6i_sendMessage(_c, 0, m);
  cMsg_n3srV9vp_sendMessage(_c, 0, m);
  cMsg_kj8jGsgI_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_qE4Qprjc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_V9aFkZtK_sendMessage(_c, 0, m);
  cMsg_pFx7z45r_sendMessage(_c, 0, m);
  cMsg_kj8jGsgI_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_zPEgIPtb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2a27WnkY_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ngivbmgu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_JODVyzho_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2a27WnkY_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ngivbmgu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rLkh09M8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_kHXJOOWo_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Zaf0x0Zv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fG6lY5by_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_kHXJOOWo_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Zaf0x0Zv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_HIncY3ir_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Q6yHOLVI_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_j6PsPTyO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rCzyAiZV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Q6yHOLVI_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_j6PsPTyO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_tzHaMyKw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_zPEgIPtb_sendMessage(_c, 0, m);
  cMsg_fG6lY5by_sendMessage(_c, 0, m);
  cMsg_rCzyAiZV_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Vw0dDa6F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_JODVyzho_sendMessage(_c, 0, m);
  cMsg_rLkh09M8_sendMessage(_c, 0, m);
  cMsg_rCzyAiZV_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_hDpspC6p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_JODVyzho_sendMessage(_c, 0, m);
  cMsg_fG6lY5by_sendMessage(_c, 0, m);
  cMsg_HIncY3ir_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Hf3uWUi5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Mvb8e4QF_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_5Ulrtg1E_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_I7XzREcD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Mvb8e4QF_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_5Ulrtg1E_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_C1W82p68_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_JYg5mQwL_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_e7GZ1JIN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zelEGOo4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_JYg5mQwL_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_e7GZ1JIN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_26S0cfmN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_QKgnepKu_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AhqpiOId_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KApHiuAd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_QKgnepKu_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AhqpiOId_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SR8WNT7g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_I7XzREcD_sendMessage(_c, 0, m);
  cMsg_zelEGOo4_sendMessage(_c, 0, m);
  cMsg_KApHiuAd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_CgQ1V1fg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_I7XzREcD_sendMessage(_c, 0, m);
  cMsg_C1W82p68_sendMessage(_c, 0, m);
  cMsg_26S0cfmN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_oKBsfdRn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_Hf3uWUi5_sendMessage(_c, 0, m);
  cMsg_C1W82p68_sendMessage(_c, 0, m);
  cMsg_KApHiuAd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_HuFcQIqG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_V9aFkZtK_sendMessage(_c, 0, m);
  cMsg_n3srV9vp_sendMessage(_c, 0, m);
  cMsg_g73pXvLq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_o7exz0eS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_Hf3uWUi5_sendMessage(_c, 0, m);
  cMsg_zelEGOo4_sendMessage(_c, 0, m);
  cMsg_26S0cfmN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_CunFlHmT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_sbd9rsvK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_vooCBlOG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_WPX31R9R, 0, m, &cSlice_WPX31R9R_sendMessage);
      break;
    }
    default: {
      cSwitchcase_jcNqXgjO_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_WPX31R9R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_A29hsmg9_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_A29hsmg9_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_35veTBZ2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_A29hsmg9_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_A29hsmg9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_J8JX4lev_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_hQYHxfWz_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_5LJSIxSV, 0, m, &cSlice_5LJSIxSV_sendMessage);
      break;
    }
    default: {
      cSwitchcase_R8lz57Ye_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_5LJSIxSV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_RWGO2q4x_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_RWGO2q4x_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_ZLxderQc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_RWGO2q4x_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_RWGO2q4x_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_A4fKPXII_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_R8lz57Ye_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_7jSXxCnZ, 0, m, &cSlice_7jSXxCnZ_sendMessage);
      break;
    }
    default: {
      cSwitchcase_vooCBlOG_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_7jSXxCnZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_QH9jydR0_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_QH9jydR0_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_7gvbRUho_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_QH9jydR0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_QH9jydR0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_YV9QQVZn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jcNqXgjO_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_xb3PVWVd, 0, m, &cSlice_xb3PVWVd_sendMessage);
      break;
    }
    default: {
      cSwitchcase_Cr6YEBAI_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_xb3PVWVd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_zYnr0dk3_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_zYnr0dk3_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_U8f1GYSP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_zYnr0dk3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_zYnr0dk3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_S4Fliwp5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Cr6YEBAI_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_y1ZiOLig, 0, m, &cSlice_y1ZiOLig_sendMessage);
      break;
    }
    default: {
      cSwitchcase_r1cCo2Jd_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_y1ZiOLig_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_JIjhg19M_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_JIjhg19M_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_2XTjAuWf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_JIjhg19M_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_JIjhg19M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_zXbmkPu5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_u4CEOwnb_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_cZL2Qf4Q_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_cZL2Qf4Q_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_EPtySQS8, 1, m, &cDelay_EPtySQS8_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0LpL8o1A_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_EPtySQS8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_EPtySQS8, m);
  cMsg_KESsN50N_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_cZL2Qf4Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_EPtySQS8, 0, m, &cDelay_EPtySQS8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0LpL8o1A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_EPtySQS8, 0, m, &cDelay_EPtySQS8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_N9g6K5gT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_5b0lFFpz, HV_BINOP_SUBTRACT, 0, m, &cBinop_5b0lFFpz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_2GPfXjW3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_LMXfCSVD_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_UMU9VdC7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_LMXfCSVD_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_1TvLgphb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_N9g6K5gT, 0, m, &cVar_N9g6K5gT_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_p5pbPN1Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bQOuW2Bv_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_kCWIBGNz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6bdDMwb9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_5b0lFFpz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_9mf0wntR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9mf0wntR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_o7c8L0Ls_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_o7c8L0Ls_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_743N0iwV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_FMbJVgOU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_KQMxn4bR_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_DjvUEP0h_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_KQMxn4bR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_2GPfXjW3, 1, m, &cIf_2GPfXjW3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DjvUEP0h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_Pwzjf6TU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_uQocxxxn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UMU9VdC7, 1, m, &cIf_UMU9VdC7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Pwzjf6TU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_2GPfXjW3, 0, m, &cIf_2GPfXjW3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_743N0iwV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UMU9VdC7, 0, m, &cIf_UMU9VdC7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bQOuW2Bv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_6bdDMwb9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_N9g6K5gT, 0, m, &cVar_N9g6K5gT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_kCWIBGNz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_5b0lFFpz, HV_BINOP_SUBTRACT, 1, m, &cBinop_5b0lFFpz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_LMXfCSVD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UgCbAvGG, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_r1cCo2Jd_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_rtlMGM82, 0, m, &cSlice_rtlMGM82_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_rtlMGM82_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_DexM7IZL_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_DexM7IZL_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_5HobttMZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_DexM7IZL_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_DexM7IZL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_8GN31U0Z_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_NjuL8Qmj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_1ZQImshA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_v4GmXdBD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TG9l5fhg, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_xnuFMFB2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_VA6MSucp, 0, m, &cVar_VA6MSucp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_sbd9rsvK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_3SItWbzE, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_4Nq7kGEW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_KUNDykty_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_kaPfBmdg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_u4CEOwnb_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_KESsN50N_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8SoWHIbM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_KUNDykty_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8SoWHIbM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_2kCOOm1v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_KWCcQnxz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_KWCcQnxz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_ESw3vLjk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ESw3vLjk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_dhOv2gwo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_3e1g4Oco_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_zqS7rVJG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_96UmmIHQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_dS7yf57A_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_XTEu5phC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_IFrCYwCT_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_KcCE9QCM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_CXZeSufB_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_6pLq9WCy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_HA2PK7Yx_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_iDWrxXdj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_uYT5u3sf_sendMessage);
  cMsg_vmhAXEGP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_uYT5u3sf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_iDWrxXdj, 1, m, &cVar_iDWrxXdj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_S8wlpfjv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_9v2n5T9E_sendMessage(_c, 0, m);
  cMsg_2L4OhUAt_sendMessage(_c, 0, m);
  cMsg_g1y33AJH_sendMessage(_c, 0, m);
  cMsg_fZpL2DKt_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_XxdMCgLU_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x764F1C36: { // "mix"
      cSlice_onMessage(_c, &Context(_c)->cSlice_cnysjtO6, 0, m, &cSlice_cnysjtO6_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_cnysjtO6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_y0eIbHn2_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_y0eIbHn2_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_dpdFeYaw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_y0eIbHn2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_y0eIbHn2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_3CoH8hpN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_mufOr9pE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.0f, 0, m, &cBinop_nwevphAs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_nwevphAs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 2.0f, 0, m, &cBinop_bb6C87n6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_bb6C87n6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 1.0f, 0, m, &cBinop_k9ktlSPm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_k9ktlSPm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_k9zMXvI6_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_QoGkOSHj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QoGkOSHj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_yMRO4b9c_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_k9zMXvI6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_qr60d9gc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_qr60d9gc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_OcvdyWq2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_yMRO4b9c_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_coL8wv6L_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_coL8wv6L_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_pqBbO9Eb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_OcvdyWq2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_PTlJE51K_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_pqBbO9Eb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_JqqUCuZr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_PTlJE51K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_436i1fPs, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_JqqUCuZr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_I7Tzfx8d, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_IKg4xYBS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_trGs2zR4_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_UzVH5hEP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_EBqSyL47_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_htUfVxDG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CMyveEq0_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_A49vg3Mi_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_vlu50F0h_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_SU0Ry4Ae_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_lEsa2dBO_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NvQsOJhl_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_CMyveEq0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4zNxHbD6_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_A49vg3Mi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_7jb7D0Iy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_vlu50F0h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_LjdrJR1a_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_SU0Ry4Ae_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_CmMTNk1X_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_lEsa2dBO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uIqQs6J4_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_NvQsOJhl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_j68IKSQP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_L8wtyRkz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_z4trlv7S_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_0UziSUln_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_mb47BrtT_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_c2nQJ34L_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_XTChdCQL_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_XTChdCQL_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Zz4G0H1s, 1, m, &cDelay_Zz4G0H1s_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_heIco6an_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Zz4G0H1s_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Zz4G0H1s, m);
  cIf_onMessage(_c, &Context(_c)->cIf_L8wtyRkz, 0, m, &cIf_L8wtyRkz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_XTChdCQL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Zz4G0H1s, 0, m, &cDelay_Zz4G0H1s_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_heIco6an_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Zz4G0H1s, 0, m, &cDelay_Zz4G0H1s_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_EPcmIItU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ipT1HE1Z_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_xswcv1Ck_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_98moOXZP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_jMNS643L_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_8zwceFS7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_8zwceFS7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_c2nQJ34L_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_jMNS643L_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_L8wtyRkz, 1, m, &cIf_L8wtyRkz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xswcv1Ck_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0UziSUln, 0, m, &cIf_0UziSUln_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ipT1HE1Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0UziSUln, 1, m, &cIf_0UziSUln_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_mb47BrtT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_NoLSESq7, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_z4trlv7S_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_NoLSESq7, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_MSJ2rhuD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_5VNggVJO_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_IjxalCac_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_1XcNUZej_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_czo10FOv_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_SipgeQTX_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_SipgeQTX_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_3CI2ijKb, 1, m, &cDelay_3CI2ijKb_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_kPEXHg7e_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_3CI2ijKb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_3CI2ijKb, m);
  cIf_onMessage(_c, &Context(_c)->cIf_MSJ2rhuD, 0, m, &cIf_MSJ2rhuD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SipgeQTX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_3CI2ijKb, 0, m, &cDelay_3CI2ijKb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_kPEXHg7e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_3CI2ijKb, 0, m, &cDelay_3CI2ijKb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_w91aA4Qc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Yl4uCDQo_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_VQYiIA5M_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_zR6uvewQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0NEEUbcx_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_3ONa5HFS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_3ONa5HFS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_czo10FOv_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_0NEEUbcx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_MSJ2rhuD, 1, m, &cIf_MSJ2rhuD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_VQYiIA5M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_IjxalCac, 0, m, &cIf_IjxalCac_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Yl4uCDQo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_IjxalCac, 1, m, &cIf_IjxalCac_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1XcNUZej_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_9zwD0dMe, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_5VNggVJO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_9zwD0dMe, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_A7xbyVIp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_fTad8wn6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_fgtl7ym7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_WV8beEqu_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_tKxqYFKz_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_EXbG2Agq_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_EXbG2Agq_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_afD7ypye, 1, m, &cDelay_afD7ypye_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_czkYWblj_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_afD7ypye_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_afD7ypye, m);
  cIf_onMessage(_c, &Context(_c)->cIf_A7xbyVIp, 0, m, &cIf_A7xbyVIp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EXbG2Agq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_afD7ypye, 0, m, &cDelay_afD7ypye_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_czkYWblj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_afD7ypye, 0, m, &cDelay_afD7ypye_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_EUdo34X0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_MtobHPzs_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_SVw84MN2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_zJ3CtooD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_CkXxI4V3_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_JZGSmEB3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_JZGSmEB3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_tKxqYFKz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_CkXxI4V3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_A7xbyVIp, 1, m, &cIf_A7xbyVIp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_SVw84MN2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_fgtl7ym7, 0, m, &cIf_fgtl7ym7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_MtobHPzs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_fgtl7ym7, 1, m, &cIf_fgtl7ym7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_WV8beEqu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_uuoTx4ER, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_fTad8wn6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_uuoTx4ER, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_6bFljx5e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_SLu4KyDi_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_9L2IP2FC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_1ABn8SUb_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_OgH3KIJY_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_qQaNvIst_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_qQaNvIst_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_udYkQLCD, 1, m, &cDelay_udYkQLCD_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_p1OXYhIZ_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_udYkQLCD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_udYkQLCD, m);
  cIf_onMessage(_c, &Context(_c)->cIf_6bFljx5e, 0, m, &cIf_6bFljx5e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_qQaNvIst_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_udYkQLCD, 0, m, &cDelay_udYkQLCD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_p1OXYhIZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_udYkQLCD, 0, m, &cDelay_udYkQLCD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_3NSPhsOv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_prsDknRU_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ie05193a_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_DgDmw2hv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LfUaUxxK_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0a0sHOZI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0a0sHOZI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_OgH3KIJY_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_LfUaUxxK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_6bFljx5e, 1, m, &cIf_6bFljx5e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_prsDknRU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9L2IP2FC, 1, m, &cIf_9L2IP2FC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ie05193a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9L2IP2FC, 0, m, &cIf_9L2IP2FC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1ABn8SUb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_hT4P1smt, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_SLu4KyDi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_hT4P1smt, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_5VbxKJwu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_v8w5vgag_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_lAQTjRw3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ezDxTaM4_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_tH2o7WiI_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_dhiw0V1w_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_dhiw0V1w_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_CrjxS59O, 1, m, &cDelay_CrjxS59O_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Plf5Y8ev_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_CrjxS59O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_CrjxS59O, m);
  cIf_onMessage(_c, &Context(_c)->cIf_5VbxKJwu, 0, m, &cIf_5VbxKJwu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dhiw0V1w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_CrjxS59O, 0, m, &cDelay_CrjxS59O_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Plf5Y8ev_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_CrjxS59O, 0, m, &cDelay_CrjxS59O_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Zx1foh0r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_nFEsbeMJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HJunB774_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_O7gksBJw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_VFhNvrYx_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_pi4E2h8K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_pi4E2h8K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_tH2o7WiI_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_VFhNvrYx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_5VbxKJwu, 1, m, &cIf_5VbxKJwu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HJunB774_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_lAQTjRw3, 0, m, &cIf_lAQTjRw3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_nFEsbeMJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_lAQTjRw3, 1, m, &cIf_lAQTjRw3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ezDxTaM4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_CTmP5gSj, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_v8w5vgag_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_CTmP5gSj, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_vSYmOfBc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_qNUd0IDM_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_ve8qkMIr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ZnUNp71t_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_O8PLwGGG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_MjZLIcGY_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_MjZLIcGY_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_f1rjqsra, 1, m, &cDelay_f1rjqsra_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_wYsPseqJ_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_f1rjqsra_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_f1rjqsra, m);
  cIf_onMessage(_c, &Context(_c)->cIf_vSYmOfBc, 0, m, &cIf_vSYmOfBc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_MjZLIcGY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_f1rjqsra, 0, m, &cDelay_f1rjqsra_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wYsPseqJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_f1rjqsra, 0, m, &cDelay_f1rjqsra_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_HOiuLpUI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_p8yee4MG_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NHYAXIXN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_eOQC656U_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_a5GsUmh8_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_15B1JHUV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_15B1JHUV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_O8PLwGGG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_a5GsUmh8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_vSYmOfBc, 1, m, &cIf_vSYmOfBc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NHYAXIXN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ve8qkMIr, 0, m, &cIf_ve8qkMIr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_p8yee4MG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ve8qkMIr, 1, m, &cIf_ve8qkMIr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ZnUNp71t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_50dw8X8z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_qNUd0IDM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_50dw8X8z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_zM00CmwM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_9vuD7ilz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_d8rxEpyn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_dbVDplN4_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_p4AjS0zO_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_1PR6OFQN_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_1PR6OFQN_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_kIJpx1QL, 1, m, &cDelay_kIJpx1QL_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6SavGVCV_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_kIJpx1QL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_kIJpx1QL, m);
  cIf_onMessage(_c, &Context(_c)->cIf_zM00CmwM, 0, m, &cIf_zM00CmwM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1PR6OFQN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_kIJpx1QL, 0, m, &cDelay_kIJpx1QL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_6SavGVCV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_kIJpx1QL, 0, m, &cDelay_kIJpx1QL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_upSIpuo1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ViaHvSAg_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_7deQKLsF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Iw5Apks7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_31uw7IdN_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_56Kt3D5V_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_56Kt3D5V_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_p4AjS0zO_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_31uw7IdN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zM00CmwM, 1, m, &cIf_zM00CmwM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_7deQKLsF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_d8rxEpyn, 0, m, &cIf_d8rxEpyn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ViaHvSAg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_d8rxEpyn, 1, m, &cIf_d8rxEpyn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dbVDplN4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_hn3MwQZ7, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_9vuD7ilz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_hn3MwQZ7, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_m937nwGI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ewyjwjAo_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_mPuzOd7I_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_0MgJyN2J_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Ju9NlxMq_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_BaF5uC2g_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_BaF5uC2g_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_4ZUuceVg, 1, m, &cDelay_4ZUuceVg_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_APFDIj1D_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_4ZUuceVg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_4ZUuceVg, m);
  cIf_onMessage(_c, &Context(_c)->cIf_m937nwGI, 0, m, &cIf_m937nwGI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BaF5uC2g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_4ZUuceVg, 0, m, &cDelay_4ZUuceVg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_APFDIj1D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_4ZUuceVg, 0, m, &cDelay_4ZUuceVg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_6tsNv3FW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_l4N3Ozdw_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_azSsFxAb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_m6wKZoWx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_VeY0Zmj6_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_qQveBCE8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_qQveBCE8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_Ju9NlxMq_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_VeY0Zmj6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_m937nwGI, 1, m, &cIf_m937nwGI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_azSsFxAb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_mPuzOd7I, 0, m, &cIf_mPuzOd7I_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_l4N3Ozdw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_mPuzOd7I, 1, m, &cIf_mPuzOd7I_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_0MgJyN2J_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gsvLugpM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_ewyjwjAo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gsvLugpM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_UblHqpPZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_NzSbjJ27_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_zAu43UMQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_5KlJzxtV_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_evEb4mnZ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_88Daur60_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_88Daur60_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_vISuwMCw, 1, m, &cDelay_vISuwMCw_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_RkbYu24O_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_vISuwMCw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_vISuwMCw, m);
  cIf_onMessage(_c, &Context(_c)->cIf_UblHqpPZ, 0, m, &cIf_UblHqpPZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_88Daur60_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_vISuwMCw, 0, m, &cDelay_vISuwMCw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_RkbYu24O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_vISuwMCw, 0, m, &cDelay_vISuwMCw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_wEifnLql_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_H5AAdvLO_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_jQ7hl7iX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_2KYI6Fp1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gGxyiu12_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ume7TCqS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gGxyiu12_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_UblHqpPZ, 1, m, &cIf_UblHqpPZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ume7TCqS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_evEb4mnZ_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_jQ7hl7iX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zAu43UMQ, 0, m, &cIf_zAu43UMQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_H5AAdvLO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zAu43UMQ, 1, m, &cIf_zAu43UMQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_5KlJzxtV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_6UeFb5mq, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_NzSbjJ27_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_6UeFb5mq, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_BW6Km1Rc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_fXAuFSsx_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_8sJEbFQd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_HE1OkNNu_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_IFDfjicv_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_RPpJfo1X_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_RPpJfo1X_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_7YSZMAR9, 1, m, &cDelay_7YSZMAR9_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_F4caWXRe_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_7YSZMAR9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_7YSZMAR9, m);
  cIf_onMessage(_c, &Context(_c)->cIf_BW6Km1Rc, 0, m, &cIf_BW6Km1Rc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_RPpJfo1X_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_7YSZMAR9, 0, m, &cDelay_7YSZMAR9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_F4caWXRe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_7YSZMAR9, 0, m, &cDelay_7YSZMAR9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AIY1tp8f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ueEOFueF_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FoWODY3a_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lTCf9X7h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_z6Joh7w6_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_wyEwhiAP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_z6Joh7w6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_BW6Km1Rc, 1, m, &cIf_BW6Km1Rc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wyEwhiAP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_IFDfjicv_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_ueEOFueF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_8sJEbFQd, 1, m, &cIf_8sJEbFQd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FoWODY3a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_8sJEbFQd, 0, m, &cIf_8sJEbFQd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_HE1OkNNu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_yOW9VxDn, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_fXAuFSsx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_yOW9VxDn, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_Fnrjvx0J_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_iHEd6Vik_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_KBs2cKeO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ESZmkSae_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_xIfV6yE8_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_WU6sqC4D_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_WU6sqC4D_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_co7FWI8X, 1, m, &cDelay_co7FWI8X_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_b99W8DZt_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_co7FWI8X_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_co7FWI8X, m);
  cIf_onMessage(_c, &Context(_c)->cIf_Fnrjvx0J, 0, m, &cIf_Fnrjvx0J_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_WU6sqC4D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_co7FWI8X, 0, m, &cDelay_co7FWI8X_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_b99W8DZt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_co7FWI8X, 0, m, &cDelay_co7FWI8X_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_UxZLnqyi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LeEAx3RT_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0oH7Yuje_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_o5Fi46jD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_B1gCFbWz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_yZF5kyoK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_B1gCFbWz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Fnrjvx0J, 1, m, &cIf_Fnrjvx0J_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_yZF5kyoK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_xIfV6yE8_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_LeEAx3RT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_KBs2cKeO, 1, m, &cIf_KBs2cKeO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0oH7Yuje_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_KBs2cKeO, 0, m, &cIf_KBs2cKeO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ESZmkSae_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_XvhwWoGj, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_iHEd6Vik_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_XvhwWoGj, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_Q86cQbEO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_HIsSPCp6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_nozFIrBj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_b3pTDzOF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_QLGUMt2K_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_pjVYtCcN_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_pjVYtCcN_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_7iwVzLMc, 1, m, &cDelay_7iwVzLMc_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_hmPuFxXv_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_7iwVzLMc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_7iwVzLMc, m);
  cIf_onMessage(_c, &Context(_c)->cIf_Q86cQbEO, 0, m, &cIf_Q86cQbEO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_pjVYtCcN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_7iwVzLMc, 0, m, &cDelay_7iwVzLMc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_hmPuFxXv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_7iwVzLMc, 0, m, &cDelay_7iwVzLMc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Kmsa3UHw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_OPKeAjV8_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_DMLLDveX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_t4b22iDO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LUBTYWNU_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_JPDvBpKZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_JPDvBpKZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_QLGUMt2K_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_LUBTYWNU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Q86cQbEO, 1, m, &cIf_Q86cQbEO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DMLLDveX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_nozFIrBj, 0, m, &cIf_nozFIrBj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_OPKeAjV8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_nozFIrBj, 1, m, &cIf_nozFIrBj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_b3pTDzOF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TFhGkTXs, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_HIsSPCp6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TFhGkTXs, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4zNxHbD6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_fIB4SUg1_sendMessage(_c, 0, m);
  cMsg_MYEpksdJ_sendMessage(_c, 0, m);
  cMsg_HWB1yBWd_sendMessage(_c, 0, m);
  cMsg_9ZXfQmDg_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_7jb7D0Iy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_fIB4SUg1_sendMessage(_c, 0, m);
  cMsg_fsqWczxZ_sendMessage(_c, 0, m);
  cMsg_Kwox3sDN_sendMessage(_c, 0, m);
  cMsg_Qk4EEEZY_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_LjdrJR1a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_WHBw8pBf_sendMessage(_c, 0, m);
  cMsg_fsqWczxZ_sendMessage(_c, 0, m);
  cMsg_9ZXfQmDg_sendMessage(_c, 0, m);
  cMsg_fze8wDI3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_CmMTNk1X_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_WHBw8pBf_sendMessage(_c, 0, m);
  cMsg_D7cULozh_sendMessage(_c, 0, m);
  cMsg_HWB1yBWd_sendMessage(_c, 0, m);
  cMsg_lasHJwyf_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_uIqQs6J4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_jZZS0o3l_sendMessage(_c, 0, m);
  cMsg_Qk4EEEZY_sendMessage(_c, 0, m);
  cMsg_eXleT9ht_sendMessage(_c, 0, m);
  cMsg_HWB1yBWd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_j68IKSQP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_jZZS0o3l_sendMessage(_c, 0, m);
  cMsg_lasHJwyf_sendMessage(_c, 0, m);
  cMsg_fsqWczxZ_sendMessage(_c, 0, m);
  cMsg_nfwIExy4_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_KhwO0G71_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_98moOXZP_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_EPcmIItU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_yocV9qbz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_98moOXZP_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_EPcmIItU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_asPOM6Tc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zR6uvewQ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_w91aA4Qc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_s3w9lx3j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zR6uvewQ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_w91aA4Qc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_P6xR8MxO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zJ3CtooD_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_EUdo34X0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_epxjb9Qw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_zJ3CtooD_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_EUdo34X0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fIB4SUg1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_KhwO0G71_sendMessage(_c, 0, m);
  cMsg_s3w9lx3j_sendMessage(_c, 0, m);
  cMsg_epxjb9Qw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_WHBw8pBf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_yocV9qbz_sendMessage(_c, 0, m);
  cMsg_asPOM6Tc_sendMessage(_c, 0, m);
  cMsg_epxjb9Qw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_jZZS0o3l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_yocV9qbz_sendMessage(_c, 0, m);
  cMsg_s3w9lx3j_sendMessage(_c, 0, m);
  cMsg_P6xR8MxO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_XdM5qayG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_DgDmw2hv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_3NSPhsOv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SMGzej83_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_DgDmw2hv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_3NSPhsOv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_WoYLVKVc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_O7gksBJw_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Zx1foh0r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_VwCxSqVC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_O7gksBJw_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Zx1foh0r_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_MDcVFZW5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_eOQC656U_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_HOiuLpUI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wx3ciBbv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_eOQC656U_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_HOiuLpUI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fsqWczxZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_SMGzej83_sendMessage(_c, 0, m);
  cMsg_VwCxSqVC_sendMessage(_c, 0, m);
  cMsg_wx3ciBbv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_MYEpksdJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_SMGzej83_sendMessage(_c, 0, m);
  cMsg_WoYLVKVc_sendMessage(_c, 0, m);
  cMsg_MDcVFZW5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_D7cULozh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_XdM5qayG_sendMessage(_c, 0, m);
  cMsg_VwCxSqVC_sendMessage(_c, 0, m);
  cMsg_MDcVFZW5_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_hTX6W3Cy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lTCf9X7h_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AIY1tp8f_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_VauoHG3H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lTCf9X7h_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AIY1tp8f_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dTTmkPEP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_o5Fi46jD_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_UxZLnqyi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_lBDicXL4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_o5Fi46jD_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_UxZLnqyi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_kGX4EBDH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_t4b22iDO_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Kmsa3UHw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_nZMttRD8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_t4b22iDO_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Kmsa3UHw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_lasHJwyf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_hTX6W3Cy_sendMessage(_c, 0, m);
  cMsg_lBDicXL4_sendMessage(_c, 0, m);
  cMsg_nZMttRD8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Qk4EEEZY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_VauoHG3H_sendMessage(_c, 0, m);
  cMsg_dTTmkPEP_sendMessage(_c, 0, m);
  cMsg_nZMttRD8_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9ZXfQmDg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_VauoHG3H_sendMessage(_c, 0, m);
  cMsg_lBDicXL4_sendMessage(_c, 0, m);
  cMsg_kGX4EBDH_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_duF7oCOA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Iw5Apks7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_upSIpuo1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_e672Vj6Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Iw5Apks7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_upSIpuo1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_94o47SFV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_m6wKZoWx_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_6tsNv3FW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_qz8wLnoU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_m6wKZoWx_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_6tsNv3FW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wfSex42Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2KYI6Fp1_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_wEifnLql_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_t9qyKUCQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_2KYI6Fp1_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_wEifnLql_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_HWB1yBWd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_e672Vj6Y_sendMessage(_c, 0, m);
  cMsg_qz8wLnoU_sendMessage(_c, 0, m);
  cMsg_t9qyKUCQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Kwox3sDN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_e672Vj6Y_sendMessage(_c, 0, m);
  cMsg_94o47SFV_sendMessage(_c, 0, m);
  cMsg_wfSex42Z_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_nfwIExy4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_duF7oCOA_sendMessage(_c, 0, m);
  cMsg_94o47SFV_sendMessage(_c, 0, m);
  cMsg_t9qyKUCQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_eXleT9ht_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_XdM5qayG_sendMessage(_c, 0, m);
  cMsg_WoYLVKVc_sendMessage(_c, 0, m);
  cMsg_wx3ciBbv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_fze8wDI3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_duF7oCOA_sendMessage(_c, 0, m);
  cMsg_qz8wLnoU_sendMessage(_c, 0, m);
  cMsg_wfSex42Z_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_RrKG4xNY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_xmO0jwYv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_FsAicdo4_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_ORV7sPxf, 0, m, &cSlice_ORV7sPxf_sendMessage);
      break;
    }
    default: {
      cSwitchcase_1nZLjqmQ_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_ORV7sPxf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_DLR8WeDO_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_DLR8WeDO_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_GcjTuTPW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_DLR8WeDO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_DLR8WeDO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_a9dr0eMq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_QqY8NFjw_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_ZQBi7ycl, 0, m, &cSlice_ZQBi7ycl_sendMessage);
      break;
    }
    default: {
      cSwitchcase_9mbg6v5q_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_ZQBi7ycl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_vDLRIxUF_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_vDLRIxUF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_TqBzNWYb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_vDLRIxUF_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_vDLRIxUF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_tUPKHbIY_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_9mbg6v5q_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_gGsz0OV6, 0, m, &cSlice_gGsz0OV6_sendMessage);
      break;
    }
    default: {
      cSwitchcase_FsAicdo4_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_gGsz0OV6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_wBFZeFVX_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_wBFZeFVX_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_EMHmhOt3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_wBFZeFVX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_wBFZeFVX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_DNvwUqp6_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_1nZLjqmQ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_ZjcWOpCE, 0, m, &cSlice_ZjcWOpCE_sendMessage);
      break;
    }
    default: {
      cSwitchcase_Vm4fimDk_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_ZjcWOpCE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_fc5gqdUD_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_fc5gqdUD_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_wnRiXiy4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_fc5gqdUD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_fc5gqdUD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_PPp5Evmm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Vm4fimDk_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_OzKhaLG7, 0, m, &cSlice_OzKhaLG7_sendMessage);
      break;
    }
    default: {
      cSwitchcase_2fb5sMlu_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_OzKhaLG7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_2DRUgaqe_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_2DRUgaqe_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_zIcGzlGa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_2DRUgaqe_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_2DRUgaqe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_x4AqsGDQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_iwXSyZz7_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_lnYeMgw8_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_lnYeMgw8_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_ZMFtjtjR, 1, m, &cDelay_ZMFtjtjR_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_rN7QvHEO_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_ZMFtjtjR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_ZMFtjtjR, m);
  cMsg_E5tigj8F_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_lnYeMgw8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_ZMFtjtjR, 0, m, &cDelay_ZMFtjtjR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_rN7QvHEO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_ZMFtjtjR, 0, m, &cDelay_ZMFtjtjR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_3oZhXiWn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_vdf5PFgh, HV_BINOP_SUBTRACT, 0, m, &cBinop_vdf5PFgh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_qj2CWrww_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_h30wf4eu_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_OTZsxrs1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_h30wf4eu_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_002yFlyr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_3oZhXiWn, 0, m, &cVar_3oZhXiWn_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_EYQBnZ5o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0kl0HwCu_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_67Gey9kq_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_PIdCaULA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_vdf5PFgh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_SWpb4O36_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_SWpb4O36_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_RhFkTQSx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_RhFkTQSx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_VdNpQUps_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_GHd4XOyE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_luZiPTBq_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_qv5vABc8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_luZiPTBq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_qj2CWrww, 1, m, &cIf_qj2CWrww_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_qv5vABc8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_NPNnJRGg_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_fiQxGdwy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_OTZsxrs1, 1, m, &cIf_OTZsxrs1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_NPNnJRGg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_qj2CWrww, 0, m, &cIf_qj2CWrww_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_VdNpQUps_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_OTZsxrs1, 0, m, &cIf_OTZsxrs1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0kl0HwCu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_PIdCaULA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_3oZhXiWn, 0, m, &cVar_3oZhXiWn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_67Gey9kq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_vdf5PFgh, HV_BINOP_SUBTRACT, 1, m, &cBinop_vdf5PFgh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_h30wf4eu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QJkbFjMq, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_2fb5sMlu_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_Xg6TJzfE, 0, m, &cSlice_Xg6TJzfE_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_Xg6TJzfE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_ARE9DCP0_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_ARE9DCP0_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_GAdyyTVt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_ARE9DCP0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_ARE9DCP0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_DY3OxtQJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_trGs2zR4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_VSisnH6c, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_EBqSyL47_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_LGpLTtuO, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_oMSCdJKD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_UzVH5hEP, 0, m, &cVar_UzVH5hEP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_xmO0jwYv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_adrmUAO1, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_siCr5DxI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_UI6wCtu0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_7YohoXTS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_iwXSyZz7_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_E5tigj8F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ei6ijquT, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_UI6wCtu0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ei6ijquT, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_sujyDzAq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_NEoRfUWY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_NEoRfUWY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_zaeDjthu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_zaeDjthu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_q3t8fdtk, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_ZhevNCBF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4XJiweFZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_HRbU7CeG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_XyU2Cni4_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jSX6asPJ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_xK2WQR1l_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ofI7937M_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_kJlY6Bic_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UskGqfiq_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_2bsHRhgq_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_mLZjSYlv_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_xK2WQR1l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_nGMDAt4D_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_ofI7937M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_I7HBiRXu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_kJlY6Bic_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_KOWhwJvR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_UskGqfiq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_n47CklvO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_2bsHRhgq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_VlJnTsKx_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_mLZjSYlv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_wyO6XSrk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_caZyFUBz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_jawpBXtj_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_OFuOxfHS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_BMWWa4lH_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_0AAXLjje_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_IhEuTAWs_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_IhEuTAWs_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_EWBaq5zv, 1, m, &cDelay_EWBaq5zv_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_baIO91xk_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_EWBaq5zv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_EWBaq5zv, m);
  cIf_onMessage(_c, &Context(_c)->cIf_caZyFUBz, 0, m, &cIf_caZyFUBz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_IhEuTAWs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_EWBaq5zv, 0, m, &cDelay_EWBaq5zv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_baIO91xk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_EWBaq5zv, 0, m, &cDelay_EWBaq5zv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_mCd22YKd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_B76vfmuY_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zdH9iPuA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_fCRgXbbA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_mARfWPTy_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_lMVcXlSw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_mARfWPTy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_caZyFUBz, 1, m, &cIf_caZyFUBz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_lMVcXlSw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_0AAXLjje_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_B76vfmuY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_OFuOxfHS, 1, m, &cIf_OFuOxfHS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zdH9iPuA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_OFuOxfHS, 0, m, &cIf_OFuOxfHS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BMWWa4lH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_PBMJsRgR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_jawpBXtj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_PBMJsRgR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_4CIBc9GO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_YbM1aTib_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_BHcKLl7Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_QZKPIhkD_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_zElroXPt_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_VsboC3JE_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_VsboC3JE_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_UcdJVJ4g, 1, m, &cDelay_UcdJVJ4g_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_dpg7gzKg_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_UcdJVJ4g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_UcdJVJ4g, m);
  cIf_onMessage(_c, &Context(_c)->cIf_4CIBc9GO, 0, m, &cIf_4CIBc9GO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_VsboC3JE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_UcdJVJ4g, 0, m, &cDelay_UcdJVJ4g_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_dpg7gzKg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_UcdJVJ4g, 0, m, &cDelay_UcdJVJ4g_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_K1mqi2aL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_GWehNNsB_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QR9hGIC5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_7DQ9V25g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_2apXDPOB_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_248S7ApX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_248S7ApX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_zElroXPt_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_2apXDPOB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_4CIBc9GO, 1, m, &cIf_4CIBc9GO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_GWehNNsB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_BHcKLl7Z, 1, m, &cIf_BHcKLl7Z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QR9hGIC5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_BHcKLl7Z, 0, m, &cIf_BHcKLl7Z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_QZKPIhkD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_IJBRAbAK, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_YbM1aTib_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_IJBRAbAK, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_nQICfSTM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_yKHQ7Aqm_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_6J2nIAA1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_grX6eBR9_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_1vzh3cPX_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_XRoHVuEE_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_XRoHVuEE_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Ab2V712U, 1, m, &cDelay_Ab2V712U_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_xfiWWMW6_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Ab2V712U_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Ab2V712U, m);
  cIf_onMessage(_c, &Context(_c)->cIf_nQICfSTM, 0, m, &cIf_nQICfSTM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_XRoHVuEE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Ab2V712U, 0, m, &cDelay_Ab2V712U_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xfiWWMW6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Ab2V712U, 0, m, &cDelay_Ab2V712U_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_2Ke3ff0w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_uc9kNPi9_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_9F3YRTYk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Qwt8MvMd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_NsJipQLm_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_t2KsLOra_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NsJipQLm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_nQICfSTM, 1, m, &cIf_nQICfSTM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_t2KsLOra_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_1vzh3cPX_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_9F3YRTYk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_6J2nIAA1, 0, m, &cIf_6J2nIAA1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_uc9kNPi9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_6J2nIAA1, 1, m, &cIf_6J2nIAA1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_grX6eBR9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8iFLSOGV, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_yKHQ7Aqm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8iFLSOGV, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_4KqHx45W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_5bQA1SPJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_9gYmSYUC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_LmWOL20F_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_3EOIOQCb_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_KuhBYLXY_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_KuhBYLXY_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_0kDvDfOB, 1, m, &cDelay_0kDvDfOB_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_wXq0jCAm_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_0kDvDfOB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_0kDvDfOB, m);
  cIf_onMessage(_c, &Context(_c)->cIf_4KqHx45W, 0, m, &cIf_4KqHx45W_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KuhBYLXY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_0kDvDfOB, 0, m, &cDelay_0kDvDfOB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wXq0jCAm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_0kDvDfOB, 0, m, &cDelay_0kDvDfOB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rTPdEawi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_udPPU9az_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_5XK78Ywr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XiuwYGHC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_eiCzlHT3_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_754cxZzV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_eiCzlHT3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_4KqHx45W, 1, m, &cIf_4KqHx45W_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_754cxZzV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_3EOIOQCb_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_udPPU9az_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9gYmSYUC, 1, m, &cIf_9gYmSYUC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_5XK78Ywr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9gYmSYUC, 0, m, &cIf_9gYmSYUC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_LmWOL20F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ogPoI4Gm, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_5bQA1SPJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_ogPoI4Gm, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_t1IWtrGQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_S0w3EWGB_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_95YLLLcm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_OkasPhMR_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Z1mBZ3EI_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_3rimgBmw_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_3rimgBmw_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_BqYhunFc, 1, m, &cDelay_BqYhunFc_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_K7r1fi8d_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_BqYhunFc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_BqYhunFc, m);
  cIf_onMessage(_c, &Context(_c)->cIf_t1IWtrGQ, 0, m, &cIf_t1IWtrGQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3rimgBmw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_BqYhunFc, 0, m, &cDelay_BqYhunFc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_K7r1fi8d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_BqYhunFc, 0, m, &cDelay_BqYhunFc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_rkrrKYVf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ohd9nZHa_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_RyYyxFpe_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_YAY0VbtA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_W3Ui7Ur0_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_x1U6aD5Y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_W3Ui7Ur0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_t1IWtrGQ, 1, m, &cIf_t1IWtrGQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_x1U6aD5Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_Z1mBZ3EI_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_ohd9nZHa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_95YLLLcm, 1, m, &cIf_95YLLLcm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_RyYyxFpe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_95YLLLcm, 0, m, &cIf_95YLLLcm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_OkasPhMR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UXOMloGm, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_S0w3EWGB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_UXOMloGm, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_GXFf6yqN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_mlsSvk0t_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_WLOUqG7R_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_cVs1tcTz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ahHrnsMW_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_NOx8sIbL_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_NOx8sIbL_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_FkVbRG5i, 1, m, &cDelay_FkVbRG5i_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FKNq9Dn7_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_FkVbRG5i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_FkVbRG5i, m);
  cIf_onMessage(_c, &Context(_c)->cIf_GXFf6yqN, 0, m, &cIf_GXFf6yqN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_NOx8sIbL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_FkVbRG5i, 0, m, &cDelay_FkVbRG5i_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FKNq9Dn7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_FkVbRG5i, 0, m, &cDelay_FkVbRG5i_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XKTnxC06_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_vpcKx6rz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_mNe64I4Q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_9afq0xuT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_IFjDzEcB_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_g9ct8ATV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_g9ct8ATV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ahHrnsMW_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_IFjDzEcB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_GXFf6yqN, 1, m, &cIf_GXFf6yqN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_vpcKx6rz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WLOUqG7R, 1, m, &cIf_WLOUqG7R_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_mNe64I4Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WLOUqG7R, 0, m, &cIf_WLOUqG7R_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_cVs1tcTz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_maSlUTpQ, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_mlsSvk0t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_maSlUTpQ, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_6rLq8vL2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_leUzPip0_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_yiiFDu2y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_fc6xcd5P_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_rbxhY4Lf_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_ASi2CBsQ_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_ASi2CBsQ_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_btzddTrh, 1, m, &cDelay_btzddTrh_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0YSfJuNq_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_btzddTrh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_btzddTrh, m);
  cIf_onMessage(_c, &Context(_c)->cIf_6rLq8vL2, 0, m, &cIf_6rLq8vL2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ASi2CBsQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_btzddTrh, 0, m, &cDelay_btzddTrh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0YSfJuNq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_btzddTrh, 0, m, &cDelay_btzddTrh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_G1OubQtu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_rkZPQXt1_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_WGI5a80p_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_scYbDRCP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_TIp0xioU_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_K1h6p9rO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_K1h6p9rO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_rbxhY4Lf_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_TIp0xioU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_6rLq8vL2, 1, m, &cIf_6rLq8vL2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WGI5a80p_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_yiiFDu2y, 0, m, &cIf_yiiFDu2y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_rkZPQXt1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_yiiFDu2y, 1, m, &cIf_yiiFDu2y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fc6xcd5P_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QDmTEdfD, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_leUzPip0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_QDmTEdfD, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_6FzwUjcc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_KQj1gEKT_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_XBugU6D8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_zEGJ6dBm_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_R8MOvxHc_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_s7vycQFv_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_s7vycQFv_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_nlYULWDE, 1, m, &cDelay_nlYULWDE_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_3JGSclCu_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_nlYULWDE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_nlYULWDE, m);
  cIf_onMessage(_c, &Context(_c)->cIf_6FzwUjcc, 0, m, &cIf_6FzwUjcc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_s7vycQFv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_nlYULWDE, 0, m, &cDelay_nlYULWDE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_3JGSclCu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_nlYULWDE, 0, m, &cDelay_nlYULWDE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_KUbrqLE7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gzRwdRs2_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0ecH6GIn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ItaMPZKS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_E6RxKsgb_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_w8Px7WNU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_w8Px7WNU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_R8MOvxHc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_E6RxKsgb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_6FzwUjcc, 1, m, &cIf_6FzwUjcc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0ecH6GIn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_XBugU6D8, 0, m, &cIf_XBugU6D8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gzRwdRs2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_XBugU6D8, 1, m, &cIf_XBugU6D8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zEGJ6dBm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_4ncutjso, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_KQj1gEKT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_4ncutjso, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_Wnf71lMp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_EREyipOC_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_DNK6Qfnu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_zf3jmMel_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_qbppClb5_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_pefYgDw8_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_pefYgDw8_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Mo0r9DGV, 1, m, &cDelay_Mo0r9DGV_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_XxNwqM61_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Mo0r9DGV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Mo0r9DGV, m);
  cIf_onMessage(_c, &Context(_c)->cIf_Wnf71lMp, 0, m, &cIf_Wnf71lMp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_pefYgDw8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Mo0r9DGV, 0, m, &cDelay_Mo0r9DGV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_XxNwqM61_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Mo0r9DGV, 0, m, &cDelay_Mo0r9DGV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_oGjbRcmg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wbLqbHYy_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_3DGODedD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MhFheYLv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_0msMok95_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sUXazao9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sUXazao9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_qbppClb5_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_0msMok95_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Wnf71lMp, 1, m, &cIf_Wnf71lMp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_3DGODedD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DNK6Qfnu, 0, m, &cIf_DNK6Qfnu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wbLqbHYy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DNK6Qfnu, 1, m, &cIf_DNK6Qfnu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zf3jmMel_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_X2BBOH47, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_EREyipOC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_X2BBOH47, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_S9GtnnA1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_vHztRKYF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_dBf3MJOh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_bBkIxxje_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_HUeZc6eS_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_8cVEOgJU_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_8cVEOgJU_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_3GoOmsEk, 1, m, &cDelay_3GoOmsEk_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_72cyUle3_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_3GoOmsEk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_3GoOmsEk, m);
  cIf_onMessage(_c, &Context(_c)->cIf_S9GtnnA1, 0, m, &cIf_S9GtnnA1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_8cVEOgJU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_3GoOmsEk, 0, m, &cDelay_3GoOmsEk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_72cyUle3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_3GoOmsEk, 0, m, &cDelay_3GoOmsEk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MjqieJc1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_dacYymad_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_aCC1jFmo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_OlQOZdXU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_oeqMTtqL_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_NcCgw9Lo_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_NcCgw9Lo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_HUeZc6eS_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_oeqMTtqL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_S9GtnnA1, 1, m, &cIf_S9GtnnA1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_dacYymad_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_dBf3MJOh, 1, m, &cIf_dBf3MJOh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aCC1jFmo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_dBf3MJOh, 0, m, &cIf_dBf3MJOh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_bBkIxxje_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_T4AXGh4t, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_vHztRKYF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_T4AXGh4t, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_sj1MqZ0s_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_hAh0Bwmm_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_7W2L7zt1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Kiv78CsI_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_DKnrOYOa_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_dna28xPS_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_dna28xPS_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_bSz74kbb, 1, m, &cDelay_bSz74kbb_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_S6QPIGMx_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_bSz74kbb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_bSz74kbb, m);
  cIf_onMessage(_c, &Context(_c)->cIf_sj1MqZ0s, 0, m, &cIf_sj1MqZ0s_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dna28xPS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_bSz74kbb, 0, m, &cDelay_bSz74kbb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_S6QPIGMx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_bSz74kbb, 0, m, &cDelay_bSz74kbb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WNpP9MCg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_xKz0VHjo_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_GvumUzfc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_unqqPnFu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_A6CwB7g9_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_2OjBis7j_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2OjBis7j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_DKnrOYOa_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_A6CwB7g9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_sj1MqZ0s, 1, m, &cIf_sj1MqZ0s_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_GvumUzfc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_7W2L7zt1, 0, m, &cIf_7W2L7zt1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xKz0VHjo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_7W2L7zt1, 1, m, &cIf_7W2L7zt1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Kiv78CsI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_Ed0waiOo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_hAh0Bwmm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_Ed0waiOo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_MSPSodx5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_F8Kq9NUV_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_wnBD7SOE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_8IZHJz1O_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_aJSVdcn2_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Mz9pxQop_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Mz9pxQop_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_N9fWxzNY, 1, m, &cDelay_N9fWxzNY_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_W1uA7bQL_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_N9fWxzNY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_N9fWxzNY, m);
  cIf_onMessage(_c, &Context(_c)->cIf_MSPSodx5, 0, m, &cIf_MSPSodx5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Mz9pxQop_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_N9fWxzNY, 0, m, &cDelay_N9fWxzNY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_W1uA7bQL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_N9fWxzNY, 0, m, &cDelay_N9fWxzNY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_n3n5g1kY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DCkJHrHT_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_SzgyhMwl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4bskloKe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gwh2sayg_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ux21o4NG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ux21o4NG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_aJSVdcn2_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_gwh2sayg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_MSPSodx5, 1, m, &cIf_MSPSodx5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_SzgyhMwl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wnBD7SOE, 0, m, &cIf_wnBD7SOE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DCkJHrHT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wnBD7SOE, 1, m, &cIf_wnBD7SOE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_8IZHJz1O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_cVGc06ZX, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_F8Kq9NUV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_cVGc06ZX, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_nGMDAt4D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_FkKcb2ir_sendMessage(_c, 0, m);
  cMsg_X99QcWCW_sendMessage(_c, 0, m);
  cMsg_2tBbXe27_sendMessage(_c, 0, m);
  cMsg_ygtIIAuD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_I7HBiRXu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_FkKcb2ir_sendMessage(_c, 0, m);
  cMsg_cZJKDzbJ_sendMessage(_c, 0, m);
  cMsg_I5yJcCm8_sendMessage(_c, 0, m);
  cMsg_L96UCWaZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_KOWhwJvR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_r65KdTw8_sendMessage(_c, 0, m);
  cMsg_cZJKDzbJ_sendMessage(_c, 0, m);
  cMsg_ygtIIAuD_sendMessage(_c, 0, m);
  cMsg_VbtPfOO3_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_n47CklvO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_r65KdTw8_sendMessage(_c, 0, m);
  cMsg_fZGZrU5Q_sendMessage(_c, 0, m);
  cMsg_2tBbXe27_sendMessage(_c, 0, m);
  cMsg_ueR7xrTs_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_VlJnTsKx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_lH7HnAO1_sendMessage(_c, 0, m);
  cMsg_L96UCWaZ_sendMessage(_c, 0, m);
  cMsg_Xo2TxbVI_sendMessage(_c, 0, m);
  cMsg_2tBbXe27_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_wyO6XSrk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_lH7HnAO1_sendMessage(_c, 0, m);
  cMsg_ueR7xrTs_sendMessage(_c, 0, m);
  cMsg_cZJKDzbJ_sendMessage(_c, 0, m);
  cMsg_9y0fGOLU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_yaZTi8Al_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_fCRgXbbA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_mCd22YKd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_0O3vt7fr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_fCRgXbbA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_mCd22YKd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oWnTSmkv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_7DQ9V25g_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_K1mqi2aL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_70knoQJt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_7DQ9V25g_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_K1mqi2aL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_2ePN442k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Qwt8MvMd_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_2Ke3ff0w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_eW4Lwj3y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_Qwt8MvMd_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_2Ke3ff0w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_FkKcb2ir_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_yaZTi8Al_sendMessage(_c, 0, m);
  cMsg_70knoQJt_sendMessage(_c, 0, m);
  cMsg_eW4Lwj3y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_r65KdTw8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_0O3vt7fr_sendMessage(_c, 0, m);
  cMsg_oWnTSmkv_sendMessage(_c, 0, m);
  cMsg_eW4Lwj3y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_lH7HnAO1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_0O3vt7fr_sendMessage(_c, 0, m);
  cMsg_70knoQJt_sendMessage(_c, 0, m);
  cMsg_2ePN442k_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_4pkBR9O3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_XiuwYGHC_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rTPdEawi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_xbqpjLPX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_XiuwYGHC_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rTPdEawi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1huPF7Va_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_YAY0VbtA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rkrrKYVf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_4EdVQDho_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_YAY0VbtA_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_rkrrKYVf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rBnjSyoN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_9afq0xuT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XKTnxC06_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_sMA4eQx1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_9afq0xuT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XKTnxC06_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_cZJKDzbJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_xbqpjLPX_sendMessage(_c, 0, m);
  cMsg_4EdVQDho_sendMessage(_c, 0, m);
  cMsg_sMA4eQx1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_X99QcWCW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_xbqpjLPX_sendMessage(_c, 0, m);
  cMsg_1huPF7Va_sendMessage(_c, 0, m);
  cMsg_rBnjSyoN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_fZGZrU5Q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_4pkBR9O3_sendMessage(_c, 0, m);
  cMsg_4EdVQDho_sendMessage(_c, 0, m);
  cMsg_rBnjSyoN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_cJ7VkNzl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_OlQOZdXU_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_MjqieJc1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dD53CXo5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_OlQOZdXU_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_MjqieJc1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_UDMj40TY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_unqqPnFu_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_WNpP9MCg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_bPYUzpKz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_unqqPnFu_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_WNpP9MCg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_AY28k0ta_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_4bskloKe_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_n3n5g1kY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Uc7tt85Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_4bskloKe_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_n3n5g1kY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ueR7xrTs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_cJ7VkNzl_sendMessage(_c, 0, m);
  cMsg_bPYUzpKz_sendMessage(_c, 0, m);
  cMsg_Uc7tt85Y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_L96UCWaZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_dD53CXo5_sendMessage(_c, 0, m);
  cMsg_UDMj40TY_sendMessage(_c, 0, m);
  cMsg_Uc7tt85Y_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ygtIIAuD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_dD53CXo5_sendMessage(_c, 0, m);
  cMsg_bPYUzpKz_sendMessage(_c, 0, m);
  cMsg_AY28k0ta_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_AfDd34oy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_scYbDRCP_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_G1OubQtu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Dfe9Kj9L_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_scYbDRCP_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_G1OubQtu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_iLbjQT0M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ItaMPZKS_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_KUbrqLE7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_JvQ7sCeK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ItaMPZKS_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_KUbrqLE7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BdoDFTjA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MhFheYLv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_oGjbRcmg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Rl9P74Vm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MhFheYLv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_oGjbRcmg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_2tBbXe27_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_Dfe9Kj9L_sendMessage(_c, 0, m);
  cMsg_JvQ7sCeK_sendMessage(_c, 0, m);
  cMsg_Rl9P74Vm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_I5yJcCm8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_Dfe9Kj9L_sendMessage(_c, 0, m);
  cMsg_iLbjQT0M_sendMessage(_c, 0, m);
  cMsg_BdoDFTjA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9y0fGOLU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_AfDd34oy_sendMessage(_c, 0, m);
  cMsg_iLbjQT0M_sendMessage(_c, 0, m);
  cMsg_Rl9P74Vm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Xo2TxbVI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_4pkBR9O3_sendMessage(_c, 0, m);
  cMsg_1huPF7Va_sendMessage(_c, 0, m);
  cMsg_sMA4eQx1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_VbtPfOO3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_AfDd34oy_sendMessage(_c, 0, m);
  cMsg_JvQ7sCeK_sendMessage(_c, 0, m);
  cMsg_BdoDFTjA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_nPytwmId_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_mkSzHtdR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_2zlpFbln_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_2L1iEtek, 0, m, &cSlice_2L1iEtek_sendMessage);
      break;
    }
    default: {
      cSwitchcase_n96RyrF9_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_2L1iEtek_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_263CXWpZ_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_263CXWpZ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_tjwCGUUl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_263CXWpZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_263CXWpZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_tlvShJ1H_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_5vQTqlUP_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_JAQZg1fS, 0, m, &cSlice_JAQZg1fS_sendMessage);
      break;
    }
    default: {
      cSwitchcase_XkfgxOBv_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_JAQZg1fS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_9auOr8Tj_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_9auOr8Tj_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_OkxvPbCq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_9auOr8Tj_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_9auOr8Tj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_XGxhpSj1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_XkfgxOBv_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_J2NH60K7, 0, m, &cSlice_J2NH60K7_sendMessage);
      break;
    }
    default: {
      cSwitchcase_2zlpFbln_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_J2NH60K7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_nIIEF36g_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_nIIEF36g_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_PP4tW2MC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_nIIEF36g_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_nIIEF36g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_Wkpv9Vsl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_n96RyrF9_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_9ffvyDTC, 0, m, &cSlice_9ffvyDTC_sendMessage);
      break;
    }
    default: {
      cSwitchcase_HYO0WQQ7_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_9ffvyDTC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_hnYmdSyd_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_hnYmdSyd_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_hhvHu7AA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_hnYmdSyd_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_hnYmdSyd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_v2skW5DT_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_HYO0WQQ7_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_wYaC3vOp, 0, m, &cSlice_wYaC3vOp_sendMessage);
      break;
    }
    default: {
      cSwitchcase_fdmchHmX_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_wYaC3vOp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_axgp68d2_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_axgp68d2_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_AehKCGix_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_axgp68d2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_axgp68d2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_Fejs2us2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_aSu4jjtk_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_hwJQsWAr_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_hwJQsWAr_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_eBvITYP9, 1, m, &cDelay_eBvITYP9_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QcgPAM47_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_eBvITYP9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_eBvITYP9, m);
  cMsg_p2ZCzu8S_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_hwJQsWAr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_eBvITYP9, 0, m, &cDelay_eBvITYP9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QcgPAM47_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_eBvITYP9, 0, m, &cDelay_eBvITYP9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_ETd0PQUF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Yryw9a9t, HV_BINOP_SUBTRACT, 0, m, &cBinop_Yryw9a9t_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_SQw7j4CQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_gRArnzsc_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_58L0i0bl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_gRArnzsc_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_CMuR8qRn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_ETd0PQUF, 0, m, &cVar_ETd0PQUF_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_45tOAySj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_HTEWsMW3_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_LFGXT6BU_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_eZTDNsE9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Yryw9a9t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_gdTYyOEy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_gdTYyOEy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_MBw40ict_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MBw40ict_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_6b2iNLS1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_3QvSI6ca_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_hlhGIy1W_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_BonGuI1N_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_hlhGIy1W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_SQw7j4CQ, 1, m, &cIf_SQw7j4CQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_BonGuI1N_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_zBjIpUdo_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_TEW7q8Ph_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_58L0i0bl, 1, m, &cIf_58L0i0bl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zBjIpUdo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_SQw7j4CQ, 0, m, &cIf_SQw7j4CQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_6b2iNLS1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_58L0i0bl, 0, m, &cIf_58L0i0bl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HTEWsMW3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_eZTDNsE9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_ETd0PQUF, 0, m, &cVar_ETd0PQUF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LFGXT6BU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_Yryw9a9t, HV_BINOP_SUBTRACT, 1, m, &cBinop_Yryw9a9t_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_gRArnzsc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_t0DjmaiP, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_fdmchHmX_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_8DYE7hEE, 0, m, &cSlice_8DYE7hEE_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_8DYE7hEE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_rZ9dfEXy_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_rZ9dfEXy_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_WEJNRm8c_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_rZ9dfEXy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_rZ9dfEXy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_QbQDH3nH_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_4XJiweFZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_n4y0vey0, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_XyU2Cni4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_zT4WjDAp, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_t5IZN8NH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_HRbU7CeG, 0, m, &cVar_HRbU7CeG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_mkSzHtdR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_D3xFLs9K, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_JKj5JLUj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_BimSpBVc_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_Fpdhw952_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_aSu4jjtk_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_p2ZCzu8S_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_225hgjbz, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_BimSpBVc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_225hgjbz, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_BZgUcS0d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_Bxusgypj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Bxusgypj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4ExRyK8l_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_4ExRyK8l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_6VexcZJF, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_KI0DoguC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_07PL4HOW_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_s7gCHDLK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_N96UVIz1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_atwbwh98_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_N1BvDbWD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_NLJ5dtoh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_312BWiZO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_UxYXUtWZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_lM2GMkcC_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_MEeaWoyX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_zb2RDkXP_sendMessage);
  cMsg_ExknxxEi_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_zb2RDkXP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_MEeaWoyX, 1, m, &cVar_MEeaWoyX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_eBumaAjW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_obnvhoqt_sendMessage(_c, 0, m);
  cMsg_3fjLN9TB_sendMessage(_c, 0, m);
  cMsg_SG95B5Pd_sendMessage(_c, 0, m);
  cMsg_pupfkkbm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ZYsLHD91_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x764F1C36: { // "mix"
      cSlice_onMessage(_c, &Context(_c)->cSlice_FTxEpQYV, 0, m, &cSlice_FTxEpQYV_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_FTxEpQYV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_FsuQL90f_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_FsuQL90f_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_2IIWTMSO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_FsuQL90f_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_FsuQL90f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_3lKBO5ei_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_bGJYQYOv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 0.0f, 0, m, &cBinop_TyB5OzJt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_TyB5OzJt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 2.0f, 0, m, &cBinop_4XlSz3oC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_4XlSz3oC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 1.0f, 0, m, &cBinop_hYfCOFV0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_hYfCOFV0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_A5GWewdK_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_TktMO0uD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_A5GWewdK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_sJciBnYv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_TktMO0uD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_nSDQ84V9_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_sJciBnYv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_SgZoMOCr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_nSDQ84V9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_v27B66yP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_v27B66yP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, 0.5f, 0, m, &cBinop_bNdHMVNB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_SgZoMOCr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_1NOTpLXH_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cBinop_bNdHMVNB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_x6rKN79S_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_1NOTpLXH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_rbElxj1N, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_x6rKN79S_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 100.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_jkHkClBU, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_V63JdNu0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_wk1Mu0Dy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_U4imW5zu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_1DqcyPkV_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ZzFQmilo_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_3LSXwC2q_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_xXkIjGEG_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zlo7P0Zh_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CBdhe8xu_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_1IJ6PL8Z_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_OIqpvMxJ_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_3LSXwC2q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_WTBZNfSm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_xXkIjGEG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_E63GVUJZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_zlo7P0Zh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_oQXzNH8a_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_CBdhe8xu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_1kA0RIhi_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_1IJ6PL8Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_lCZavKbn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_OIqpvMxJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_eb3sOJVD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_hRuxZLH4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_PfyiS5tv_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_0u47Q93g_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_3pfBCM5s_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_NU0aGB8h_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_9LmDEHcU_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_9LmDEHcU_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_TKHxeOY0, 1, m, &cDelay_TKHxeOY0_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_o3FYOxA3_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_TKHxeOY0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_TKHxeOY0, m);
  cIf_onMessage(_c, &Context(_c)->cIf_hRuxZLH4, 0, m, &cIf_hRuxZLH4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_9LmDEHcU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_TKHxeOY0, 0, m, &cDelay_TKHxeOY0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_o3FYOxA3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_TKHxeOY0, 0, m, &cDelay_TKHxeOY0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_u2oVBdjv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_SwIM1d19_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_9C1zsQ0A_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_8ecpJgjT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_qlDb62eB_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_B70zbV39_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_qlDb62eB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_hRuxZLH4, 1, m, &cIf_hRuxZLH4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_B70zbV39_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_NU0aGB8h_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_SwIM1d19_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0u47Q93g, 1, m, &cIf_0u47Q93g_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_9C1zsQ0A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_0u47Q93g, 0, m, &cIf_0u47Q93g_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3pfBCM5s_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TkJqBeom, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_PfyiS5tv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_TkJqBeom, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_OjfOlQA4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_xgyEFmWt_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_cmSooiFj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ZU1LDVJ4_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ft0A1rCq_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_BOIdvIRw_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_BOIdvIRw_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_RYKcHfG2, 1, m, &cDelay_RYKcHfG2_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_jdkp7rxj_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_RYKcHfG2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_RYKcHfG2, m);
  cIf_onMessage(_c, &Context(_c)->cIf_OjfOlQA4, 0, m, &cIf_OjfOlQA4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BOIdvIRw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_RYKcHfG2, 0, m, &cDelay_RYKcHfG2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_jdkp7rxj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_RYKcHfG2, 0, m, &cDelay_RYKcHfG2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_6MZPBjVk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_nyIKJ70H_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LWS9Dpon_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZNGoC8XF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_nmPMnx1H_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zD7yTBXi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_nmPMnx1H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_OjfOlQA4, 1, m, &cIf_OjfOlQA4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zD7yTBXi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ft0A1rCq_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_nyIKJ70H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_cmSooiFj, 1, m, &cIf_cmSooiFj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LWS9Dpon_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_cmSooiFj, 0, m, &cIf_cmSooiFj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ZU1LDVJ4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JgktvltE, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_xgyEFmWt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JgktvltE, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_cbvnHtRb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_VFDlrQqd_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_BO8YMi5m_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_N7kYao12_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_kIV8lbiw_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_ZpsHzE2N_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_ZpsHzE2N_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_VIsEsFSd, 1, m, &cDelay_VIsEsFSd_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_24H0pD7l_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_VIsEsFSd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_VIsEsFSd, m);
  cIf_onMessage(_c, &Context(_c)->cIf_cbvnHtRb, 0, m, &cIf_cbvnHtRb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ZpsHzE2N_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_VIsEsFSd, 0, m, &cDelay_VIsEsFSd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_24H0pD7l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_VIsEsFSd, 0, m, &cDelay_VIsEsFSd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WtjIkMo8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wbl4QqsY_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_B3kB8hY4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_9gI7Z3yx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_aFl4Svjt_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_DBHhkxQI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DBHhkxQI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_kIV8lbiw_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_aFl4Svjt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_cbvnHtRb, 1, m, &cIf_cbvnHtRb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wbl4QqsY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_BO8YMi5m, 1, m, &cIf_BO8YMi5m_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_B3kB8hY4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_BO8YMi5m, 0, m, &cIf_BO8YMi5m_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_N7kYao12_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JSIk82TP, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_VFDlrQqd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_JSIk82TP, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_DxP0nUcX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_JE1kyudl_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_THQTpFV7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Tsvtr8y8_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_qvRRWijj_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_4WO7EpQx_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_4WO7EpQx_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_45PVxpEO, 1, m, &cDelay_45PVxpEO_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_XlW1qLcX_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_45PVxpEO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_45PVxpEO, m);
  cIf_onMessage(_c, &Context(_c)->cIf_DxP0nUcX, 0, m, &cIf_DxP0nUcX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_4WO7EpQx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_45PVxpEO, 0, m, &cDelay_45PVxpEO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_XlW1qLcX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_45PVxpEO, 0, m, &cDelay_45PVxpEO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_o1wdWkIL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ZIihGC7O_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_31Bq6tst_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_iMF0yTlN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_JhnFbcD3_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_n41NlDHx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_n41NlDHx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_qvRRWijj_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_JhnFbcD3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DxP0nUcX, 1, m, &cIf_DxP0nUcX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_31Bq6tst_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_THQTpFV7, 0, m, &cIf_THQTpFV7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ZIihGC7O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_THQTpFV7, 1, m, &cIf_THQTpFV7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Tsvtr8y8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_VnrnqskK, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_JE1kyudl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_VnrnqskK, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_epnifYji_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_NnRtkzhp_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_WkaOda4Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_13LljQdj_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ebit0cdk_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_ME7sfFYX_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_ME7sfFYX_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_1dFP3pfI, 1, m, &cDelay_1dFP3pfI_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_XEAhEVbg_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_1dFP3pfI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_1dFP3pfI, m);
  cIf_onMessage(_c, &Context(_c)->cIf_epnifYji, 0, m, &cIf_epnifYji_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ME7sfFYX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_1dFP3pfI, 0, m, &cDelay_1dFP3pfI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_XEAhEVbg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_1dFP3pfI, 0, m, &cDelay_1dFP3pfI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_92DNpp4h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_B1mdULB6_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_74zMpemq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_H9PaKDHS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_nodGdIBc_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Xrv6Xihw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_nodGdIBc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_epnifYji, 1, m, &cIf_epnifYji_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Xrv6Xihw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ebit0cdk_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_B1mdULB6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WkaOda4Z, 1, m, &cIf_WkaOda4Z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_74zMpemq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WkaOda4Z, 0, m, &cIf_WkaOda4Z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_13LljQdj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_c6oIfRZl, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_NnRtkzhp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_c6oIfRZl, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_GwLfWoaq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_jwmp3hQn_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_VQZG141H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_1DDKKXwf_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_VgABlKk0_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_I9X4PUi2_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_I9X4PUi2_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_PIlmjk51, 1, m, &cDelay_PIlmjk51_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Tuala2vl_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_PIlmjk51_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_PIlmjk51, m);
  cIf_onMessage(_c, &Context(_c)->cIf_GwLfWoaq, 0, m, &cIf_GwLfWoaq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_I9X4PUi2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_PIlmjk51, 0, m, &cDelay_PIlmjk51_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Tuala2vl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_PIlmjk51, 0, m, &cDelay_PIlmjk51_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_aov70aVS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_RbXm7rTX_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_niy2gBdV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_WrIpY8qT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zmUXSFI1_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_JOgjniL6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_zmUXSFI1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_GwLfWoaq, 1, m, &cIf_GwLfWoaq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_JOgjniL6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_VgABlKk0_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_RbXm7rTX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_VQZG141H, 1, m, &cIf_VQZG141H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_niy2gBdV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_VQZG141H, 0, m, &cIf_VQZG141H_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1DDKKXwf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_FqCHeqQA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_jwmp3hQn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_FqCHeqQA, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_2ctlVBqJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_kD1YDCgY_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_Yp3U8Zld_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_BevEQxgJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_CYFzwCQv_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_6SBmHYMy_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_6SBmHYMy_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_aVjOEWaY, 1, m, &cDelay_aVjOEWaY_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_G5oBxuTm_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_aVjOEWaY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_aVjOEWaY, m);
  cIf_onMessage(_c, &Context(_c)->cIf_2ctlVBqJ, 0, m, &cIf_2ctlVBqJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_6SBmHYMy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_aVjOEWaY, 0, m, &cDelay_aVjOEWaY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_G5oBxuTm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_aVjOEWaY, 0, m, &cDelay_aVjOEWaY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZiHztxF7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_VhDsnfL5_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_FS5sg6UZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_g0pOFrD9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_mmAMH8u0_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_WtV3OojC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WtV3OojC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_CYFzwCQv_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_mmAMH8u0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_2ctlVBqJ, 1, m, &cIf_2ctlVBqJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_VhDsnfL5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Yp3U8Zld, 1, m, &cIf_Yp3U8Zld_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FS5sg6UZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Yp3U8Zld, 0, m, &cIf_Yp3U8Zld_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BevEQxgJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_6MXdMybS, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_kD1YDCgY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_6MXdMybS, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_vI5gjHdj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_DtIn1Fc6_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_eTwT6q0C_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_xnFMB5vJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jFjtqiis_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_CE5rsh3l_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_CE5rsh3l_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_mxjt0dz0, 1, m, &cDelay_mxjt0dz0_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_9amPGhAg_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_mxjt0dz0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_mxjt0dz0, m);
  cIf_onMessage(_c, &Context(_c)->cIf_vI5gjHdj, 0, m, &cIf_vI5gjHdj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_CE5rsh3l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_mxjt0dz0, 0, m, &cDelay_mxjt0dz0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_9amPGhAg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_mxjt0dz0, 0, m, &cDelay_mxjt0dz0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ZwbnOAeb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ufSDR8mH_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QwyrmXCc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_PHZ0JULH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Cr2yEloW_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_C7FgIqdO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_C7FgIqdO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_jFjtqiis_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_Cr2yEloW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_vI5gjHdj, 1, m, &cIf_vI5gjHdj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QwyrmXCc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_eTwT6q0C, 0, m, &cIf_eTwT6q0C_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ufSDR8mH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_eTwT6q0C, 1, m, &cIf_eTwT6q0C_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_xnFMB5vJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_w0usi9ji, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_DtIn1Fc6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_w0usi9ji, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_cNSl2xfz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_uF2kCdj5_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_zZSWcAwy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_zW0GZgaJ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_JhzAkDrR_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_LpwlUQPA_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_LpwlUQPA_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_g9nnTwsO, 1, m, &cDelay_g9nnTwsO_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_G4LNlz4a_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_g9nnTwsO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_g9nnTwsO, m);
  cIf_onMessage(_c, &Context(_c)->cIf_cNSl2xfz, 0, m, &cIf_cNSl2xfz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_LpwlUQPA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_g9nnTwsO, 0, m, &cDelay_g9nnTwsO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_G4LNlz4a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_g9nnTwsO, 0, m, &cDelay_g9nnTwsO_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AGND4gMW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_q3f5wj3B_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_KPNyp05o_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1Qhbl2II_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_9OI3e2Vz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LfsgNMbG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LfsgNMbG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_JhzAkDrR_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_9OI3e2Vz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_cNSl2xfz, 1, m, &cIf_cNSl2xfz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_KPNyp05o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zZSWcAwy, 0, m, &cIf_zZSWcAwy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_q3f5wj3B_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_zZSWcAwy, 1, m, &cIf_zZSWcAwy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_zW0GZgaJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_3ZMyjMzM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uF2kCdj5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_3ZMyjMzM, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_wMn8wSN9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_tiFlWjAM_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_risoWZno_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_qBrTtHa9_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_zF2qHojL_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_2S4h3M1M_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_2S4h3M1M_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_hURFSPh0, 1, m, &cDelay_hURFSPh0_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_WTiUaUVg_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_hURFSPh0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_hURFSPh0, m);
  cIf_onMessage(_c, &Context(_c)->cIf_wMn8wSN9, 0, m, &cIf_wMn8wSN9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_2S4h3M1M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_hURFSPh0, 0, m, &cDelay_hURFSPh0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WTiUaUVg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_hURFSPh0, 0, m, &cDelay_hURFSPh0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_T58YhCEQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Oy1YPWAc_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_kVSq6wz5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_OO7X8qSZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_SjFLChGt_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_q2Iiavzw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_q2Iiavzw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_zF2qHojL_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_SjFLChGt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wMn8wSN9, 1, m, &cIf_wMn8wSN9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_kVSq6wz5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_risoWZno, 0, m, &cIf_risoWZno_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Oy1YPWAc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_risoWZno, 1, m, &cIf_risoWZno_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_qBrTtHa9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YTWmwqf5, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_tiFlWjAM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YTWmwqf5, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_ai5Ea2jd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_RmhvVgzt_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_ofd9W6lr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_7mnxxv4t_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_FRVsKcgw_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_EEDqzoxF_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_EEDqzoxF_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Ax5oo9Yw, 1, m, &cDelay_Ax5oo9Yw_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_H9gyBHcg_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Ax5oo9Yw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Ax5oo9Yw, m);
  cIf_onMessage(_c, &Context(_c)->cIf_ai5Ea2jd, 0, m, &cIf_ai5Ea2jd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_EEDqzoxF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Ax5oo9Yw, 0, m, &cDelay_Ax5oo9Yw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_H9gyBHcg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Ax5oo9Yw, 0, m, &cDelay_Ax5oo9Yw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Q4nt6SCt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_PcMNFUnS_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UYwRCmvh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_xJLiLQaG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_xnXtCw2n_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_rSpckhuE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_rSpckhuE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_FRVsKcgw_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_xnXtCw2n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ai5Ea2jd, 1, m, &cIf_ai5Ea2jd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_UYwRCmvh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ofd9W6lr, 0, m, &cIf_ofd9W6lr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_PcMNFUnS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ofd9W6lr, 1, m, &cIf_ofd9W6lr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_7mnxxv4t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8b1XMKYt, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_RmhvVgzt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_8b1XMKYt, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_8xTDJBHs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Ywrj2NsB_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_dOUAAhdB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_5JQMhBUn_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_witRyrv2_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_g6ker1Lp_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_g6ker1Lp_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_kvWP1jXg, 1, m, &cDelay_kvWP1jXg_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_U3MDOzyF_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_kvWP1jXg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_kvWP1jXg, m);
  cIf_onMessage(_c, &Context(_c)->cIf_8xTDJBHs, 0, m, &cIf_8xTDJBHs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_g6ker1Lp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_kvWP1jXg, 0, m, &cDelay_kvWP1jXg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_U3MDOzyF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_kvWP1jXg, 0, m, &cDelay_kvWP1jXg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_IlP3XyNj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_OC2BWfC5_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_95BYdQhl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_NP6yuu4T_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_CvhISO2n_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CdIbobsp_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_CdIbobsp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_witRyrv2_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_CvhISO2n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_8xTDJBHs, 1, m, &cIf_8xTDJBHs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_95BYdQhl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_dOUAAhdB, 0, m, &cIf_dOUAAhdB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_OC2BWfC5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_dOUAAhdB, 1, m, &cIf_dOUAAhdB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_5JQMhBUn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gpmk9514, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_Ywrj2NsB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_gpmk9514, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_WTBZNfSm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_fgICUwZg_sendMessage(_c, 0, m);
  cMsg_ks9lkcXt_sendMessage(_c, 0, m);
  cMsg_tsYVhHmF_sendMessage(_c, 0, m);
  cMsg_Wo0qg7F7_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_E63GVUJZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_fgICUwZg_sendMessage(_c, 0, m);
  cMsg_hCPoi1VH_sendMessage(_c, 0, m);
  cMsg_RwAgs0Ec_sendMessage(_c, 0, m);
  cMsg_CbdtyEXA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_oQXzNH8a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_9an6b5ed_sendMessage(_c, 0, m);
  cMsg_hCPoi1VH_sendMessage(_c, 0, m);
  cMsg_Wo0qg7F7_sendMessage(_c, 0, m);
  cMsg_C0Om6EEy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_1kA0RIhi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_9an6b5ed_sendMessage(_c, 0, m);
  cMsg_e8tM3vMX_sendMessage(_c, 0, m);
  cMsg_tsYVhHmF_sendMessage(_c, 0, m);
  cMsg_TeUP8dQM_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_lCZavKbn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_U6R8bKLV_sendMessage(_c, 0, m);
  cMsg_CbdtyEXA_sendMessage(_c, 0, m);
  cMsg_ODKoXl6W_sendMessage(_c, 0, m);
  cMsg_tsYVhHmF_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_eb3sOJVD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_U6R8bKLV_sendMessage(_c, 0, m);
  cMsg_TeUP8dQM_sendMessage(_c, 0, m);
  cMsg_hCPoi1VH_sendMessage(_c, 0, m);
  cMsg_TLIzCMzS_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_u98BvvG8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_8ecpJgjT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_u2oVBdjv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SF5cmRAt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_8ecpJgjT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_u2oVBdjv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_940mFrP4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ZNGoC8XF_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_6MZPBjVk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_R6lkeOJD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ZNGoC8XF_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_6MZPBjVk_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ajENhS27_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_9gI7Z3yx_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_WtjIkMo8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_UT7J6tKm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_9gI7Z3yx_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_WtjIkMo8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fgICUwZg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_u98BvvG8_sendMessage(_c, 0, m);
  cMsg_R6lkeOJD_sendMessage(_c, 0, m);
  cMsg_UT7J6tKm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9an6b5ed_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_SF5cmRAt_sendMessage(_c, 0, m);
  cMsg_940mFrP4_sendMessage(_c, 0, m);
  cMsg_UT7J6tKm_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_U6R8bKLV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_SF5cmRAt_sendMessage(_c, 0, m);
  cMsg_R6lkeOJD_sendMessage(_c, 0, m);
  cMsg_ajENhS27_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_lVdlDsQC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_iMF0yTlN_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_o1wdWkIL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_hfG8WTUy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_iMF0yTlN_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_o1wdWkIL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Y7Il9e8n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_H9PaKDHS_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_92DNpp4h_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_jj176Woz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_H9PaKDHS_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_92DNpp4h_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ATNyqKNQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_WrIpY8qT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aov70aVS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_es0L5yiw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_WrIpY8qT_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aov70aVS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_hCPoi1VH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_hfG8WTUy_sendMessage(_c, 0, m);
  cMsg_jj176Woz_sendMessage(_c, 0, m);
  cMsg_es0L5yiw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ks9lkcXt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_hfG8WTUy_sendMessage(_c, 0, m);
  cMsg_Y7Il9e8n_sendMessage(_c, 0, m);
  cMsg_ATNyqKNQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_e8tM3vMX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_lVdlDsQC_sendMessage(_c, 0, m);
  cMsg_jj176Woz_sendMessage(_c, 0, m);
  cMsg_ATNyqKNQ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_MMdGKvxZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_OO7X8qSZ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_T58YhCEQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3Lxk2drj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_OO7X8qSZ_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_T58YhCEQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Sslwtvo9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_xJLiLQaG_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Q4nt6SCt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_68EQBfnJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_xJLiLQaG_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_Q4nt6SCt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_TmBuklyg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_NP6yuu4T_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_IlP3XyNj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_a5DiSW4l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_NP6yuu4T_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_IlP3XyNj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_TeUP8dQM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_MMdGKvxZ_sendMessage(_c, 0, m);
  cMsg_68EQBfnJ_sendMessage(_c, 0, m);
  cMsg_a5DiSW4l_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_CbdtyEXA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_3Lxk2drj_sendMessage(_c, 0, m);
  cMsg_Sslwtvo9_sendMessage(_c, 0, m);
  cMsg_a5DiSW4l_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Wo0qg7F7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_3Lxk2drj_sendMessage(_c, 0, m);
  cMsg_68EQBfnJ_sendMessage(_c, 0, m);
  cMsg_TmBuklyg_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vn3zUCIt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_g0pOFrD9_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ZiHztxF7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_vkoEKydN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_g0pOFrD9_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ZiHztxF7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_8mtc9MIT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_PHZ0JULH_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ZwbnOAeb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_VOppPNn6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_PHZ0JULH_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_ZwbnOAeb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rkTJwR8u_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_1Qhbl2II_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AGND4gMW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_NzNrPkza_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_1Qhbl2II_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_AGND4gMW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_tsYVhHmF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_vkoEKydN_sendMessage(_c, 0, m);
  cMsg_VOppPNn6_sendMessage(_c, 0, m);
  cMsg_NzNrPkza_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_RwAgs0Ec_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_vkoEKydN_sendMessage(_c, 0, m);
  cMsg_8mtc9MIT_sendMessage(_c, 0, m);
  cMsg_rkTJwR8u_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_TLIzCMzS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_vn3zUCIt_sendMessage(_c, 0, m);
  cMsg_8mtc9MIT_sendMessage(_c, 0, m);
  cMsg_NzNrPkza_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ODKoXl6W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_lVdlDsQC_sendMessage(_c, 0, m);
  cMsg_Y7Il9e8n_sendMessage(_c, 0, m);
  cMsg_es0L5yiw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_C0Om6EEy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_vn3zUCIt_sendMessage(_c, 0, m);
  cMsg_VOppPNn6_sendMessage(_c, 0, m);
  cMsg_rkTJwR8u_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_HRZqLKqd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_w3gTv2Uy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_PrxsLHmG_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_0ZHN1N8B, 0, m, &cSlice_0ZHN1N8B_sendMessage);
      break;
    }
    default: {
      cSwitchcase_HVHCSalJ_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_0ZHN1N8B_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_Hwv0l2fs_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_Hwv0l2fs_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_itik8I1i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_Hwv0l2fs_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_Hwv0l2fs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_fbjDC3zp_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_PhtVSEec_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_tw7jn2zk, 0, m, &cSlice_tw7jn2zk_sendMessage);
      break;
    }
    default: {
      cSwitchcase_iGe1sgGn_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_tw7jn2zk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_B4iZLKpr_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_B4iZLKpr_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_zJ0DTtW8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_B4iZLKpr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_B4iZLKpr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_aZYlD2gk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_iGe1sgGn_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_zUVq5yTi, 0, m, &cSlice_zUVq5yTi_sendMessage);
      break;
    }
    default: {
      cSwitchcase_PrxsLHmG_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_zUVq5yTi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_QAWKsh5H_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_QAWKsh5H_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_8qFyfhTt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_QAWKsh5H_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_QAWKsh5H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_JRXNrRlD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_HVHCSalJ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_8WPMLH0M, 0, m, &cSlice_8WPMLH0M_sendMessage);
      break;
    }
    default: {
      cSwitchcase_jGwpNGma_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_8WPMLH0M_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_P7hCOE26_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_P7hCOE26_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_5WboOmCL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_P7hCOE26_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_P7hCOE26_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_69voIQmF_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jGwpNGma_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_CYoZ61Af, 0, m, &cSlice_CYoZ61Af_sendMessage);
      break;
    }
    default: {
      cSwitchcase_6RVA8wUi_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_CYoZ61Af_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_U7qIGF8V_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_U7qIGF8V_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_CmgJsT2G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_U7qIGF8V_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_U7qIGF8V_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_g7Xyl5Ts_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_K3AQ1z8T_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Vh2VBaeL_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Vh2VBaeL_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_d81baJ7W, 1, m, &cDelay_d81baJ7W_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_aLvVTm3i_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_d81baJ7W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_d81baJ7W, m);
  cMsg_18HqyApD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Vh2VBaeL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_d81baJ7W, 0, m, &cDelay_d81baJ7W_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_aLvVTm3i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_d81baJ7W, 0, m, &cDelay_d81baJ7W_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_eQp8tbe3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1jln4NUD, HV_BINOP_SUBTRACT, 0, m, &cBinop_1jln4NUD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_VzvOtQkL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_dfck31IF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_ypzjjpo4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_dfck31IF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_UvKpHaWA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_eQp8tbe3, 0, m, &cVar_eQp8tbe3_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_xQgzju4H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_s3SvLyrk_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_fvHb3ev7_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_3mYH6sks_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1jln4NUD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_qelRVDVE_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_qelRVDVE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_fV1tjvG2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_fV1tjvG2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_lbcxIk8v_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_EJrvLW04_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Jp4v3jBw_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_8AdEKAH2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_8AdEKAH2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_h7BJ9l3D_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_Jp4v3jBw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_VzvOtQkL, 1, m, &cIf_VzvOtQkL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_6IhfFONz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ypzjjpo4, 1, m, &cIf_ypzjjpo4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_h7BJ9l3D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_VzvOtQkL, 0, m, &cIf_VzvOtQkL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lbcxIk8v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ypzjjpo4, 0, m, &cIf_ypzjjpo4_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_fvHb3ev7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1jln4NUD, HV_BINOP_SUBTRACT, 1, m, &cBinop_1jln4NUD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_s3SvLyrk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_3mYH6sks_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_eQp8tbe3, 0, m, &cVar_eQp8tbe3_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dfck31IF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_2dWBcy6Z, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_6RVA8wUi_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_qWV0nL4O, 0, m, &cSlice_qWV0nL4O_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_qWV0nL4O_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_Ezola3Kz_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_Ezola3Kz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_oY29yRWB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_Ezola3Kz_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_Ezola3Kz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_L2MVKoIl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_wk1Mu0Dy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_DSPEe7yf, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_1DqcyPkV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_Xt4uhHeO, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_BWeEIQRI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_U4imW5zu, 0, m, &cVar_U4imW5zu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_w3gTv2Uy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_RLNFQdZ1, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_ZCv1Sv2f_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_K3AQ1z8T_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_hqm3oH2v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_NaV35ozw_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_18HqyApD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_lLssYiPy, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_NaV35ozw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_lLssYiPy, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_rMzTZdqJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_3oqqpg93_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_3oqqpg93_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_jlRpn0hv_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_jlRpn0hv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_emHMr1W4, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_u6HPV7b8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_7pD1IzEn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_hJTkuq3x_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_wATh4nxU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_Br4ghokq_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x0: { // "0.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_43Vock6v_sendMessage);
      break;
    }
    case 0x3F800000: { // "1.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_OI9iVdmI_sendMessage);
      break;
    }
    case 0x40000000: { // "2.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_CTZY5Tj1_sendMessage);
      break;
    }
    case 0x40400000: { // "3.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_qforK3we_sendMessage);
      break;
    }
    case 0x40800000: { // "4.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LiuM9gZQ_sendMessage);
      break;
    }
    case 0x40A00000: { // "5.0"
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HykgLWJG_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cCast_43Vock6v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_qIuK63V1_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_OI9iVdmI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_xSRykiev_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_CTZY5Tj1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_uci70nqU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_qforK3we_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_s3OkSNEN_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_LiuM9gZQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_zTt4HSv0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_HykgLWJG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_SEqcQIQB_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cIf_CgkF4SEh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_qUzncRmE_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_eLbVDaVI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_LRzsX6im_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_8eNw28ZE_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_kyj0mepR_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_kyj0mepR_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_huU0tQ7j, 1, m, &cDelay_huU0tQ7j_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_QqJE6eej_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_huU0tQ7j_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_huU0tQ7j, m);
  cIf_onMessage(_c, &Context(_c)->cIf_CgkF4SEh, 0, m, &cIf_CgkF4SEh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_kyj0mepR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_huU0tQ7j, 0, m, &cDelay_huU0tQ7j_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_QqJE6eej_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_huU0tQ7j, 0, m, &cDelay_huU0tQ7j_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_XZDtuXS2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_WB2QPc4n_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_lB10XEam_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_cEXGaF5i_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_rvQYL84T_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_4228NDP5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_4228NDP5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_8eNw28ZE_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_rvQYL84T_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_CgkF4SEh, 1, m, &cIf_CgkF4SEh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_lB10XEam_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_eLbVDaVI, 0, m, &cIf_eLbVDaVI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WB2QPc4n_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_eLbVDaVI, 1, m, &cIf_eLbVDaVI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_LRzsX6im_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_We3kXJi3, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_qUzncRmE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_We3kXJi3, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_so9jvf99_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_p4ffIEDv_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_AY6HjMZS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_J8FyP2BQ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_mVZReXA4_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Up0CBOuo_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Up0CBOuo_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_3u1NkgB6, 1, m, &cDelay_3u1NkgB6_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_hId4DYVd_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_3u1NkgB6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_3u1NkgB6, m);
  cIf_onMessage(_c, &Context(_c)->cIf_so9jvf99, 0, m, &cIf_so9jvf99_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Up0CBOuo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_3u1NkgB6, 0, m, &cDelay_3u1NkgB6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_hId4DYVd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_3u1NkgB6, 0, m, &cDelay_3u1NkgB6_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_1hBBVkXR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_DxoZZnPw_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_1bOnGUvW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_AECZFhH1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_b4QS1sqZ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_WsW8OAWB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_b4QS1sqZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_so9jvf99, 1, m, &cIf_so9jvf99_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WsW8OAWB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_mVZReXA4_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_1bOnGUvW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_AY6HjMZS, 0, m, &cIf_AY6HjMZS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_DxoZZnPw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_AY6HjMZS, 1, m, &cIf_AY6HjMZS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_J8FyP2BQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_KAUdhPPU, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_p4ffIEDv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_KAUdhPPU, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_9uEltQ6o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_lPjqIsTc_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_IhU68FEl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_IubOGooN_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_66beRf3f_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_eT0XnrPZ_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_eT0XnrPZ_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_dU8Vp9gl, 1, m, &cDelay_dU8Vp9gl_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_0rt7K4Vf_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_dU8Vp9gl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_dU8Vp9gl, m);
  cIf_onMessage(_c, &Context(_c)->cIf_9uEltQ6o, 0, m, &cIf_9uEltQ6o_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_eT0XnrPZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_dU8Vp9gl, 0, m, &cDelay_dU8Vp9gl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_0rt7K4Vf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_dU8Vp9gl, 0, m, &cDelay_dU8Vp9gl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_BHqfLcZT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_mMh8CkbQ_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UvGdNW2e_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_QCVmF91F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_FGxtWovg_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_2wroKWAr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_FGxtWovg_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_9uEltQ6o, 1, m, &cIf_9uEltQ6o_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2wroKWAr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_66beRf3f_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_UvGdNW2e_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_IhU68FEl, 0, m, &cIf_IhU68FEl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_mMh8CkbQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_IhU68FEl, 1, m, &cIf_IhU68FEl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_IubOGooN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_OPvamDGw, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_lPjqIsTc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_OPvamDGw, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_WVUMLUcL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_SSLjo78a_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_Rd4sPitH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Zc45fFKL_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_xSTh2kG6_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_yhASkPLb_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_yhASkPLb_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_R5NjUSdF, 1, m, &cDelay_R5NjUSdF_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_bmempDHi_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_R5NjUSdF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_R5NjUSdF, m);
  cIf_onMessage(_c, &Context(_c)->cIf_WVUMLUcL, 0, m, &cIf_WVUMLUcL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_yhASkPLb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_R5NjUSdF, 0, m, &cDelay_R5NjUSdF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bmempDHi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_R5NjUSdF, 0, m, &cDelay_R5NjUSdF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_HtepidvV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_1RfJoxgD_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_EbaEy36z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_lEoQ8Wk7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_rU315Hpz_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Nw2VvsjB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Nw2VvsjB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_xSTh2kG6_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_rU315Hpz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_WVUMLUcL, 1, m, &cIf_WVUMLUcL_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_EbaEy36z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Rd4sPitH, 0, m, &cIf_Rd4sPitH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_1RfJoxgD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_Rd4sPitH, 1, m, &cIf_Rd4sPitH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Zc45fFKL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_yw79kbKp, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_SSLjo78a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_yw79kbKp, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_DvTMFpFs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_nDiAnTGC_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_wydTbgUV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_V3k2kboB_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_OZa8TpLy_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_tfbUcLCl_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_tfbUcLCl_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_Et29iBH1, 1, m, &cDelay_Et29iBH1_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_xkRiGYNZ_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_Et29iBH1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_Et29iBH1, m);
  cIf_onMessage(_c, &Context(_c)->cIf_DvTMFpFs, 0, m, &cIf_DvTMFpFs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_tfbUcLCl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_Et29iBH1, 0, m, &cDelay_Et29iBH1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_xkRiGYNZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_Et29iBH1, 0, m, &cDelay_Et29iBH1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_aP9cHuTH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_9qJdGiCn_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_wx14cfnV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_MeUZ1Qaj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_5hoSsAtP_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_HFoUpKPQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_HFoUpKPQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_OZa8TpLy_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_5hoSsAtP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_DvTMFpFs, 1, m, &cIf_DvTMFpFs_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_wx14cfnV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wydTbgUV, 0, m, &cIf_wydTbgUV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_9qJdGiCn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_wydTbgUV, 1, m, &cIf_wydTbgUV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_V3k2kboB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_aG95Bhw8, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_nDiAnTGC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_aG95Bhw8, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_SSl9in7x_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_450IUKkw_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_l03sluUu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_C7j3U4uz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_oY66Td7N_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_Vlw8YQAY_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_Vlw8YQAY_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_G9mGlW6m, 1, m, &cDelay_G9mGlW6m_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_tQVT5zIe_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_G9mGlW6m_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_G9mGlW6m, m);
  cIf_onMessage(_c, &Context(_c)->cIf_SSl9in7x, 0, m, &cIf_SSl9in7x_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Vlw8YQAY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_G9mGlW6m, 0, m, &cDelay_G9mGlW6m_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_tQVT5zIe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_G9mGlW6m, 0, m, &cDelay_G9mGlW6m_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_bdqIiJLP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_pFaMS6pA_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_pLpY18oW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_73bDxHir_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_TgddmvfK_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sUI1ZkOf_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_TgddmvfK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_SSl9in7x, 1, m, &cIf_SSl9in7x_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_sUI1ZkOf_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_oY66Td7N_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_pLpY18oW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_l03sluUu, 0, m, &cIf_l03sluUu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_pFaMS6pA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_l03sluUu, 1, m, &cIf_l03sluUu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_C7j3U4uz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_B2BiVPLR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_450IUKkw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_B2BiVPLR, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_CsvwRLqz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_cTPPrwii_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_qskZZuWi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_wirBRAMs_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_xveEHH1C_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_klEvuObc_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_klEvuObc_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_05ou24Aq, 1, m, &cDelay_05ou24Aq_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_1y10mSJs_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_05ou24Aq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_05ou24Aq, m);
  cIf_onMessage(_c, &Context(_c)->cIf_CsvwRLqz, 0, m, &cIf_CsvwRLqz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_klEvuObc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_05ou24Aq, 0, m, &cDelay_05ou24Aq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_1y10mSJs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_05ou24Aq, 0, m, &cDelay_05ou24Aq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_CFe8jgZU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_3Pswiz8a_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_exQ6yeZl_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_gj3oDbFR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_OlVfhwvi_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_RS35WeAy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_OlVfhwvi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_CsvwRLqz, 1, m, &cIf_CsvwRLqz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_RS35WeAy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_xveEHH1C_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_3Pswiz8a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_qskZZuWi, 1, m, &cIf_qskZZuWi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_exQ6yeZl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_qskZZuWi, 0, m, &cIf_qskZZuWi_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wirBRAMs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_coxYZNAo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_cTPPrwii_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_coxYZNAo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_TvH7a5lK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_7eGyfxLz_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_kuhc0Rct_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_11dYak3m_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_A9P5eKii_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_KER8H3jG_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_KER8H3jG_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_gSyQyjWP, 1, m, &cDelay_gSyQyjWP_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_gdV47I5u_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_gSyQyjWP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_gSyQyjWP, m);
  cIf_onMessage(_c, &Context(_c)->cIf_TvH7a5lK, 0, m, &cIf_TvH7a5lK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KER8H3jG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_gSyQyjWP, 0, m, &cDelay_gSyQyjWP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_gdV47I5u_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_gSyQyjWP, 0, m, &cDelay_gSyQyjWP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_hJAezV2F_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_5fkGAsy9_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_uYmXaLyy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_BJbjwNUU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bqNoUJYv_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LXjoxzQu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_bqNoUJYv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_TvH7a5lK, 1, m, &cIf_TvH7a5lK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LXjoxzQu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_A9P5eKii_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_5fkGAsy9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_kuhc0Rct, 1, m, &cIf_kuhc0Rct_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_uYmXaLyy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_kuhc0Rct, 0, m, &cIf_kuhc0Rct_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_11dYak3m_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_lOjst2XV, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_7eGyfxLz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_lOjst2XV, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_VVakyEXH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_xzARr9U3_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_iTG4cmUt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_ybEuwJyF_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_ZhKt5dXX_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_a7MCRPrH_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_a7MCRPrH_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_8RC82jrB, 1, m, &cDelay_8RC82jrB_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_yxF7d5ia_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_8RC82jrB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_8RC82jrB, m);
  cIf_onMessage(_c, &Context(_c)->cIf_VVakyEXH, 0, m, &cIf_VVakyEXH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_a7MCRPrH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_8RC82jrB, 0, m, &cDelay_8RC82jrB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_yxF7d5ia_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_8RC82jrB, 0, m, &cDelay_8RC82jrB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_VSp3EGVU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_yRFBTVc2_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Pew4fdHB_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_ME2BTJ5A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_E5DTA8nU_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_lYRJRSwH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_lYRJRSwH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ZhKt5dXX_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_E5DTA8nU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_VVakyEXH, 1, m, &cIf_VVakyEXH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_yRFBTVc2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iTG4cmUt, 1, m, &cIf_iTG4cmUt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Pew4fdHB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iTG4cmUt, 0, m, &cIf_iTG4cmUt_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ybEuwJyF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_y4bKytXG, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_xzARr9U3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_y4bKytXG, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_hnW2RDF1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_Zm193lRR_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_un7OCP5w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_v2vei9tb_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_nePA1Ad5_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_h95gotDP_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_h95gotDP_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_NrfF1O3K, 1, m, &cDelay_NrfF1O3K_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_2l80HZWp_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_NrfF1O3K_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_NrfF1O3K, m);
  cIf_onMessage(_c, &Context(_c)->cIf_hnW2RDF1, 0, m, &cIf_hnW2RDF1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_h95gotDP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_NrfF1O3K, 0, m, &cDelay_NrfF1O3K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_2l80HZWp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_NrfF1O3K, 0, m, &cDelay_NrfF1O3K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_42XqqJMZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_1feZLr7w_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_JypEUcSd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_usQwU4U3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_TTmqmgIc_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_qZmdms2U_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_qZmdms2U_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_nePA1Ad5_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_TTmqmgIc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_hnW2RDF1, 1, m, &cIf_hnW2RDF1_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_JypEUcSd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_un7OCP5w, 0, m, &cIf_un7OCP5w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_1feZLr7w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_un7OCP5w, 1, m, &cIf_un7OCP5w_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_v2vei9tb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_pKRE2uPH, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_Zm193lRR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_pKRE2uPH, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_ImiqxEp5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_QqQGSYZ8_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_2RJnOJgw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_fGn0mW71_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_YYuJwrnY_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_BbX5zySV_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_BbX5zySV_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_LixpknZy, 1, m, &cDelay_LixpknZy_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_AlPpM5pC_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_LixpknZy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_LixpknZy, m);
  cIf_onMessage(_c, &Context(_c)->cIf_ImiqxEp5, 0, m, &cIf_ImiqxEp5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_BbX5zySV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_LixpknZy, 0, m, &cDelay_LixpknZy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_AlPpM5pC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_LixpknZy, 0, m, &cDelay_LixpknZy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_GbPof5JM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_o8OvEzP7_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_e2i05ZwQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_iFyt5B76_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_eeO4CO4h_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_kieH2OZx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_eeO4CO4h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_ImiqxEp5, 1, m, &cIf_ImiqxEp5_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_kieH2OZx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_YYuJwrnY_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_o8OvEzP7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_2RJnOJgw, 1, m, &cIf_2RJnOJgw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_e2i05ZwQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_2RJnOJgw, 0, m, &cIf_2RJnOJgw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fGn0mW71_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_sj6MO7R2, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_QqQGSYZ8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_sj6MO7R2, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cIf_KIefjQYA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_XKhZHmNY_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_iCmgiWL2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_jGnFoKZQ_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cSwitchcase_cDlaDdnV_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_17kMHPkd_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_17kMHPkd_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_XYZd6bwI, 1, m, &cDelay_XYZd6bwI_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_LInhYYX6_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_XYZd6bwI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_XYZd6bwI, m);
  cIf_onMessage(_c, &Context(_c)->cIf_KIefjQYA, 0, m, &cIf_KIefjQYA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_17kMHPkd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_XYZd6bwI, 0, m, &cDelay_XYZd6bwI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_LInhYYX6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_XYZd6bwI, 0, m, &cDelay_XYZd6bwI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_YIuhZmOT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_r978clEk_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_w4XnjOci_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_BLaMQmKh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Xy0l1AMl_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_54l7JNf8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_Xy0l1AMl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_KIefjQYA, 1, m, &cIf_KIefjQYA_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_54l7JNf8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_cDlaDdnV_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_r978clEk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iCmgiWL2, 1, m, &cIf_iCmgiWL2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_w4XnjOci_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_iCmgiWL2, 0, m, &cIf_iCmgiWL2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_jGnFoKZQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YtNcxyvu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_XKhZHmNY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_YtNcxyvu, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_qIuK63V1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CFS");
  cMsg_SUWHZOTq_sendMessage(_c, 0, m);
  cMsg_ZWEmQuus_sendMessage(_c, 0, m);
  cMsg_rsXOeU5N_sendMessage(_c, 0, m);
  cMsg_Npv9Qma2_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_xSRykiev_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "CSF");
  cMsg_SUWHZOTq_sendMessage(_c, 0, m);
  cMsg_ZITgL2Up_sendMessage(_c, 0, m);
  cMsg_2hZKf96w_sendMessage(_c, 0, m);
  cMsg_qPyMAwu7_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_uci70nqU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FCS");
  cMsg_fWEyz3m6_sendMessage(_c, 0, m);
  cMsg_ZITgL2Up_sendMessage(_c, 0, m);
  cMsg_Npv9Qma2_sendMessage(_c, 0, m);
  cMsg_SG7rbwkB_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_s3OkSNEN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "FSC");
  cMsg_fWEyz3m6_sendMessage(_c, 0, m);
  cMsg_9MO8iJcD_sendMessage(_c, 0, m);
  cMsg_rsXOeU5N_sendMessage(_c, 0, m);
  cMsg_aSpdTrSC_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_zTt4HSv0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SCF");
  cMsg_m0cKBvN9_sendMessage(_c, 0, m);
  cMsg_qPyMAwu7_sendMessage(_c, 0, m);
  cMsg_E9QNOpIc_sendMessage(_c, 0, m);
  cMsg_rsXOeU5N_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_SEqcQIQB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "SFC");
  cMsg_m0cKBvN9_sendMessage(_c, 0, m);
  cMsg_aSpdTrSC_sendMessage(_c, 0, m);
  cMsg_ZITgL2Up_sendMessage(_c, 0, m);
  cMsg_qvz4cLwy_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_vMBJp3AA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_cEXGaF5i_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XZDtuXS2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_316mfwRa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_cEXGaF5i_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_XZDtuXS2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dCiClDoZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_AECZFhH1_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_1hBBVkXR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_kXqEurXO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_AECZFhH1_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_1hBBVkXR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_7KjJovmk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_QCVmF91F_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_BHqfLcZT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Hg29HcsD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_QCVmF91F_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_BHqfLcZT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SUWHZOTq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_vMBJp3AA_sendMessage(_c, 0, m);
  cMsg_kXqEurXO_sendMessage(_c, 0, m);
  cMsg_Hg29HcsD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_fWEyz3m6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_316mfwRa_sendMessage(_c, 0, m);
  cMsg_dCiClDoZ_sendMessage(_c, 0, m);
  cMsg_Hg29HcsD_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_m0cKBvN9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_316mfwRa_sendMessage(_c, 0, m);
  cMsg_kXqEurXO_sendMessage(_c, 0, m);
  cMsg_7KjJovmk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_y8wfdQQq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lEoQ8Wk7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_HtepidvV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_B9s5xuS5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_lEoQ8Wk7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_HtepidvV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rxGJ4BE1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MeUZ1Qaj_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aP9cHuTH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_e55bUeZz_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_MeUZ1Qaj_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_aP9cHuTH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_uebHfy2G_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_73bDxHir_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_bdqIiJLP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_8G1yWEse_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_73bDxHir_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_bdqIiJLP_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_ZITgL2Up_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_B9s5xuS5_sendMessage(_c, 0, m);
  cMsg_e55bUeZz_sendMessage(_c, 0, m);
  cMsg_8G1yWEse_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_ZWEmQuus_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_B9s5xuS5_sendMessage(_c, 0, m);
  cMsg_rxGJ4BE1_sendMessage(_c, 0, m);
  cMsg_uebHfy2G_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_9MO8iJcD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_y8wfdQQq_sendMessage(_c, 0, m);
  cMsg_e55bUeZz_sendMessage(_c, 0, m);
  cMsg_uebHfy2G_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_26YQBo5A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_usQwU4U3_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_42XqqJMZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SotRBkLS_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_usQwU4U3_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_42XqqJMZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_q1bwHs0v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_iFyt5B76_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_GbPof5JM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_x3HfL1tD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_iFyt5B76_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_GbPof5JM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_FfhcPDNA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_BLaMQmKh_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_YIuhZmOT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_q9nX9S3E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_BLaMQmKh_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_YIuhZmOT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_aSpdTrSC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_100");
  cMsg_26YQBo5A_sendMessage(_c, 0, m);
  cMsg_x3HfL1tD_sendMessage(_c, 0, m);
  cMsg_q9nX9S3E_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_qPyMAwu7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_010");
  cMsg_SotRBkLS_sendMessage(_c, 0, m);
  cMsg_q1bwHs0v_sendMessage(_c, 0, m);
  cMsg_q9nX9S3E_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_Npv9Qma2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_001");
  cMsg_SotRBkLS_sendMessage(_c, 0, m);
  cMsg_x3HfL1tD_sendMessage(_c, 0, m);
  cMsg_FfhcPDNA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_UbSCn7iw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_gj3oDbFR_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_CFe8jgZU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_gAlpkhL8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_gj3oDbFR_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_CFe8jgZU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_KqdxyLvs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_BJbjwNUU_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_hJAezV2F_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_8UAHoJ2r_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_BJbjwNUU_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_hJAezV2F_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_wPpZIIhU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ME2BTJ5A_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_VSp3EGVU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_k4bNXKmu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_ME2BTJ5A_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_VSp3EGVU_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rsXOeU5N_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_000");
  cMsg_gAlpkhL8_sendMessage(_c, 0, m);
  cMsg_8UAHoJ2r_sendMessage(_c, 0, m);
  cMsg_k4bNXKmu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_2hZKf96w_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_011");
  cMsg_gAlpkhL8_sendMessage(_c, 0, m);
  cMsg_KqdxyLvs_sendMessage(_c, 0, m);
  cMsg_wPpZIIhU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_qvz4cLwy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_UbSCn7iw_sendMessage(_c, 0, m);
  cMsg_KqdxyLvs_sendMessage(_c, 0, m);
  cMsg_k4bNXKmu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_E9QNOpIc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_110");
  cMsg_y8wfdQQq_sendMessage(_c, 0, m);
  cMsg_rxGJ4BE1_sendMessage(_c, 0, m);
  cMsg_8G1yWEse_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_SG7rbwkB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "_101");
  cMsg_UbSCn7iw_sendMessage(_c, 0, m);
  cMsg_8UAHoJ2r_sendMessage(_c, 0, m);
  cMsg_wPpZIIhU_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_tZFaMLqy_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_oM7uUptR_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_D5BryBm0_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x3934C683: { // "fldr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_ncSalD50, 0, m, &cSlice_ncSalD50_sendMessage);
      break;
    }
    default: {
      cSwitchcase_wIME3oIx_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_ncSalD50_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_OlIRUK2v_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_OlIRUK2v_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_4tAy6BVl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_OlIRUK2v_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_OlIRUK2v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_nYfQhPMr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_jdI06ZIc_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x17E7AF38: { // "sqnc"
      cSlice_onMessage(_c, &Context(_c)->cSlice_DiXe5zAq, 0, m, &cSlice_DiXe5zAq_sendMessage);
      break;
    }
    default: {
      cSwitchcase_qYvQ7O8I_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_DiXe5zAq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_W8ewQNuX_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_W8ewQNuX_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_FHXPm0ZB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_W8ewQNuX_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_W8ewQNuX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_STQOLGXP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_qYvQ7O8I_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x941862B0: { // "crshr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_HSPt5mGs, 0, m, &cSlice_HSPt5mGs_sendMessage);
      break;
    }
    default: {
      cSwitchcase_D5BryBm0_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_HSPt5mGs_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_PvSBpXBk_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_PvSBpXBk_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_xweTUVZA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_PvSBpXBk_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_PvSBpXBk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_AaWtmiYG_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_wIME3oIx_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xAB63864D: { // "smthr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_ZCrol453, 0, m, &cSlice_ZCrol453_sendMessage);
      break;
    }
    default: {
      cSwitchcase_V2urZXgZ_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_ZCrol453_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_CaIK7mhP_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_CaIK7mhP_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_Za0eD1qF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_CaIK7mhP_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_CaIK7mhP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_rnNNEjfA_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_V2urZXgZ_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0xC4C86346: { // "lmtr"
      cSlice_onMessage(_c, &Context(_c)->cSlice_v3vavNxi, 0, m, &cSlice_v3vavNxi_sendMessage);
      break;
    }
    default: {
      cSwitchcase_E9nkaCcD_onMessage(_c, NULL, 0, m, NULL);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_v3vavNxi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_m4GNDAcn_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_m4GNDAcn_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_muKdJB5v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_m4GNDAcn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_m4GNDAcn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_8wn4a4Mx_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_vZ5r0Lfc_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x7A5B032D: { // "stop"
      cMsg_dHjGi3Ze_sendMessage(_c, 0, m);
      break;
    }
    default: {
      cMsg_dHjGi3Ze_sendMessage(_c, 0, m);
      cDelay_onMessage(_c, &Context(_c)->cDelay_cKWHqMCx, 1, m, &cDelay_cKWHqMCx_sendMessage);
      cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ULN75hEj_sendMessage);
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cDelay_cKWHqMCx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {
  cDelay_clearExecutingMessage(&Context(_c)->cDelay_cKWHqMCx, m);
  cMsg_1dyin4G0_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_dHjGi3Ze_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "clear");
  cDelay_onMessage(_c, &Context(_c)->cDelay_cKWHqMCx, 0, m, &cDelay_cKWHqMCx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_ULN75hEj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cDelay_onMessage(_c, &Context(_c)->cDelay_cKWHqMCx, 0, m, &cDelay_cKWHqMCx_sendMessage);
}

void Heavy_WSTD_M3NGLR::cVar_gQd62AtM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_tC0L9hZm, HV_BINOP_SUBTRACT, 0, m, &cBinop_tC0L9hZm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cIf_rAn8mkqv_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_TwXKbWKe_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cIf_V1yXuthD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      break;
    }
    case 1: {
      cMsg_TwXKbWKe_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::sEnv_8SZBwKcI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_gQd62AtM, 0, m, &cVar_gQd62AtM_sendMessage);
}

void Heavy_WSTD_M3NGLR::sEnv_bxy3cjVB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_fRysjQ5t_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_WrMDK5z6_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_c7O2ufDG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_tC0L9hZm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_WIoAIXGH_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_WIoAIXGH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_SUBTRACT, 0.0f, 0, m, &cBinop_Oi7HpMRn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_Oi7HpMRn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 100.0f, 0, m, &cBinop_qoBApSRF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_yuBgq1sn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ueR7Rjgu_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UeiKTwcR_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_UeiKTwcR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_CxVsIbog_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cCast_ueR7Rjgu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_rAn8mkqv, 1, m, &cIf_rAn8mkqv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_sopCKXxJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_V1yXuthD, 1, m, &cIf_V1yXuthD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_CxVsIbog_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  cIf_onMessage(_c, &Context(_c)->cIf_rAn8mkqv, 0, m, &cIf_rAn8mkqv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_qoBApSRF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cIf_onMessage(_c, &Context(_c)->cIf_V1yXuthD, 0, m, &cIf_V1yXuthD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_WrMDK5z6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_tC0L9hZm, HV_BINOP_SUBTRACT, 1, m, &cBinop_tC0L9hZm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cCast_fRysjQ5t_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
}

void Heavy_WSTD_M3NGLR::cCast_c7O2ufDG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_gQd62AtM, 0, m, &cVar_gQd62AtM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_TwXKbWKe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_AmjVNRMc, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cSwitchcase_E9nkaCcD_onMessage(HeavyContextInterface *_c, void *o, int letIn, const HvMessage *const m, void *sendMessage) {
  int msgIndex = 0;
  switch (msg_getHash(m, msgIndex)) {
    case 0x6D60E6E: { // "symbol"
      msgIndex = 1;
      break;
    }
  }
  switch (msg_getHash(m, msgIndex)) {
    case 0x811CC33F: { // "gain"
      cSlice_onMessage(_c, &Context(_c)->cSlice_9aADY5jx, 0, m, &cSlice_9aADY5jx_sendMessage);
      break;
    }
    default: {
      break;
    }
  }
}

void Heavy_WSTD_M3NGLR::cSlice_9aADY5jx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  switch (letIn) {
    case 0: {
      cSend_ayZqOdxn_sendMessage(_c, 0, m);
      break;
    }
    case 1: {
      cSend_ayZqOdxn_sendMessage(_c, 0, m);
      break;
    }
    default: return;
  }
}

void Heavy_WSTD_M3NGLR::cVar_Rcs5mJTj_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSend_ayZqOdxn_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cSend_ayZqOdxn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cReceive_a9AmNNPu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_7pD1IzEn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_sbsm5hgo, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_wATh4nxU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_hbYCTr0Q, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_1k94g1J0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_hJTkuq3x, 0, m, &cVar_hJTkuq3x_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_oM7uUptR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_36L2nBRt, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_GnDisZsT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_vZ5r0Lfc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cCast_jjNCXnLF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_pCCNcDcb_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_1dyin4G0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_qNgfQORs, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_pCCNcDcb_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  msg_setFloat(m, 1, 20.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_qNgfQORs, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_u5oEmakO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setFloat(m, 0, 100.0f);
  msg_setElementToFrom(m, 1, n, 0);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 0.0f, 0, m, &cBinop_bStFahao_sendMessage);
}

void Heavy_WSTD_M3NGLR::cBinop_bStFahao_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_h8eifypZ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_h8eifypZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setElementToFrom(m, 0, n, 0);
  msg_setFloat(m, 1, 10.0f);
  sLine_onMessage(_c, &Context(_c)->sLine_15uNlUt4, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cVar_oLzkn6qr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_dHlkn6WK_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_n3jzoP0d_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_fq8gGOEc_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cVar_X6yszb3Y_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_1r7S29s4_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cMsg_l8cRGrNQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "sqnc");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_munyIyvq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "crshr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_YUA5fLMD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_WEbLxt2K, 0, m, &cVar_WEbLxt2K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_DtdQVDRt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "fldr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uUTCtfBo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "smthr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_oqAqAIEo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "lmtr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_uYYB9uJL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 512.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_dZUNub7K, 0, m, &cVar_dZUNub7K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_rI49sh6k_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_h8fgoD7X, 0, m, &cVar_h8fgoD7X_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_B2SJsuMT, 0, m, &cVar_B2SJsuMT_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_vksrWJt7, 0, m, &cVar_vksrWJt7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_CMSDTOOJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 50.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_YUA5fLMD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_k1YiH2TN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "highvol");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_q6dw9q4N_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_etpstFjG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_4o5i0Dh8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "midvol");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_q6dw9q4N_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_etpstFjG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_eIbtwcrF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "lowvol");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_q6dw9q4N_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_etpstFjG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_hkvz5J2l_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "midfreq");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_q6dw9q4N_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_etpstFjG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_ShTlb9K0_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "midq");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_q6dw9q4N_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_etpstFjG_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_dnBMwePR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1337.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_lNtHYdg0, 0, m, &cVar_lNtHYdg0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_XzgL3I3o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_pF65netw, 0, m, &cVar_pF65netw_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_nFdg7mi8, 0, m, &cVar_nFdg7mi8_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_BAig76qw, 0, m, &cVar_BAig76qw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_Bi4ShFsU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "mix");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_zqS7rVJG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "sqnc");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_CXZeSufB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "crshr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_6N78UjYa_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_96UmmIHQ, 0, m, &cVar_96UmmIHQ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_IFrCYwCT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "fldr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_HA2PK7Yx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "smthr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_vmhAXEGP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "lmtr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_9v2n5T9E_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 512.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_KcCE9QCM, 0, m, &cVar_KcCE9QCM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_2L4OhUAt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_XTEu5phC, 0, m, &cVar_XTEu5phC_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_6pLq9WCy, 0, m, &cVar_6pLq9WCy_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_iDWrxXdj, 0, m, &cVar_iDWrxXdj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_g1y33AJH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 50.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_6N78UjYa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dS7yf57A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "mix");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_07PL4HOW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "sqnc");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_312BWiZO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "crshr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cBinop_dKl6HIsE_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_s7gCHDLK, 0, m, &cVar_s7gCHDLK_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_N1BvDbWD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "fldr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_lM2GMkcC_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "smthr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_ExknxxEi_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "lmtr");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_obnvhoqt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 512.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_NLJ5dtoh, 0, m, &cVar_NLJ5dtoh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_3fjLN9TB_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 1.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_atwbwh98, 0, m, &cVar_atwbwh98_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_UxYXUtWZ, 0, m, &cVar_UxYXUtWZ_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_MEeaWoyX, 0, m, &cVar_MEeaWoyX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_SG95B5Pd_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 50.0f);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_dKl6HIsE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_N96UVIz1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "mix");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_b2vyTGUK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_oLzkn6qr, 0, m, &cVar_oLzkn6qr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_dHlkn6WK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "gain");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_3Lkzwf8D_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_YsGlCJbz_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_hQYHxfWz_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_fZpL2DKt_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_n3jzoP0d, 0, m, &cVar_n3jzoP0d_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_fq8gGOEc_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "gain");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_XxdMCgLU_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_QqY8NFjw_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_5vQTqlUP_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cMsg_pupfkkbm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(1);
  msg_init(m, 1, msg_getTimestamp(n));
  msg_setFloat(m, 0, 0.0f);
  cVar_onMessage(_c, &Context(_c)->cVar_X6yszb3Y, 0, m, &cVar_X6yszb3Y_sendMessage);
}

void Heavy_WSTD_M3NGLR::cMsg_1r7S29s4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {
  HvMessage *m = nullptr;
  m = HV_MESSAGE_ON_STACK(2);
  msg_init(m, 2, msg_getTimestamp(n));
  msg_setSymbol(m, 0, "gain");
  msg_setElementToFrom(m, 1, n, 0);
  cSwitchcase_ZYsLHD91_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_PhtVSEec_onMessage(_c, NULL, 0, m, NULL);
  cSwitchcase_jdI06ZIc_onMessage(_c, NULL, 0, m, NULL);
}

void Heavy_WSTD_M3NGLR::cReceive_SotfbAwO_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_4AvLmp5R_sendMessage(_c, 0, m);
  cMsg_Xxi37HEJ_sendMessage(_c, 0, m);
  cMsg_uka3hlWw_sendMessage(_c, 0, m);
  cMsg_TZXtMJ9Y_sendMessage(_c, 0, m);
  cMsg_h2BYsSjj_sendMessage(_c, 0, m);
  cMsg_slxOIbhy_sendMessage(_c, 0, m);
  cVar_onMessage(_c, &Context(_c)->cVar_uz9xgzOh, 0, m, &cVar_uz9xgzOh_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_Ux8L40qC, 0, m, &cVar_Ux8L40qC_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_cl6eqZL4, 0, m, &cVar_cl6eqZL4_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_gcmj1Fey, 0, m, &cVar_gcmj1Fey_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_Td91SR6z, 0, m, &cVar_Td91SR6z_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_FjuiMfmI, 0, m, &cVar_FjuiMfmI_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_uU9rdlxg, 0, m, &cVar_uU9rdlxg_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_gAg6qjBM, 0, m, &cVar_gAg6qjBM_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_ISmXXI6S, 0, m, &cVar_ISmXXI6S_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_bXjj2nBB, 0, m, &cVar_bXjj2nBB_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_xxyCg8kW, 0, m, &cVar_xxyCg8kW_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_2o3gsLoE, 0, m, &cVar_2o3gsLoE_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_FMyaqghN, 0, m, &cVar_FMyaqghN_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_IvPeriUE, 0, m, &cVar_IvPeriUE_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_g7qup9cy, 0, m, &cVar_g7qup9cy_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_6zbdNNKu, 0, m, &cVar_6zbdNNKu_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_w1SnImWl, 0, m, &cVar_w1SnImWl_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_VDjwZrPZ, 0, m, &cVar_VDjwZrPZ_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_kD4dVJ2R, 0, m, &cVar_kD4dVJ2R_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_xdySbQS6, 0, m, &cVar_xdySbQS6_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_nlMVnmMc, 0, m, &cVar_nlMVnmMc_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_nK7oTbDD, 0, m, &cVar_nK7oTbDD_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_TF1GwCm9, 0, m, &cVar_TF1GwCm9_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_iCMuYcKz, 0, m, &cVar_iCMuYcKz_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_xJ0RrljN, 0, m, &cVar_xJ0RrljN_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_8FTilLlK, 0, m, &cVar_8FTilLlK_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_OUnPsAXz, 0, m, &cVar_OUnPsAXz_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_K1MN64FH, 0, m, &cVar_K1MN64FH_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_35veTBZ2, 0, m, &cVar_35veTBZ2_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_ZLxderQc, 0, m, &cVar_ZLxderQc_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_7gvbRUho, 0, m, &cVar_7gvbRUho_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_U8f1GYSP, 0, m, &cVar_U8f1GYSP_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_2XTjAuWf, 0, m, &cVar_2XTjAuWf_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_5HobttMZ, 0, m, &cVar_5HobttMZ_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_GcjTuTPW, 0, m, &cVar_GcjTuTPW_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_TqBzNWYb, 0, m, &cVar_TqBzNWYb_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_EMHmhOt3, 0, m, &cVar_EMHmhOt3_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_wnRiXiy4, 0, m, &cVar_wnRiXiy4_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_zIcGzlGa, 0, m, &cVar_zIcGzlGa_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_GAdyyTVt, 0, m, &cVar_GAdyyTVt_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_tjwCGUUl, 0, m, &cVar_tjwCGUUl_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_OkxvPbCq, 0, m, &cVar_OkxvPbCq_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_PP4tW2MC, 0, m, &cVar_PP4tW2MC_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_hhvHu7AA, 0, m, &cVar_hhvHu7AA_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_AehKCGix, 0, m, &cVar_AehKCGix_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_WEJNRm8c, 0, m, &cVar_WEJNRm8c_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_itik8I1i, 0, m, &cVar_itik8I1i_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_zJ0DTtW8, 0, m, &cVar_zJ0DTtW8_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_8qFyfhTt, 0, m, &cVar_8qFyfhTt_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_5WboOmCL, 0, m, &cVar_5WboOmCL_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_CmgJsT2G, 0, m, &cVar_CmgJsT2G_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_oY29yRWB, 0, m, &cVar_oY29yRWB_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_4tAy6BVl, 0, m, &cVar_4tAy6BVl_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_FHXPm0ZB, 0, m, &cVar_FHXPm0ZB_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_xweTUVZA, 0, m, &cVar_xweTUVZA_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_Za0eD1qF, 0, m, &cVar_Za0eD1qF_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_muKdJB5v, 0, m, &cVar_muKdJB5v_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_Rcs5mJTj, 0, m, &cVar_Rcs5mJTj_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_ocFZCKCX, 0, m, &cVar_ocFZCKCX_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_dpdFeYaw, 0, m, &cVar_dpdFeYaw_sendMessage);
  cVar_onMessage(_c, &Context(_c)->cVar_2IIWTMSO, 0, m, &cVar_2IIWTMSO_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_PZD91OeI_sendMessage);
  cMsg_XzgL3I3o_sendMessage(_c, 0, m);
  cMsg_dnBMwePR_sendMessage(_c, 0, m);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_S8wlpfjv_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_eBumaAjW_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_Wj2JpUW3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_pF65netw, 0, m, &cVar_pF65netw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_66Uw0S6A_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_nFdg7mi8, 0, m, &cVar_nFdg7mi8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_h6wmJtSK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_BAig76qw, 0, m, &cVar_BAig76qw_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_tTQhvy6q_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Ux8L40qC, 0, m, &cVar_Ux8L40qC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_1QQNA9M3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ueyLGSqM_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6qBeUsyh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_imIWhje3_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zT2j1wAY_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_hNuQUVtn_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Wd9udIyI_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_zT0kruKr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_43ZddC2v_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_e97CnG4f_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_uHyprpIa_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_xVBHH04v_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_A85DFdiu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_p64sWTJL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_sxexITpI_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_bH7gdO1c_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_XulKMGFa_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_4oZeZiLj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_H6MsTrQq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_IpwSQqMb, HV_BINOP_MULTIPLY, 0, m, &cBinop_IpwSQqMb_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_GdVfSCBI_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_wRsRKMUr_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_hgklUeqD_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_LgwrMfe7_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_S8OmFtjH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_MDOMLgPe_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_7wow9NCq_sendMessage);
  cMsg_maF2aIb6_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_MP6FeVQM_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_8E1V1uNq, HV_BINOP_MULTIPLY, 0, m, &cBinop_8E1V1uNq_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_PdTy37pm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_GoYp8bYa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_l4rrY6rq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ARLz0Qgv, HV_BINOP_MULTIPLY, 1, m, &cBinop_ARLz0Qgv_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_RYseM1Ry_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_UOYvWYah_sendMessage);
  cMsg_EHKzmQNe_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_SQamgB1D_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_1AcyTjjy, HV_BINOP_MULTIPLY, 0, m, &cBinop_1AcyTjjy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_GOQ8YMqR_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_stl2HTHH_sendMessage(_c, 0, m);
  cMsg_iBfdIBBv_sendMessage(_c, 0, m);
  cMsg_htRDScqH_sendMessage(_c, 0, m);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_FqAkigqz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_IqW4d26x_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_Bc2yqogs_sendMessage);
  cMsg_onkF0dFu_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_QeesCwR7_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Td91SR6z, 0, m, &cVar_Td91SR6z_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_vPtFdr5P_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_QtjZCZ64_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_aq66E8BJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_RFdY0G3h_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_6rkaZsed_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_wNQxy0Za_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_ae0EBhG8_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_2yKnTu0G_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_9SEoXHoU_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Jl9vc1mJ_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_e45dtl0N_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_7VtKQzLI_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_gnhAuVBc_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_BbgOv7yo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_zucpqpUi_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_q3ZRd5Ge_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_Sala7CIy_sendMessage);
  cCast_onMessage(_c, HV_CAST_FLOAT, 0, m, &cCast_3B5hXH9F_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_UT34Tkd2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_T9CfvvHV, HV_BINOP_MULTIPLY, 0, m, &cBinop_T9CfvvHV_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_ReTLutGG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_C9yKus9m_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_JnHsxTiR_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_q5WTJwHo_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_4AMm70Da_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_URR6d0NG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_KaNCp8Jk_sendMessage);
  cMsg_X9jCmQhl_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_N9y7codF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_CY1gO20Q, HV_BINOP_MULTIPLY, 0, m, &cBinop_CY1gO20Q_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_OLCCp2mq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_YZhNLIar_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_oJhZG77o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_ssTyQGf0, HV_BINOP_MULTIPLY, 1, m, &cBinop_ssTyQGf0_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -1.0f, 0, m, &cBinop_l092jTZK_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_AgoAhiOw_sendMessage);
  cMsg_zZAObVjr_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_scIqzx6W_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_onMessage(_c, &Context(_c)->cBinop_jCxaUzve, HV_BINOP_MULTIPLY, 0, m, &cBinop_jCxaUzve_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_Zx7qRj0a_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_nQ6k9lYQ_sendMessage(_c, 0, m);
  cMsg_bULR6DH9_sendMessage(_c, 0, m);
  cMsg_E2p3pDhh_sendMessage(_c, 0, m);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MULTIPLY, -2.0f, 0, m, &cBinop_sKJcIsIm_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_dsYqG1bK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_ADD, 1.0f, 0, m, &cBinop_XWZCxUU1_sendMessage);
  cMsg_3JTJ98Me_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_M7vmizJ4_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_lNtHYdg0, 0, m, &cVar_lNtHYdg0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_lk6laZNw_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Uudm7beN, 0, m, &cVar_Uudm7beN_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_MvFpJRdZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_dZUNub7K, 0, m, &cVar_dZUNub7K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_waxN58Ni_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_h8fgoD7X, 0, m, &cVar_h8fgoD7X_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_LdKvmYkh_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_B2SJsuMT, 0, m, &cVar_B2SJsuMT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_yRvUW9ON_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_vksrWJt7, 0, m, &cVar_vksrWJt7_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_RQ8lnqju_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_YUA5fLMD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_e2HxpxhL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 1.0f, 0, m, &cBinop_5OJXsxOJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_44TTkb8o_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_M0mXCjwu, 0, m, &cVar_M0mXCjwu_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_4gMPFIf6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_YU5vles9, 0, m, &cVar_YU5vles9_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_J4se2wAL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_wsDP8LGG_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_sRMDF2PX_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_1SuaqSbl_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_UixkLqjv_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_bnA3puum_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_aIQ9p1Wr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_2vBJDoM5_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_LyxbLvGZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_WnXNA3q8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_PXypk2IJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_J8JX4lev_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_Od1XQm8K, 0, m, &cVar_Od1XQm8K_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_S4Fliwp5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_CunFlHmT, 0, m, &cVar_CunFlHmT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_YV9QQVZn_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_xnuFMFB2_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_A4fKPXII_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_7ibaKm4w_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_4Nq7kGEW_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_kaPfBmdg_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_zXbmkPu5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_FMbJVgOU_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_uQocxxxn_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_8GN31U0Z_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_2kCOOm1v_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_3CoH8hpN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 1.0f, 0, m, &cBinop_mufOr9pE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_a9dr0eMq_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_IKg4xYBS, 0, m, &cVar_IKg4xYBS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_PPp5Evmm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_RrKG4xNY, 0, m, &cVar_RrKG4xNY_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_DNvwUqp6_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_oMSCdJKD_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_tUPKHbIY_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_htUfVxDG_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_siCr5DxI_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_7YohoXTS_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_x4AqsGDQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_GHd4XOyE_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_fiQxGdwy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_DY3OxtQJ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_sujyDzAq_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_tlvShJ1H_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_ZhevNCBF, 0, m, &cVar_ZhevNCBF_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_v2skW5DT_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_nPytwmId, 0, m, &cVar_nPytwmId_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_Wkpv9Vsl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_t5IZN8NH_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_XGxhpSj1_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_jSX6asPJ_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_JKj5JLUj_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_Fpdhw952_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_Fejs2us2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_3QvSI6ca_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_TEW7q8Ph_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_QbQDH3nH_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_BZgUcS0d_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_guIVuCz8_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_3e1g4Oco, 0, m, &cVar_3e1g4Oco_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_2wWaOjBZ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_KcCE9QCM, 0, m, &cVar_KcCE9QCM_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_ERaNPiVQ_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_XTEu5phC, 0, m, &cVar_XTEu5phC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_HNfz30dK_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_6pLq9WCy, 0, m, &cVar_6pLq9WCy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_g4fXhjYP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_iDWrxXdj, 0, m, &cVar_iDWrxXdj_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_rEDPzlnV_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_6N78UjYa_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_3lKBO5ei_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MIN, 1.0f, 0, m, &cBinop_bGJYQYOv_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_fbjDC3zp_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_V63JdNu0, 0, m, &cVar_V63JdNu0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_69voIQmF_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_HRZqLKqd, 0, m, &cVar_HRZqLKqd_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_JRXNrRlD_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_BWeEIQRI_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_aZYlD2gk_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_ZzFQmilo_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_hqm3oH2v_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_ZCv1Sv2f_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_g7Xyl5Ts_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_EJrvLW04_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_6IhfFONz_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_L2MVKoIl_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_rMzTZdqJ_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_nYfQhPMr_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_u6HPV7b8, 0, m, &cVar_u6HPV7b8_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_rnNNEjfA_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_tZFaMLqy, 0, m, &cVar_tZFaMLqy_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_AaWtmiYG_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_MAX, 2.0f, 0, m, &cBinop_1k94g1J0_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_STQOLGXP_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cSwitchcase_Br4ghokq_onMessage(_c, NULL, 0, m, NULL);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_jjNCXnLF_sendMessage);
  cCast_onMessage(_c, HV_CAST_BANG, 0, m, &cCast_GnDisZsT_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_8wn4a4Mx_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 0.0f, 0, m, &cBinop_yuBgq1sn_sendMessage);
  cBinop_k_onMessage(_c, NULL, HV_BINOP_EQ, 1.0f, 0, m, &cBinop_sopCKXxJ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_a9AmNNPu_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cMsg_u5oEmakO_sendMessage(_c, 0, m);
}

void Heavy_WSTD_M3NGLR::cReceive_rcFogwHW_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_KI0DoguC, 0, m, &cVar_KI0DoguC_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_6clWHYdo_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_NLJ5dtoh, 0, m, &cVar_NLJ5dtoh_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_HlKHEeVN_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_atwbwh98, 0, m, &cVar_atwbwh98_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_8iCGnXsL_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_UxYXUtWZ, 0, m, &cVar_UxYXUtWZ_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_edLW5Ok2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_MEeaWoyX, 0, m, &cVar_MEeaWoyX_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_tozHwuH9_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cBinop_k_onMessage(_c, NULL, HV_BINOP_DIVIDE, 100.0f, 0, m, &cBinop_dKl6HIsE_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_5TBiuZl5_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_oLzkn6qr, 0, m, &cVar_oLzkn6qr_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_5NV2HJGm_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_n3jzoP0d, 0, m, &cVar_n3jzoP0d_sendMessage);
}

void Heavy_WSTD_M3NGLR::cReceive_lzBJpNp2_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {
  cVar_onMessage(_c, &Context(_c)->cVar_X6yszb3Y, 0, m, &cVar_X6yszb3Y_sendMessage);
}




/*
 * Context Process Implementation
 */

int Heavy_WSTD_M3NGLR::process(float **inputBuffers, float **outputBuffers, int n) {
  while (hLp_hasData(&inQueue)) {
    hv_uint32_t numBytes = 0;
    ReceiverMessagePair *p = reinterpret_cast<ReceiverMessagePair *>(hLp_getReadBuffer(&inQueue, &numBytes));
    hv_assert(numBytes >= sizeof(ReceiverMessagePair));
    scheduleMessageForReceiver(p->receiverHash, &p->msg);
    hLp_consume(&inQueue);
  }

  sendBangToReceiver(0xDD21C0EB); // send to __hv_bang~ on next cycle
  const int n4 = n & ~HV_N_SIMD_MASK; // ensure that the block size is a multiple of HV_N_SIMD

  // temporary signal vars
  hv_bufferf_t Bf0, Bf1, Bf2, Bf3, Bf4, Bf5, Bf6, Bf7, Bf8, Bf9, Bf10, Bf11, Bf12, Bf13, Bf14, Bf15, Bf16, Bf17, Bf18, Bf19, Bf20;

  // input and output vars
  hv_bufferf_t O0, O1;
  hv_bufferf_t I0, I1;

  // declare and init the zero buffer
  hv_bufferf_t ZERO; __hv_zero_f(VOf(ZERO));

  hv_uint32_t nextBlock = blockStartTimestamp;
  for (int n = 0; n < n4; n += HV_N_SIMD) {

    // process all of the messages for this block
    nextBlock += HV_N_SIMD;
    while (mq_hasMessageBefore(&mq, nextBlock)) {
      MessageNode *const node = mq_peek(&mq);
      node->sendMessage(this, node->let, node->m);
      mq_pop(&mq);
    }

    // load input buffers
    __hv_load_f(inputBuffers[0]+n, VOf(I0));
    __hv_load_f(inputBuffers[1]+n, VOf(I1));

    // zero output buffers
    __hv_zero_f(VOf(O0));
    __hv_zero_f(VOf(O1));

    // process all signal functions
    __hv_line_f(&sLine_F0JuHtHo, VOf(Bf0));
    __hv_line_f(&sLine_wp2rbr12, VOf(Bf1));
    __hv_line_f(&sLine_Ik8joamg, VOf(Bf2));
    __hv_line_f(&sLine_DpAfafPl, VOf(Bf3));
    __hv_line_f(&sLine_xglAVFA6, VOf(Bf4));
    __hv_biquad_f(&sBiquad_s_zodZRcjv, VIf(I0), VIf(Bf0), VIf(Bf1), VIf(Bf2), VIf(Bf3), VIf(Bf4), VOf(Bf4));
    __hv_line_f(&sLine_lZjeicNW, VOf(Bf3));
    __hv_mul_f(VIf(Bf4), VIf(Bf3), VOf(Bf3));
    __hv_line_f(&sLine_kimGO7Ur, VOf(Bf4));
    __hv_zero_f(VOf(Bf2));
    __hv_line_f(&sLine_nvirZ593, VOf(Bf1));
    __hv_line_f(&sLine_xnRwmx5a, VOf(Bf0));
    __hv_line_f(&sLine_l6oALZOK, VOf(Bf5));
    __hv_biquad_f(&sBiquad_s_Rh0i4upA, VIf(I0), VIf(Bf4), VIf(Bf2), VIf(Bf1), VIf(Bf0), VIf(Bf5), VOf(Bf5));
    __hv_line_f(&sLine_Rf2hEX6Q, VOf(Bf0));
    __hv_mul_f(VIf(Bf5), VIf(Bf0), VOf(Bf0));
    __hv_line_f(&sLine_ZJPBoN8j, VOf(Bf5));
    __hv_line_f(&sLine_dd6zbHbI, VOf(Bf1));
    __hv_line_f(&sLine_fnxLh9cA, VOf(Bf2));
    __hv_line_f(&sLine_IZ2Om1Jb, VOf(Bf4));
    __hv_line_f(&sLine_JINsEOAB, VOf(Bf6));
    __hv_biquad_f(&sBiquad_s_Jo96Moux, VIf(I0), VIf(Bf5), VIf(Bf1), VIf(Bf2), VIf(Bf4), VIf(Bf6), VOf(Bf6));
    __hv_line_f(&sLine_7rgd3VNh, VOf(Bf4));
    __hv_mul_f(VIf(Bf6), VIf(Bf4), VOf(Bf4));
    __hv_line_f(&sLine_ZixC1Pet, VOf(Bf6));
    __hv_line_f(&sLine_S4rC6zLn, VOf(Bf2));
    __hv_line_f(&sLine_gVCakBg5, VOf(Bf1));
    __hv_line_f(&sLine_WEkTRzNX, VOf(Bf5));
    __hv_line_f(&sLine_BYtOCRgy, VOf(Bf7));
    __hv_biquad_f(&sBiquad_s_Hi1jQCzP, VIf(I1), VIf(Bf6), VIf(Bf2), VIf(Bf1), VIf(Bf5), VIf(Bf7), VOf(Bf7));
    __hv_line_f(&sLine_wpzdnGce, VOf(Bf5));
    __hv_mul_f(VIf(Bf7), VIf(Bf5), VOf(Bf5));
    __hv_line_f(&sLine_bx236IpN, VOf(Bf7));
    __hv_zero_f(VOf(Bf1));
    __hv_line_f(&sLine_j2GoFRCw, VOf(Bf2));
    __hv_line_f(&sLine_NlY0VCZV, VOf(Bf6));
    __hv_line_f(&sLine_dkHqOCXr, VOf(Bf8));
    __hv_biquad_f(&sBiquad_s_Aw4Dg0sr, VIf(I1), VIf(Bf7), VIf(Bf1), VIf(Bf2), VIf(Bf6), VIf(Bf8), VOf(Bf8));
    __hv_line_f(&sLine_LCfTd9ID, VOf(Bf6));
    __hv_mul_f(VIf(Bf8), VIf(Bf6), VOf(Bf6));
    __hv_line_f(&sLine_4VaKHD1e, VOf(Bf8));
    __hv_line_f(&sLine_DciG475c, VOf(Bf2));
    __hv_line_f(&sLine_0WUCz3An, VOf(Bf1));
    __hv_line_f(&sLine_2er8Yv1c, VOf(Bf7));
    __hv_line_f(&sLine_G5R6lu62, VOf(Bf9));
    __hv_biquad_f(&sBiquad_s_jRhPNL2g, VIf(I1), VIf(Bf8), VIf(Bf2), VIf(Bf1), VIf(Bf7), VIf(Bf9), VOf(Bf9));
    __hv_line_f(&sLine_3F4GMRkF, VOf(Bf7));
    __hv_mul_f(VIf(Bf9), VIf(Bf7), VOf(Bf7));
    __hv_varwrite_f(&sVarf_41BxaYgu, VIf(Bf0));
    __hv_varwrite_f(&sVarf_pFxFrJAB, VIf(Bf6));
    __hv_varwrite_f(&sVarf_pqyGbl7v, VIf(Bf4));
    __hv_varwrite_f(&sVarf_v0TpMypi, VIf(Bf7));
    __hv_varread_f(&sVarf_1nmUGqu9, VOf(Bf7));
    __hv_varread_f(&sVarf_EbruvE0R, VOf(Bf4));
    __hv_varread_f(&sVarf_JBIq6J7w, VOf(Bf6));
    __hv_varread_f(&sVarf_uT3DcEaj, VOf(Bf0));
    __hv_varread_f(&sVarf_3x25e2jw, VOf(Bf9));
    __hv_varread_f(&sVarf_n28bgRkl, VOf(Bf1));
    __hv_line_f(&sLine_rby24EH1, VOf(Bf2));
    __hv_line_f(&sLine_3k5g1IdH, VOf(Bf8));
    __hv_mul_f(VIf(Bf0), VIf(Bf8), VOf(Bf0));
    __hv_fma_f(VIf(Bf6), VIf(Bf2), VIf(Bf0), VOf(Bf0));
    __hv_mul_f(VIf(Bf1), VIf(Bf8), VOf(Bf8));
    __hv_fma_f(VIf(Bf9), VIf(Bf2), VIf(Bf8), VOf(Bf8));
    __hv_varwrite_f(&sVarf_JBIq6J7w, VIf(Bf7));
    __hv_varwrite_f(&sVarf_3x25e2jw, VIf(Bf4));
    __hv_varread_f(&sVarf_MtkhUlyb, VOf(Bf2));
    __hv_varread_f(&sVarf_EFqy6zps, VOf(Bf9));
    __hv_varread_f(&sVarf_V2V8jseP, VOf(Bf1));
    __hv_varread_f(&sVarf_WxH6j9Pq, VOf(Bf6));
    __hv_varread_f(&sVarf_1XkO61QH, VOf(Bf10));
    __hv_varread_f(&sVarf_pMLCtIQp, VOf(Bf11));
    __hv_varread_f(&sVarf_tQAEmRSA, VOf(Bf12));
    __hv_varwrite_f(&sVarf_jGYsXzV5, VIf(Bf7));
    __hv_varread_f(&sVarf_HdSGIVvE, VOf(Bf13));
    __hv_line_f(&sLine_tI24KERb, VOf(Bf14));
    __hv_mul_f(VIf(Bf13), VIf(Bf14), VOf(Bf14));
    __hv_varread_f(&sVarf_peXENc61, VOf(Bf15));
    __hv_line_f(&sLine_BVWA1YXY, VOf(Bf16));
    __hv_mul_f(VIf(Bf15), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_gmMYFTyo, VOf(Bf14));
    __hv_line_f(&sLine_MoU8urdf, VOf(Bf17));
    __hv_mul_f(VIf(Bf14), VIf(Bf17), VOf(Bf17));
    __hv_add_f(VIf(Bf16), VIf(Bf17), VOf(Bf17));
    __hv_varwrite_f(&sVarf_WxH6j9Pq, VIf(Bf17));
    __hv_varread_f(&sVarf_jGYsXzV5, VOf(Bf17));
    __hv_line_f(&sLine_QTfLgJMX, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_orqh9TpR, VOf(Bf18));
    __hv_mul_f(VIf(Bf14), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf16), VIf(Bf18), VOf(Bf18));
    __hv_line_f(&sLine_NWlJaD2v, VOf(Bf16));
    __hv_mul_f(VIf(Bf15), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf18), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_1XkO61QH, VIf(Bf16));
    __hv_line_f(&sLine_UERuc2TT, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_qEVXWEBh, VOf(Bf18));
    __hv_mul_f(VIf(Bf13), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf16), VIf(Bf18), VOf(Bf18));
    __hv_line_f(&sLine_TxjPaoTU, VOf(Bf16));
    __hv_mul_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf18), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_pMLCtIQp, VIf(Bf16));
    __hv_line_f(&sLine_vGHuA5zu, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_sQzKqn9s, VOf(Bf17));
    __hv_mul_f(VIf(Bf15), VIf(Bf17), VOf(Bf17));
    __hv_add_f(VIf(Bf16), VIf(Bf17), VOf(Bf17));
    __hv_line_f(&sLine_CoxZr5Ze, VOf(Bf16));
    __hv_mul_f(VIf(Bf13), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_tQAEmRSA, VIf(Bf16));
    __hv_varwrite_f(&sVarf_HdSGIVvE, VIf(Bf2));
    __hv_varwrite_f(&sVarf_peXENc61, VIf(Bf9));
    __hv_varwrite_f(&sVarf_gmMYFTyo, VIf(Bf1));
    sEnv_process(this, &sEnv_7DcPoE4W, VIf(Bf6), &sEnv_7DcPoE4W_sendMessage);
    sEnv_process(this, &sEnv_opuIzkiy, VIf(Bf7), &sEnv_opuIzkiy_sendMessage);
    __hv_line_f(&sLine_kx2wIf3T, VOf(Bf7));
    __hv_var_k_f(VOf(Bf1), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf7), VIf(Bf1), VOf(Bf1));
    __hv_var_k_f(VOf(Bf7), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf1), VIf(Bf7), VOf(Bf7));
    __hv_exp_f(VIf(Bf7), VOf(Bf7));
    __hv_mul_f(VIf(Bf6), VIf(Bf7), VOf(Bf7));
    __hv_line_f(&sLine_YtDohQAi, VOf(Bf6));
    __hv_mul_f(VIf(Bf7), VIf(Bf6), VOf(Bf6));
    __hv_line_f(&sLine_65Wdeb4Z, VOf(Bf7));
    __hv_var_k_f(VOf(Bf1), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf7), VIf(Bf1), VOf(Bf1));
    __hv_var_k_f(VOf(Bf7), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf1), VIf(Bf7), VOf(Bf7));
    __hv_exp_f(VIf(Bf7), VOf(Bf7));
    __hv_mul_f(VIf(Bf6), VIf(Bf7), VOf(Bf7));
    __hv_varread_f(&sVarf_V3ETqKtH, VOf(Bf6));
    __hv_line_f(&sLine_qgzzHDCM, VOf(Bf1));
    __hv_var_k_f(VOf(Bf9), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf1), VIf(Bf9), VOf(Bf9));
    __hv_var_k_f(VOf(Bf1), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf6), VIf(Bf9), VIf(Bf1), VOf(Bf1));
    __hv_floor_f(VIf(Bf1), VOf(Bf6));
    __hv_sub_f(VIf(Bf1), VIf(Bf6), VOf(Bf6));
    __hv_sub_f(VIf(Bf1), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf6), VIf(Bf9), VOf(Bf9));
    __hv_varwrite_f(&sVarf_MtkhUlyb, VIf(Bf9));
    __hv_varwrite_f(&sVarf_V3ETqKtH, VIf(Bf10));
    __hv_varwrite_f(&sVarf_tuMg71Ok, VIf(Bf11));
    __hv_varwrite_f(&sVarf_8Jy9FumC, VIf(Bf12));
    __hv_varread_f(&sVarf_8Jy9FumC, VOf(Bf12));
    __hv_line_f(&sLine_1RbEMkiL, VOf(Bf11));
    __hv_mul_f(VIf(Bf12), VIf(Bf11), VOf(Bf11));
    __hv_var_k_f(VOf(Bf12), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf11), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf11), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf12), VIf(Bf11), VOf(Bf11));
    __hv_mul_f(VIf(Bf11), VIf(Bf11), VOf(Bf12));
    __hv_var_k_f(VOf(Bf10), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf12), VIf(Bf10), VOf(Bf9));
    __hv_var_k_f(VOf(Bf6), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf12), VIf(Bf6), VIf(Bf10), VOf(Bf10));
    __hv_div_f(VIf(Bf9), VIf(Bf10), VOf(Bf10));
    __hv_mul_f(VIf(Bf11), VIf(Bf10), VOf(Bf10));
    __hv_varwrite_f(&sVarf_V2V8jseP, VIf(Bf10));
    __hv_varread_f(&sVarf_tuMg71Ok, VOf(Bf10));
    __hv_line_f(&sLine_dqMzI0vv, VOf(Bf11));
    __hv_var_k_f(VOf(Bf9), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf10), VIf(Bf11), VIf(Bf9), VOf(Bf9));
    __hv_var_k_f(VOf(Bf11), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf9), VIf(Bf11), VOf(Bf11));
    __hv_floor_f(VIf(Bf11), VOf(Bf9));
    __hv_sub_f(VIf(Bf11), VIf(Bf9), VOf(Bf9));
    __hv_var_k_f(VOf(Bf11), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf9), VIf(Bf11), VOf(Bf11));
    __hv_abs_f(VIf(Bf11), VOf(Bf11));
    __hv_var_k_f(VOf(Bf9), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf10), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf11), VIf(Bf9), VIf(Bf10), VOf(Bf10));
    __hv_neg_f(VIf(Bf10), VOf(Bf10));
    __hv_varwrite_f(&sVarf_EFqy6zps, VIf(Bf10));
    __hv_varwrite_f(&sVarf_uT3DcEaj, VIf(Bf7));
    __hv_varread_f(&sVarf_lSwq8mza, VOf(Bf7));
    __hv_varread_f(&sVarf_LSwUi9Zg, VOf(Bf10));
    __hv_varread_f(&sVarf_jomfJ2KG, VOf(Bf9));
    __hv_varread_f(&sVarf_a4KnZP56, VOf(Bf11));
    __hv_varread_f(&sVarf_taRtdGB2, VOf(Bf6));
    __hv_varread_f(&sVarf_YpoOV4SO, VOf(Bf12));
    __hv_varread_f(&sVarf_Zmj32A7k, VOf(Bf1));
    __hv_varwrite_f(&sVarf_kMHdIB36, VIf(Bf4));
    __hv_varread_f(&sVarf_kx32E46k, VOf(Bf2));
    __hv_line_f(&sLine_pp5JfdTE, VOf(Bf16));
    __hv_mul_f(VIf(Bf2), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_ObU8hxTg, VOf(Bf17));
    __hv_line_f(&sLine_YV3v2uOo, VOf(Bf13));
    __hv_mul_f(VIf(Bf17), VIf(Bf13), VOf(Bf13));
    __hv_add_f(VIf(Bf16), VIf(Bf13), VOf(Bf13));
    __hv_varread_f(&sVarf_b0hNdWBf, VOf(Bf16));
    __hv_line_f(&sLine_QOge56TJ, VOf(Bf15));
    __hv_mul_f(VIf(Bf16), VIf(Bf15), VOf(Bf15));
    __hv_add_f(VIf(Bf13), VIf(Bf15), VOf(Bf15));
    __hv_varwrite_f(&sVarf_a4KnZP56, VIf(Bf15));
    __hv_varread_f(&sVarf_kMHdIB36, VOf(Bf15));
    __hv_line_f(&sLine_WyFfcG1z, VOf(Bf13));
    __hv_mul_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_line_f(&sLine_ZPW00dSA, VOf(Bf18));
    __hv_mul_f(VIf(Bf16), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf13), VIf(Bf18), VOf(Bf18));
    __hv_line_f(&sLine_gBD9a38k, VOf(Bf13));
    __hv_mul_f(VIf(Bf17), VIf(Bf13), VOf(Bf13));
    __hv_add_f(VIf(Bf18), VIf(Bf13), VOf(Bf13));
    __hv_varwrite_f(&sVarf_taRtdGB2, VIf(Bf13));
    __hv_line_f(&sLine_K9QPJHn2, VOf(Bf13));
    __hv_mul_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_line_f(&sLine_UCPF3frM, VOf(Bf18));
    __hv_mul_f(VIf(Bf2), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf13), VIf(Bf18), VOf(Bf18));
    __hv_line_f(&sLine_f0HiPJD1, VOf(Bf13));
    __hv_mul_f(VIf(Bf16), VIf(Bf13), VOf(Bf13));
    __hv_add_f(VIf(Bf18), VIf(Bf13), VOf(Bf13));
    __hv_varwrite_f(&sVarf_YpoOV4SO, VIf(Bf13));
    __hv_line_f(&sLine_JpdM51Pu, VOf(Bf13));
    __hv_mul_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_line_f(&sLine_gCHyjQC6, VOf(Bf15));
    __hv_mul_f(VIf(Bf17), VIf(Bf15), VOf(Bf15));
    __hv_add_f(VIf(Bf13), VIf(Bf15), VOf(Bf15));
    __hv_line_f(&sLine_BizsZvdS, VOf(Bf13));
    __hv_mul_f(VIf(Bf2), VIf(Bf13), VOf(Bf13));
    __hv_add_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_varwrite_f(&sVarf_Zmj32A7k, VIf(Bf13));
    __hv_varwrite_f(&sVarf_kx32E46k, VIf(Bf7));
    __hv_varwrite_f(&sVarf_ObU8hxTg, VIf(Bf10));
    __hv_varwrite_f(&sVarf_b0hNdWBf, VIf(Bf9));
    sEnv_process(this, &sEnv_1TvLgphb, VIf(Bf11), &sEnv_1TvLgphb_sendMessage);
    sEnv_process(this, &sEnv_p5pbPN1Q, VIf(Bf4), &sEnv_p5pbPN1Q_sendMessage);
    __hv_line_f(&sLine_UgCbAvGG, VOf(Bf4));
    __hv_var_k_f(VOf(Bf9), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf9), VOf(Bf9));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf9), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf11), VIf(Bf4), VOf(Bf4));
    __hv_line_f(&sLine_8SoWHIbM, VOf(Bf11));
    __hv_mul_f(VIf(Bf4), VIf(Bf11), VOf(Bf11));
    __hv_line_f(&sLine_dhOv2gwo, VOf(Bf4));
    __hv_var_k_f(VOf(Bf9), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf9), VOf(Bf9));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf9), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf11), VIf(Bf4), VOf(Bf4));
    __hv_varread_f(&sVarf_T0xgSH8z, VOf(Bf11));
    __hv_line_f(&sLine_TG9l5fhg, VOf(Bf9));
    __hv_var_k_f(VOf(Bf10), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf9), VIf(Bf10), VOf(Bf10));
    __hv_var_k_f(VOf(Bf9), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf11), VIf(Bf10), VIf(Bf9), VOf(Bf9));
    __hv_floor_f(VIf(Bf9), VOf(Bf11));
    __hv_sub_f(VIf(Bf9), VIf(Bf11), VOf(Bf11));
    __hv_sub_f(VIf(Bf9), VIf(Bf11), VOf(Bf11));
    __hv_div_f(VIf(Bf11), VIf(Bf10), VOf(Bf10));
    __hv_varwrite_f(&sVarf_lSwq8mza, VIf(Bf10));
    __hv_varwrite_f(&sVarf_T0xgSH8z, VIf(Bf6));
    __hv_varwrite_f(&sVarf_Y7F7L9m7, VIf(Bf12));
    __hv_varwrite_f(&sVarf_FOfAkieo, VIf(Bf1));
    __hv_varread_f(&sVarf_FOfAkieo, VOf(Bf1));
    __hv_line_f(&sLine_3SItWbzE, VOf(Bf12));
    __hv_mul_f(VIf(Bf1), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf1), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf12), VIf(Bf1), VOf(Bf1));
    __hv_var_k_f(VOf(Bf12), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf1), VIf(Bf12), VOf(Bf12));
    __hv_mul_f(VIf(Bf12), VIf(Bf12), VOf(Bf1));
    __hv_var_k_f(VOf(Bf6), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf1), VIf(Bf6), VOf(Bf10));
    __hv_var_k_f(VOf(Bf11), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf1), VIf(Bf11), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf10), VIf(Bf6), VOf(Bf6));
    __hv_mul_f(VIf(Bf12), VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_jomfJ2KG, VIf(Bf6));
    __hv_varread_f(&sVarf_Y7F7L9m7, VOf(Bf6));
    __hv_line_f(&sLine_1ZQImshA, VOf(Bf12));
    __hv_var_k_f(VOf(Bf10), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf6), VIf(Bf12), VIf(Bf10), VOf(Bf10));
    __hv_var_k_f(VOf(Bf12), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf10), VIf(Bf12), VOf(Bf12));
    __hv_floor_f(VIf(Bf12), VOf(Bf10));
    __hv_sub_f(VIf(Bf12), VIf(Bf10), VOf(Bf10));
    __hv_var_k_f(VOf(Bf12), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf10), VIf(Bf12), VOf(Bf12));
    __hv_abs_f(VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf10), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf6), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf12), VIf(Bf10), VIf(Bf6), VOf(Bf6));
    __hv_neg_f(VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_LSwUi9Zg, VIf(Bf6));
    __hv_varwrite_f(&sVarf_n28bgRkl, VIf(Bf4));
    __hv_varwrite_f(&sVarf_HxKXweV4, VIf(Bf0));
    __hv_varread_f(&sVarf_41BxaYgu, VOf(Bf0));
    __hv_varread_f(&sVarf_pFxFrJAB, VOf(Bf4));
    __hv_varread_f(&sVarf_bk9mmres, VOf(Bf6));
    __hv_varread_f(&sVarf_DNU44Dvk, VOf(Bf10));
    __hv_varread_f(&sVarf_pc28uIXQ, VOf(Bf12));
    __hv_varread_f(&sVarf_48Gbr0d4, VOf(Bf11));
    __hv_line_f(&sLine_I7Tzfx8d, VOf(Bf1));
    __hv_line_f(&sLine_436i1fPs, VOf(Bf9));
    __hv_mul_f(VIf(Bf10), VIf(Bf9), VOf(Bf10));
    __hv_fma_f(VIf(Bf6), VIf(Bf1), VIf(Bf10), VOf(Bf10));
    __hv_mul_f(VIf(Bf11), VIf(Bf9), VOf(Bf9));
    __hv_fma_f(VIf(Bf12), VIf(Bf1), VIf(Bf9), VOf(Bf9));
    __hv_varwrite_f(&sVarf_bk9mmres, VIf(Bf0));
    __hv_varwrite_f(&sVarf_pc28uIXQ, VIf(Bf4));
    __hv_varread_f(&sVarf_4RrETmif, VOf(Bf1));
    __hv_varread_f(&sVarf_7RHfUzV5, VOf(Bf12));
    __hv_varread_f(&sVarf_P1GSsvHB, VOf(Bf11));
    __hv_varread_f(&sVarf_uejxXEBp, VOf(Bf6));
    __hv_varread_f(&sVarf_U9U3U44s, VOf(Bf7));
    __hv_varread_f(&sVarf_jgqydgcA, VOf(Bf13));
    __hv_varread_f(&sVarf_eW3QVMwf, VOf(Bf15));
    __hv_varwrite_f(&sVarf_wDDwfKGv, VIf(Bf0));
    __hv_varread_f(&sVarf_4lIqFWRH, VOf(Bf2));
    __hv_line_f(&sLine_yOW9VxDn, VOf(Bf17));
    __hv_mul_f(VIf(Bf2), VIf(Bf17), VOf(Bf17));
    __hv_varread_f(&sVarf_Flh0Ug9o, VOf(Bf18));
    __hv_line_f(&sLine_XvhwWoGj, VOf(Bf16));
    __hv_mul_f(VIf(Bf18), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_aG2OgZqR, VOf(Bf17));
    __hv_line_f(&sLine_TFhGkTXs, VOf(Bf14));
    __hv_mul_f(VIf(Bf17), VIf(Bf14), VOf(Bf14));
    __hv_add_f(VIf(Bf16), VIf(Bf14), VOf(Bf14));
    __hv_varwrite_f(&sVarf_uejxXEBp, VIf(Bf14));
    __hv_varread_f(&sVarf_wDDwfKGv, VOf(Bf14));
    __hv_line_f(&sLine_NoLSESq7, VOf(Bf16));
    __hv_mul_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_hT4P1smt, VOf(Bf19));
    __hv_mul_f(VIf(Bf17), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf16), VIf(Bf19), VOf(Bf19));
    __hv_line_f(&sLine_hn3MwQZ7, VOf(Bf16));
    __hv_mul_f(VIf(Bf18), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf19), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_U9U3U44s, VIf(Bf16));
    __hv_line_f(&sLine_9zwD0dMe, VOf(Bf16));
    __hv_mul_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_CTmP5gSj, VOf(Bf19));
    __hv_mul_f(VIf(Bf2), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf16), VIf(Bf19), VOf(Bf19));
    __hv_line_f(&sLine_gsvLugpM, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf19), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_jgqydgcA, VIf(Bf16));
    __hv_line_f(&sLine_uuoTx4ER, VOf(Bf16));
    __hv_mul_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_50dw8X8z, VOf(Bf14));
    __hv_mul_f(VIf(Bf18), VIf(Bf14), VOf(Bf14));
    __hv_add_f(VIf(Bf16), VIf(Bf14), VOf(Bf14));
    __hv_line_f(&sLine_6UeFb5mq, VOf(Bf16));
    __hv_mul_f(VIf(Bf2), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_eW3QVMwf, VIf(Bf16));
    __hv_varwrite_f(&sVarf_4lIqFWRH, VIf(Bf1));
    __hv_varwrite_f(&sVarf_Flh0Ug9o, VIf(Bf12));
    __hv_varwrite_f(&sVarf_aG2OgZqR, VIf(Bf11));
    sEnv_process(this, &sEnv_002yFlyr, VIf(Bf6), &sEnv_002yFlyr_sendMessage);
    sEnv_process(this, &sEnv_EYQBnZ5o, VIf(Bf0), &sEnv_EYQBnZ5o_sendMessage);
    __hv_line_f(&sLine_QJkbFjMq, VOf(Bf0));
    __hv_var_k_f(VOf(Bf11), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf0), VIf(Bf11), VOf(Bf11));
    __hv_var_k_f(VOf(Bf0), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf11), VIf(Bf0), VOf(Bf0));
    __hv_exp_f(VIf(Bf0), VOf(Bf0));
    __hv_mul_f(VIf(Bf6), VIf(Bf0), VOf(Bf0));
    __hv_line_f(&sLine_ei6ijquT, VOf(Bf6));
    __hv_mul_f(VIf(Bf0), VIf(Bf6), VOf(Bf6));
    __hv_line_f(&sLine_q3t8fdtk, VOf(Bf0));
    __hv_var_k_f(VOf(Bf11), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf0), VIf(Bf11), VOf(Bf11));
    __hv_var_k_f(VOf(Bf0), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf11), VIf(Bf0), VOf(Bf0));
    __hv_exp_f(VIf(Bf0), VOf(Bf0));
    __hv_mul_f(VIf(Bf6), VIf(Bf0), VOf(Bf0));
    __hv_varread_f(&sVarf_2Ndn9KQG, VOf(Bf6));
    __hv_line_f(&sLine_LGpLTtuO, VOf(Bf11));
    __hv_var_k_f(VOf(Bf12), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf11), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf11), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf6), VIf(Bf12), VIf(Bf11), VOf(Bf11));
    __hv_floor_f(VIf(Bf11), VOf(Bf6));
    __hv_sub_f(VIf(Bf11), VIf(Bf6), VOf(Bf6));
    __hv_sub_f(VIf(Bf11), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf6), VIf(Bf12), VOf(Bf12));
    __hv_varwrite_f(&sVarf_4RrETmif, VIf(Bf12));
    __hv_varwrite_f(&sVarf_2Ndn9KQG, VIf(Bf7));
    __hv_varwrite_f(&sVarf_JmPsMI7s, VIf(Bf13));
    __hv_varwrite_f(&sVarf_Vg5GgGs1, VIf(Bf15));
    __hv_varread_f(&sVarf_Vg5GgGs1, VOf(Bf15));
    __hv_line_f(&sLine_adrmUAO1, VOf(Bf13));
    __hv_mul_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_var_k_f(VOf(Bf15), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf13), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf13), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_mul_f(VIf(Bf13), VIf(Bf13), VOf(Bf15));
    __hv_var_k_f(VOf(Bf7), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf15), VIf(Bf7), VOf(Bf12));
    __hv_var_k_f(VOf(Bf6), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf15), VIf(Bf6), VIf(Bf7), VOf(Bf7));
    __hv_div_f(VIf(Bf12), VIf(Bf7), VOf(Bf7));
    __hv_mul_f(VIf(Bf13), VIf(Bf7), VOf(Bf7));
    __hv_varwrite_f(&sVarf_P1GSsvHB, VIf(Bf7));
    __hv_varread_f(&sVarf_JmPsMI7s, VOf(Bf7));
    __hv_line_f(&sLine_VSisnH6c, VOf(Bf13));
    __hv_var_k_f(VOf(Bf12), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf7), VIf(Bf13), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf13), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf12), VIf(Bf13), VOf(Bf13));
    __hv_floor_f(VIf(Bf13), VOf(Bf12));
    __hv_sub_f(VIf(Bf13), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf13), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf12), VIf(Bf13), VOf(Bf13));
    __hv_abs_f(VIf(Bf13), VOf(Bf13));
    __hv_var_k_f(VOf(Bf12), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf7), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf13), VIf(Bf12), VIf(Bf7), VOf(Bf7));
    __hv_neg_f(VIf(Bf7), VOf(Bf7));
    __hv_varwrite_f(&sVarf_7RHfUzV5, VIf(Bf7));
    __hv_varwrite_f(&sVarf_DNU44Dvk, VIf(Bf0));
    __hv_varread_f(&sVarf_8rWqOBQG, VOf(Bf0));
    __hv_varread_f(&sVarf_cn3rsmrh, VOf(Bf7));
    __hv_varread_f(&sVarf_za0OzJrY, VOf(Bf12));
    __hv_varread_f(&sVarf_EhlJ6qIm, VOf(Bf13));
    __hv_varread_f(&sVarf_ID0Faufj, VOf(Bf6));
    __hv_varread_f(&sVarf_Fwr3BRtP, VOf(Bf15));
    __hv_varread_f(&sVarf_T5EXXI9b, VOf(Bf11));
    __hv_varwrite_f(&sVarf_4xOJDS4F, VIf(Bf4));
    __hv_varread_f(&sVarf_YNpMpOkZ, VOf(Bf1));
    __hv_line_f(&sLine_T4AXGh4t, VOf(Bf16));
    __hv_mul_f(VIf(Bf1), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_4ByEU52T, VOf(Bf14));
    __hv_line_f(&sLine_Ed0waiOo, VOf(Bf2));
    __hv_mul_f(VIf(Bf14), VIf(Bf2), VOf(Bf2));
    __hv_add_f(VIf(Bf16), VIf(Bf2), VOf(Bf2));
    __hv_varread_f(&sVarf_X8KJBSnl, VOf(Bf16));
    __hv_line_f(&sLine_cVGc06ZX, VOf(Bf18));
    __hv_mul_f(VIf(Bf16), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf2), VIf(Bf18), VOf(Bf18));
    __hv_varwrite_f(&sVarf_EhlJ6qIm, VIf(Bf18));
    __hv_varread_f(&sVarf_4xOJDS4F, VOf(Bf18));
    __hv_line_f(&sLine_PBMJsRgR, VOf(Bf2));
    __hv_mul_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_line_f(&sLine_ogPoI4Gm, VOf(Bf19));
    __hv_mul_f(VIf(Bf16), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf2), VIf(Bf19), VOf(Bf19));
    __hv_line_f(&sLine_QDmTEdfD, VOf(Bf2));
    __hv_mul_f(VIf(Bf14), VIf(Bf2), VOf(Bf2));
    __hv_add_f(VIf(Bf19), VIf(Bf2), VOf(Bf2));
    __hv_varwrite_f(&sVarf_ID0Faufj, VIf(Bf2));
    __hv_line_f(&sLine_IJBRAbAK, VOf(Bf2));
    __hv_mul_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_line_f(&sLine_UXOMloGm, VOf(Bf19));
    __hv_mul_f(VIf(Bf1), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf2), VIf(Bf19), VOf(Bf19));
    __hv_line_f(&sLine_4ncutjso, VOf(Bf2));
    __hv_mul_f(VIf(Bf16), VIf(Bf2), VOf(Bf2));
    __hv_add_f(VIf(Bf19), VIf(Bf2), VOf(Bf2));
    __hv_varwrite_f(&sVarf_Fwr3BRtP, VIf(Bf2));
    __hv_line_f(&sLine_8iFLSOGV, VOf(Bf2));
    __hv_mul_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_line_f(&sLine_maSlUTpQ, VOf(Bf18));
    __hv_mul_f(VIf(Bf14), VIf(Bf18), VOf(Bf18));
    __hv_add_f(VIf(Bf2), VIf(Bf18), VOf(Bf18));
    __hv_line_f(&sLine_X2BBOH47, VOf(Bf2));
    __hv_mul_f(VIf(Bf1), VIf(Bf2), VOf(Bf2));
    __hv_add_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_varwrite_f(&sVarf_T5EXXI9b, VIf(Bf2));
    __hv_varwrite_f(&sVarf_YNpMpOkZ, VIf(Bf0));
    __hv_varwrite_f(&sVarf_4ByEU52T, VIf(Bf7));
    __hv_varwrite_f(&sVarf_X8KJBSnl, VIf(Bf12));
    sEnv_process(this, &sEnv_CMuR8qRn, VIf(Bf13), &sEnv_CMuR8qRn_sendMessage);
    sEnv_process(this, &sEnv_45tOAySj, VIf(Bf4), &sEnv_45tOAySj_sendMessage);
    __hv_line_f(&sLine_t0DjmaiP, VOf(Bf4));
    __hv_var_k_f(VOf(Bf12), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf12), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf13), VIf(Bf4), VOf(Bf4));
    __hv_line_f(&sLine_225hgjbz, VOf(Bf13));
    __hv_mul_f(VIf(Bf4), VIf(Bf13), VOf(Bf13));
    __hv_line_f(&sLine_6VexcZJF, VOf(Bf4));
    __hv_var_k_f(VOf(Bf12), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf12), VOf(Bf12));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf12), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf13), VIf(Bf4), VOf(Bf4));
    __hv_varread_f(&sVarf_Zv8fG32K, VOf(Bf13));
    __hv_line_f(&sLine_zT4WjDAp, VOf(Bf12));
    __hv_var_k_f(VOf(Bf7), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf12), VIf(Bf7), VOf(Bf7));
    __hv_var_k_f(VOf(Bf12), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf13), VIf(Bf7), VIf(Bf12), VOf(Bf12));
    __hv_floor_f(VIf(Bf12), VOf(Bf13));
    __hv_sub_f(VIf(Bf12), VIf(Bf13), VOf(Bf13));
    __hv_sub_f(VIf(Bf12), VIf(Bf13), VOf(Bf13));
    __hv_div_f(VIf(Bf13), VIf(Bf7), VOf(Bf7));
    __hv_varwrite_f(&sVarf_8rWqOBQG, VIf(Bf7));
    __hv_varwrite_f(&sVarf_Zv8fG32K, VIf(Bf6));
    __hv_varwrite_f(&sVarf_dnuNoNdK, VIf(Bf15));
    __hv_varwrite_f(&sVarf_PAZ7iB8d, VIf(Bf11));
    __hv_varread_f(&sVarf_PAZ7iB8d, VOf(Bf11));
    __hv_line_f(&sLine_D3xFLs9K, VOf(Bf15));
    __hv_mul_f(VIf(Bf11), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf11), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf15), VIf(Bf11), VOf(Bf11));
    __hv_var_k_f(VOf(Bf15), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf11), VIf(Bf15), VOf(Bf15));
    __hv_mul_f(VIf(Bf15), VIf(Bf15), VOf(Bf11));
    __hv_var_k_f(VOf(Bf6), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf11), VIf(Bf6), VOf(Bf7));
    __hv_var_k_f(VOf(Bf13), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf11), VIf(Bf13), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf7), VIf(Bf6), VOf(Bf6));
    __hv_mul_f(VIf(Bf15), VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_za0OzJrY, VIf(Bf6));
    __hv_varread_f(&sVarf_dnuNoNdK, VOf(Bf6));
    __hv_line_f(&sLine_n4y0vey0, VOf(Bf15));
    __hv_var_k_f(VOf(Bf7), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf6), VIf(Bf15), VIf(Bf7), VOf(Bf7));
    __hv_var_k_f(VOf(Bf15), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf7), VIf(Bf15), VOf(Bf15));
    __hv_floor_f(VIf(Bf15), VOf(Bf7));
    __hv_sub_f(VIf(Bf15), VIf(Bf7), VOf(Bf7));
    __hv_var_k_f(VOf(Bf15), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf7), VIf(Bf15), VOf(Bf15));
    __hv_abs_f(VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf7), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf6), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf15), VIf(Bf7), VIf(Bf6), VOf(Bf6));
    __hv_neg_f(VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_cn3rsmrh, VIf(Bf6));
    __hv_varwrite_f(&sVarf_48Gbr0d4, VIf(Bf4));
    __hv_varwrite_f(&sVarf_qp6MUgTI, VIf(Bf10));
    __hv_varread_f(&sVarf_pqyGbl7v, VOf(Bf10));
    __hv_varread_f(&sVarf_v0TpMypi, VOf(Bf4));
    __hv_varread_f(&sVarf_NaU2Rj6s, VOf(Bf6));
    __hv_varread_f(&sVarf_mjGnpXDN, VOf(Bf7));
    __hv_varread_f(&sVarf_XKKLtN6L, VOf(Bf15));
    __hv_varread_f(&sVarf_bxAz3npH, VOf(Bf13));
    __hv_line_f(&sLine_jkHkClBU, VOf(Bf11));
    __hv_line_f(&sLine_rbElxj1N, VOf(Bf12));
    __hv_mul_f(VIf(Bf7), VIf(Bf12), VOf(Bf7));
    __hv_fma_f(VIf(Bf6), VIf(Bf11), VIf(Bf7), VOf(Bf7));
    __hv_mul_f(VIf(Bf13), VIf(Bf12), VOf(Bf12));
    __hv_fma_f(VIf(Bf15), VIf(Bf11), VIf(Bf12), VOf(Bf12));
    __hv_varwrite_f(&sVarf_NaU2Rj6s, VIf(Bf10));
    __hv_varwrite_f(&sVarf_XKKLtN6L, VIf(Bf4));
    __hv_varread_f(&sVarf_LiCkyoL0, VOf(Bf11));
    __hv_varread_f(&sVarf_05FRLHZ2, VOf(Bf15));
    __hv_varread_f(&sVarf_oOVEUhU7, VOf(Bf13));
    __hv_varread_f(&sVarf_WuCVYk7a, VOf(Bf6));
    __hv_varread_f(&sVarf_ueUGLXBc, VOf(Bf0));
    __hv_varread_f(&sVarf_Gi6co6Sw, VOf(Bf2));
    __hv_varread_f(&sVarf_G1t2167a, VOf(Bf18));
    __hv_varwrite_f(&sVarf_2Id1rHJp, VIf(Bf10));
    __hv_varread_f(&sVarf_fPkbEyM4, VOf(Bf1));
    __hv_line_f(&sLine_YTWmwqf5, VOf(Bf14));
    __hv_mul_f(VIf(Bf1), VIf(Bf14), VOf(Bf14));
    __hv_varread_f(&sVarf_7Snmmtkh, VOf(Bf19));
    __hv_line_f(&sLine_8b1XMKYt, VOf(Bf16));
    __hv_mul_f(VIf(Bf19), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_VRWRTgoG, VOf(Bf14));
    __hv_line_f(&sLine_gpmk9514, VOf(Bf17));
    __hv_mul_f(VIf(Bf14), VIf(Bf17), VOf(Bf17));
    __hv_add_f(VIf(Bf16), VIf(Bf17), VOf(Bf17));
    __hv_varwrite_f(&sVarf_WuCVYk7a, VIf(Bf17));
    __hv_varread_f(&sVarf_2Id1rHJp, VOf(Bf17));
    __hv_line_f(&sLine_TkJqBeom, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_VnrnqskK, VOf(Bf20));
    __hv_mul_f(VIf(Bf14), VIf(Bf20), VOf(Bf20));
    __hv_add_f(VIf(Bf16), VIf(Bf20), VOf(Bf20));
    __hv_line_f(&sLine_6MXdMybS, VOf(Bf16));
    __hv_mul_f(VIf(Bf19), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf20), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_ueUGLXBc, VIf(Bf16));
    __hv_line_f(&sLine_JgktvltE, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_c6oIfRZl, VOf(Bf20));
    __hv_mul_f(VIf(Bf1), VIf(Bf20), VOf(Bf20));
    __hv_add_f(VIf(Bf16), VIf(Bf20), VOf(Bf20));
    __hv_line_f(&sLine_w0usi9ji, VOf(Bf16));
    __hv_mul_f(VIf(Bf14), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf20), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_Gi6co6Sw, VIf(Bf16));
    __hv_line_f(&sLine_JSIk82TP, VOf(Bf16));
    __hv_mul_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_line_f(&sLine_FqCHeqQA, VOf(Bf17));
    __hv_mul_f(VIf(Bf19), VIf(Bf17), VOf(Bf17));
    __hv_add_f(VIf(Bf16), VIf(Bf17), VOf(Bf17));
    __hv_line_f(&sLine_3ZMyjMzM, VOf(Bf16));
    __hv_mul_f(VIf(Bf1), VIf(Bf16), VOf(Bf16));
    __hv_add_f(VIf(Bf17), VIf(Bf16), VOf(Bf16));
    __hv_varwrite_f(&sVarf_G1t2167a, VIf(Bf16));
    __hv_varwrite_f(&sVarf_fPkbEyM4, VIf(Bf11));
    __hv_varwrite_f(&sVarf_7Snmmtkh, VIf(Bf15));
    __hv_varwrite_f(&sVarf_VRWRTgoG, VIf(Bf13));
    sEnv_process(this, &sEnv_UvKpHaWA, VIf(Bf6), &sEnv_UvKpHaWA_sendMessage);
    sEnv_process(this, &sEnv_xQgzju4H, VIf(Bf10), &sEnv_xQgzju4H_sendMessage);
    __hv_line_f(&sLine_2dWBcy6Z, VOf(Bf10));
    __hv_var_k_f(VOf(Bf13), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf10), VIf(Bf13), VOf(Bf13));
    __hv_var_k_f(VOf(Bf10), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf13), VIf(Bf10), VOf(Bf10));
    __hv_exp_f(VIf(Bf10), VOf(Bf10));
    __hv_mul_f(VIf(Bf6), VIf(Bf10), VOf(Bf10));
    __hv_line_f(&sLine_lLssYiPy, VOf(Bf6));
    __hv_mul_f(VIf(Bf10), VIf(Bf6), VOf(Bf6));
    __hv_line_f(&sLine_emHMr1W4, VOf(Bf10));
    __hv_var_k_f(VOf(Bf13), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf10), VIf(Bf13), VOf(Bf13));
    __hv_var_k_f(VOf(Bf10), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf13), VIf(Bf10), VOf(Bf10));
    __hv_exp_f(VIf(Bf10), VOf(Bf10));
    __hv_mul_f(VIf(Bf6), VIf(Bf10), VOf(Bf10));
    __hv_varread_f(&sVarf_z9KQRjps, VOf(Bf6));
    __hv_line_f(&sLine_Xt4uhHeO, VOf(Bf13));
    __hv_var_k_f(VOf(Bf15), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf13), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf13), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf6), VIf(Bf15), VIf(Bf13), VOf(Bf13));
    __hv_floor_f(VIf(Bf13), VOf(Bf6));
    __hv_sub_f(VIf(Bf13), VIf(Bf6), VOf(Bf6));
    __hv_sub_f(VIf(Bf13), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf6), VIf(Bf15), VOf(Bf15));
    __hv_varwrite_f(&sVarf_LiCkyoL0, VIf(Bf15));
    __hv_varwrite_f(&sVarf_z9KQRjps, VIf(Bf0));
    __hv_varwrite_f(&sVarf_VkqYtiKj, VIf(Bf2));
    __hv_varwrite_f(&sVarf_EBebXL78, VIf(Bf18));
    __hv_varread_f(&sVarf_EBebXL78, VOf(Bf18));
    __hv_line_f(&sLine_RLNFQdZ1, VOf(Bf2));
    __hv_mul_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_var_k_f(VOf(Bf18), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf2), VIf(Bf18), VOf(Bf18));
    __hv_var_k_f(VOf(Bf2), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf18), VIf(Bf2), VOf(Bf2));
    __hv_mul_f(VIf(Bf2), VIf(Bf2), VOf(Bf18));
    __hv_var_k_f(VOf(Bf0), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf18), VIf(Bf0), VOf(Bf15));
    __hv_var_k_f(VOf(Bf6), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf18), VIf(Bf6), VIf(Bf0), VOf(Bf0));
    __hv_div_f(VIf(Bf15), VIf(Bf0), VOf(Bf0));
    __hv_mul_f(VIf(Bf2), VIf(Bf0), VOf(Bf0));
    __hv_varwrite_f(&sVarf_oOVEUhU7, VIf(Bf0));
    __hv_varread_f(&sVarf_VkqYtiKj, VOf(Bf0));
    __hv_line_f(&sLine_DSPEe7yf, VOf(Bf2));
    __hv_var_k_f(VOf(Bf15), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf0), VIf(Bf2), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf2), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf15), VIf(Bf2), VOf(Bf2));
    __hv_floor_f(VIf(Bf2), VOf(Bf15));
    __hv_sub_f(VIf(Bf2), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf2), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf15), VIf(Bf2), VOf(Bf2));
    __hv_abs_f(VIf(Bf2), VOf(Bf2));
    __hv_var_k_f(VOf(Bf15), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf0), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf2), VIf(Bf15), VIf(Bf0), VOf(Bf0));
    __hv_neg_f(VIf(Bf0), VOf(Bf0));
    __hv_varwrite_f(&sVarf_05FRLHZ2, VIf(Bf0));
    __hv_varwrite_f(&sVarf_mjGnpXDN, VIf(Bf10));
    __hv_varread_f(&sVarf_7GfR5yr8, VOf(Bf10));
    __hv_varread_f(&sVarf_efaBIdJA, VOf(Bf0));
    __hv_varread_f(&sVarf_CH5wshEn, VOf(Bf15));
    __hv_varread_f(&sVarf_vpdEas9q, VOf(Bf2));
    __hv_varread_f(&sVarf_xRsmaKFr, VOf(Bf6));
    __hv_varread_f(&sVarf_HEmOZanm, VOf(Bf18));
    __hv_varread_f(&sVarf_4H4yP1z5, VOf(Bf13));
    __hv_varwrite_f(&sVarf_XAr338KV, VIf(Bf4));
    __hv_varread_f(&sVarf_kLtPkszU, VOf(Bf11));
    __hv_line_f(&sLine_pKRE2uPH, VOf(Bf16));
    __hv_mul_f(VIf(Bf11), VIf(Bf16), VOf(Bf16));
    __hv_varread_f(&sVarf_Z0ksGvbs, VOf(Bf17));
    __hv_line_f(&sLine_sj6MO7R2, VOf(Bf1));
    __hv_mul_f(VIf(Bf17), VIf(Bf1), VOf(Bf1));
    __hv_add_f(VIf(Bf16), VIf(Bf1), VOf(Bf1));
    __hv_varread_f(&sVarf_xNccdKfL, VOf(Bf16));
    __hv_line_f(&sLine_YtNcxyvu, VOf(Bf19));
    __hv_mul_f(VIf(Bf16), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf1), VIf(Bf19), VOf(Bf19));
    __hv_varwrite_f(&sVarf_vpdEas9q, VIf(Bf19));
    __hv_varread_f(&sVarf_XAr338KV, VOf(Bf19));
    __hv_line_f(&sLine_We3kXJi3, VOf(Bf1));
    __hv_mul_f(VIf(Bf19), VIf(Bf1), VOf(Bf1));
    __hv_line_f(&sLine_yw79kbKp, VOf(Bf20));
    __hv_mul_f(VIf(Bf16), VIf(Bf20), VOf(Bf20));
    __hv_add_f(VIf(Bf1), VIf(Bf20), VOf(Bf20));
    __hv_line_f(&sLine_coxYZNAo, VOf(Bf1));
    __hv_mul_f(VIf(Bf17), VIf(Bf1), VOf(Bf1));
    __hv_add_f(VIf(Bf20), VIf(Bf1), VOf(Bf1));
    __hv_varwrite_f(&sVarf_xRsmaKFr, VIf(Bf1));
    __hv_line_f(&sLine_KAUdhPPU, VOf(Bf1));
    __hv_mul_f(VIf(Bf19), VIf(Bf1), VOf(Bf1));
    __hv_line_f(&sLine_aG95Bhw8, VOf(Bf20));
    __hv_mul_f(VIf(Bf11), VIf(Bf20), VOf(Bf20));
    __hv_add_f(VIf(Bf1), VIf(Bf20), VOf(Bf20));
    __hv_line_f(&sLine_lOjst2XV, VOf(Bf1));
    __hv_mul_f(VIf(Bf16), VIf(Bf1), VOf(Bf1));
    __hv_add_f(VIf(Bf20), VIf(Bf1), VOf(Bf1));
    __hv_varwrite_f(&sVarf_HEmOZanm, VIf(Bf1));
    __hv_line_f(&sLine_OPvamDGw, VOf(Bf1));
    __hv_mul_f(VIf(Bf19), VIf(Bf1), VOf(Bf1));
    __hv_line_f(&sLine_B2BiVPLR, VOf(Bf19));
    __hv_mul_f(VIf(Bf17), VIf(Bf19), VOf(Bf19));
    __hv_add_f(VIf(Bf1), VIf(Bf19), VOf(Bf19));
    __hv_line_f(&sLine_y4bKytXG, VOf(Bf1));
    __hv_mul_f(VIf(Bf11), VIf(Bf1), VOf(Bf1));
    __hv_add_f(VIf(Bf19), VIf(Bf1), VOf(Bf1));
    __hv_varwrite_f(&sVarf_4H4yP1z5, VIf(Bf1));
    __hv_varwrite_f(&sVarf_kLtPkszU, VIf(Bf10));
    __hv_varwrite_f(&sVarf_Z0ksGvbs, VIf(Bf0));
    __hv_varwrite_f(&sVarf_xNccdKfL, VIf(Bf15));
    sEnv_process(this, &sEnv_8SZBwKcI, VIf(Bf2), &sEnv_8SZBwKcI_sendMessage);
    sEnv_process(this, &sEnv_bxy3cjVB, VIf(Bf4), &sEnv_bxy3cjVB_sendMessage);
    __hv_line_f(&sLine_AmjVNRMc, VOf(Bf4));
    __hv_var_k_f(VOf(Bf15), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf15), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf2), VIf(Bf4), VOf(Bf4));
    __hv_line_f(&sLine_qNgfQORs, VOf(Bf2));
    __hv_mul_f(VIf(Bf4), VIf(Bf2), VOf(Bf2));
    __hv_line_f(&sLine_15uNlUt4, VOf(Bf4));
    __hv_var_k_f(VOf(Bf15), 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f, 100.0f);
    __hv_sub_f(VIf(Bf4), VIf(Bf15), VOf(Bf15));
    __hv_var_k_f(VOf(Bf4), 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f, 0.115129254649702f);
    __hv_mul_f(VIf(Bf15), VIf(Bf4), VOf(Bf4));
    __hv_exp_f(VIf(Bf4), VOf(Bf4));
    __hv_mul_f(VIf(Bf2), VIf(Bf4), VOf(Bf4));
    __hv_varread_f(&sVarf_0LBjjQmS, VOf(Bf2));
    __hv_line_f(&sLine_hbYCTr0Q, VOf(Bf15));
    __hv_var_k_f(VOf(Bf0), 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f);
    __hv_div_f(VIf(Bf15), VIf(Bf0), VOf(Bf0));
    __hv_var_k_f(VOf(Bf15), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_fma_f(VIf(Bf2), VIf(Bf0), VIf(Bf15), VOf(Bf15));
    __hv_floor_f(VIf(Bf15), VOf(Bf2));
    __hv_sub_f(VIf(Bf15), VIf(Bf2), VOf(Bf2));
    __hv_sub_f(VIf(Bf15), VIf(Bf2), VOf(Bf2));
    __hv_div_f(VIf(Bf2), VIf(Bf0), VOf(Bf0));
    __hv_varwrite_f(&sVarf_7GfR5yr8, VIf(Bf0));
    __hv_varwrite_f(&sVarf_0LBjjQmS, VIf(Bf6));
    __hv_varwrite_f(&sVarf_saq1Q8WE, VIf(Bf18));
    __hv_varwrite_f(&sVarf_3MJfekF8, VIf(Bf13));
    __hv_varread_f(&sVarf_3MJfekF8, VOf(Bf13));
    __hv_line_f(&sLine_36L2nBRt, VOf(Bf18));
    __hv_mul_f(VIf(Bf13), VIf(Bf18), VOf(Bf18));
    __hv_var_k_f(VOf(Bf13), 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f, 3.0f);
    __hv_min_f(VIf(Bf18), VIf(Bf13), VOf(Bf13));
    __hv_var_k_f(VOf(Bf18), -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f, -3.0f);
    __hv_max_f(VIf(Bf13), VIf(Bf18), VOf(Bf18));
    __hv_mul_f(VIf(Bf18), VIf(Bf18), VOf(Bf13));
    __hv_var_k_f(VOf(Bf6), 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f, 27.0f);
    __hv_add_f(VIf(Bf13), VIf(Bf6), VOf(Bf0));
    __hv_var_k_f(VOf(Bf2), 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f, 9.0f);
    __hv_fma_f(VIf(Bf13), VIf(Bf2), VIf(Bf6), VOf(Bf6));
    __hv_div_f(VIf(Bf0), VIf(Bf6), VOf(Bf6));
    __hv_mul_f(VIf(Bf18), VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_CH5wshEn, VIf(Bf6));
    __hv_varread_f(&sVarf_saq1Q8WE, VOf(Bf6));
    __hv_line_f(&sLine_sbsm5hgo, VOf(Bf18));
    __hv_var_k_f(VOf(Bf0), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fma_f(VIf(Bf6), VIf(Bf18), VIf(Bf0), VOf(Bf0));
    __hv_var_k_f(VOf(Bf18), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_div_f(VIf(Bf0), VIf(Bf18), VOf(Bf18));
    __hv_floor_f(VIf(Bf18), VOf(Bf0));
    __hv_sub_f(VIf(Bf18), VIf(Bf0), VOf(Bf0));
    __hv_var_k_f(VOf(Bf18), 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f);
    __hv_sub_f(VIf(Bf0), VIf(Bf18), VOf(Bf18));
    __hv_abs_f(VIf(Bf18), VOf(Bf18));
    __hv_var_k_f(VOf(Bf0), 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f, 4.0f);
    __hv_var_k_f(VOf(Bf6), 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
    __hv_fms_f(VIf(Bf18), VIf(Bf0), VIf(Bf6), VOf(Bf6));
    __hv_neg_f(VIf(Bf6), VOf(Bf6));
    __hv_varwrite_f(&sVarf_efaBIdJA, VIf(Bf6));
    __hv_varwrite_f(&sVarf_bxAz3npH, VIf(Bf4));
    __hv_varwrite_f(&sVarf_na221FsL, VIf(Bf7));
    __hv_varwrite_f(&sVarf_sIlCiHDe, VIf(Bf8));
    __hv_varwrite_f(&sVarf_HRDkZKL3, VIf(Bf9));
    __hv_varwrite_f(&sVarf_D9WIgVKX, VIf(Bf12));
    __hv_varwrite_f(&sVarf_1nmUGqu9, VIf(Bf3));
    __hv_varwrite_f(&sVarf_EbruvE0R, VIf(Bf5));
    __hv_varread_f(&sVarf_sIlCiHDe, VOf(Bf5));
    __hv_varread_f(&sVarf_HRDkZKL3, VOf(Bf3));
    __hv_varread_f(&sVarf_D9WIgVKX, VOf(Bf12));
    __hv_add_f(VIf(Bf3), VIf(Bf12), VOf(Bf12));
    __hv_add_f(VIf(Bf5), VIf(Bf12), VOf(Bf12));
    __hv_add_f(VIf(Bf12), VIf(O1), VOf(O1));
    __hv_varread_f(&sVarf_HxKXweV4, VOf(Bf12));
    __hv_varread_f(&sVarf_qp6MUgTI, VOf(Bf5));
    __hv_varread_f(&sVarf_na221FsL, VOf(Bf3));
    __hv_add_f(VIf(Bf5), VIf(Bf3), VOf(Bf3));
    __hv_add_f(VIf(Bf12), VIf(Bf3), VOf(Bf3));
    __hv_add_f(VIf(Bf3), VIf(O0), VOf(O0));

    // save output vars to output buffer
    __hv_store_f(outputBuffers[0]+n, VIf(O0));
    __hv_store_f(outputBuffers[1]+n, VIf(O1));
  }

  blockStartTimestamp = nextBlock;

  return n4; // return the number of frames processed

}

int Heavy_WSTD_M3NGLR::processInline(float *inputBuffers, float *outputBuffers, int n4) {
  hv_assert(!(n4 & HV_N_SIMD_MASK)); // ensure that n4 is a multiple of HV_N_SIMD

  // define the heavy input buffer for 2 channel(s)
  float **const bIn = reinterpret_cast<float **>(hv_alloca(2*sizeof(float *)));
  bIn[0] = inputBuffers+(0*n4);
  bIn[1] = inputBuffers+(1*n4);

  // define the heavy output buffer for 2 channel(s)
  float **const bOut = reinterpret_cast<float **>(hv_alloca(2*sizeof(float *)));
  bOut[0] = outputBuffers+(0*n4);
  bOut[1] = outputBuffers+(1*n4);

  int n = process(bIn, bOut, n4);
  return n;
}

int Heavy_WSTD_M3NGLR::processInlineInterleaved(float *inputBuffers, float *outputBuffers, int n4) {
  hv_assert(n4 & ~HV_N_SIMD_MASK); // ensure that n4 is a multiple of HV_N_SIMD

  // define the heavy input buffer for 2 channel(s), uninterleave
  float *const bIn = reinterpret_cast<float *>(hv_alloca(2*n4*sizeof(float)));
  #if HV_SIMD_SSE || HV_SIMD_AVX
  for (int i = 0, j = 0; j < n4; j += 4, i += 8) {
    __m128 a = _mm_load_ps(inputBuffers+i);                // LRLR
    __m128 b = _mm_load_ps(inputBuffers+4+i);              // LRLR
    __m128 x = _mm_shuffle_ps(a, b, _MM_SHUFFLE(2,0,2,0)); // LLLL
    __m128 y = _mm_shuffle_ps(a, b, _MM_SHUFFLE(3,1,3,1)); // RRRR
    _mm_store_ps(bIn+j, x);
    _mm_store_ps(bIn+n4+j, y);
  }
  #elif HV_SIMD_NEON
  for (int i = 0, j = 0; j < n4; j += 4, i += 8) {
    float32x4x2_t a = vld2q_f32(inputBuffers+i); // load and uninterleave
    vst1q_f32(bIn+j, a.val[0]);
    vst1q_f32(bIn+n4+j, a.val[1]);
  }
  #else // HV_SIMD_NONE
  for (int j = 0; j < n4; ++j) {
    bIn[0*n4+j] = inputBuffers[0+2*j];
    bIn[1*n4+j] = inputBuffers[1+2*j];
  }
  #endif

  // define the heavy output buffer for 2 channel(s)
  float *const bOut = reinterpret_cast<float *>(hv_alloca(2*n4*sizeof(float)));

  int n = processInline(bIn, bOut, n4);

  // interleave the heavy output into the output buffer
  #if HV_SIMD_AVX
  for (int i = 0, j = 0; j < n4; j += 8, i += 16) {
    __m256 x = _mm256_load_ps(bOut+j);    // LLLLLLLL
    __m256 y = _mm256_load_ps(bOut+n4+j); // RRRRRRRR
    __m256 a = _mm256_unpacklo_ps(x, y);  // LRLRLRLR
    __m256 b = _mm256_unpackhi_ps(x, y);  // LRLRLRLR
    _mm256_store_ps(outputBuffers+i, a);
    _mm256_store_ps(outputBuffers+8+i, b);
  }
  #elif HV_SIMD_SSE
  for (int i = 0, j = 0; j < n4; j += 4, i += 8) {
    __m128 x = _mm_load_ps(bOut+j);    // LLLL
    __m128 y = _mm_load_ps(bOut+n4+j); // RRRR
    __m128 a = _mm_unpacklo_ps(x, y);  // LRLR
    __m128 b = _mm_unpackhi_ps(x, y);  // LRLR
    _mm_store_ps(outputBuffers+i, a);
    _mm_store_ps(outputBuffers+4+i, b);
  }
  #elif HV_SIMD_NEON
  // https://community.arm.com/groups/processors/blog/2012/03/13/coding-for-neon--part-5-rearranging-vectors
  for (int i = 0, j = 0; j < n4; j += 4, i += 8) {
    float32x4_t x = vld1q_f32(bOut+j);
    float32x4_t y = vld1q_f32(bOut+n4+j);
    float32x4x2_t z = {x, y};
    vst2q_f32(outputBuffers+i, z); // interleave and store
  }
  #else // HV_SIMD_NONE
  for (int i = 0; i < 2; ++i) {
    for (int j = 0; j < n4; ++j) {
      outputBuffers[i+2*j] = bOut[i*n4+j];
    }
  }
  #endif

  return n;
}
